/*
 * Domain.hpp
 *
 *  Created on: Sep 11, 2013
 *      Author: ckeme_01
 */

/**
 * \file Domain.hpp
 * \brief File to contain the Domain class.
 */
#ifndef DOMAIN_HPP_
#define DOMAIN_HPP_

// C++ header
#include <cstdlib>
#include <string>

/**
 * \file Domain.hpp
 * File containing the Domain class.
 */

namespace MDAT
{

/**
 * \brief Class to represent a domain.
 * \details The class is mainly designed for Pfam domains. It can store all the necessary values.
 */
class Domain
{

private:
	std::string _name;
	std::string _accession;
	size_t _seq_start;
	size_t _seq_end;
	size_t _env_start;
	size_t _env_end;
	size_t _hmm_start;
	size_t _hmm_end;
	int _repeated;
	double _evalue;
	int _id;
	int _original_position;

public:

	/**@{*/
	//! \name Constructores & Destructors
	/**
	 * \brief Standard domain constructor
	 */
	Domain();

	/**
	 * \brief Constructor with minimal Domain information.
	 * @param name Name of the domain.
	 * @param accession The accession number.
	 * @param seq_start First position of the domain.
	 * @param seq_end Last position of the domain.
	 */
	Domain(const std::string &name, const std::string &accession, size_t seq_start, size_t seq_end);

	/**
	 * \brief Complete Domain constructor.
	 * @param name Name of the domain.
	 * @param accession The accession number.
	 * @param seq_start First position of the domain.
	 * @param seq_end Last position of the domain.
	 * @param env_start First position of the envelope.
	 * @param env_end Last position of the envelope.
	 * @param hmm_start First position of the HMM match.
	 * @param hmm_end Last position of hte HMM match.
	 * @param evalue The evalue of the the domain.
	 * @param id The id of the domain.
	 */
	Domain(const std::string &name, const std::string &accession, size_t seq_start, size_t seq_end, size_t env_start, size_t env_end, size_t hmm_start, size_t hmm_end, double evalue, int id=-1);

	/**
	 * \brief Destructor
	 */
	virtual ~Domain();
	/**@}*/


	/**@{*/
	//! \name Basic functions
	/**
	 * \brief Returns the name of the domain.
	 * @return The name.
	 */
	std::string
	name() const
	{
		return _name;
	}

	/**
	 * \brief Returns the accesion number of the domain.
	 * @return The accession number.
	 */
	std::string
	accession() const
	{
		return _accession;
	}

	/**
	 * \brief Returns the accesion number of the domain.
	 * @return The accession number.
	 */
	void
	accession(const std::string &accession)
	{
		_accession=accession;
	}

	/**
	 * \brief Returns the first position of the domain.
	 * @return The first position of the domain.
	 */
	size_t
	seq_start() const
	{
		return _seq_start;
	}

	/**
	 * \brief Returns the the last position of the domain.
	 * @return The last position of the domain.
	 */
	size_t
	seq_end() const
	{
		return _seq_end;
	}

	/**
	 * \brief Returns the first position of the envelope.
	 * @return The first position of the envelope.
	 */
	size_t
	env_start() const
	{
		return _env_start;
	}

	/**
	 * \brief Returns the last position of the envelope.
	 * @return The last position of the envelope.
	 */
	size_t env_end() const
	{
		return _env_end;
	}

	/**
	 * \brief Sets the first position of the domain.
	 * @param val The new value.
	 */
	void
	seq_start(size_t val)
	{
		_seq_start=val;
	}

	/**
	 * \brief Sets the last position of the domain.
	 * @param val The new value.
	 */
	void
	seq_end(size_t val)
	{
		_seq_end=val;
	}

	/**
	 * \brief Sets the first position of the envelope.
	 * @param val The new value.
	 */
	void
	env_start(size_t val)
	{
		_env_start=val;
	}

	/**
	 * \brief Sets the last position of the envelope.
	 * @param val The new value.
	 */
	void
	env_end(size_t val)
	{
		_env_end=val;
	}

	/**
	 * \brief Returns the first position of the HMM match.
	 * @return The first position.
	 */
	size_t hmm_start() const
	{
		return _hmm_start;
	}

	/**
	 * \brief Returns the last position of the HMM match.
	 * @return The the last position.
	 */
	size_t hmm_end() const
	{
		return _hmm_end;
	}

	/**
	 * \brief Sets the first position of the HMM match.
	 * @param val The new value.
	 */
	void
	hmm_start(size_t val)
	{
		_hmm_start = val;
	}

	/**
	 * \brief Sets the last position of the HMM match.
	 * @param val The new value.
	 */
	void
	hmm_end(size_t val)
	{
		_hmm_end=val;
	}


	/**
	 * \brief Sets how often the domain is repeated.
	 * @param val The number of repeates.
	 */
	void
	repeated(int val)
	{
		_repeated=val;
	}

	/**
	 * \brief Returns how often the domain is repeated.
	 * @return The number of repeats.
	 */
	int
	repeated() const
	{
		return _repeated;
	}

	/**
	 * \brief Returns the evalue of the domain.
	 * @return The evalue.
	 */
	double evalue() const
	{
		return _evalue;
	}


	/**
	 * \brief Returns the id of the domain.
	 * @return The id of the domain.
	 */
	int id() const
	{
		return _id;
	}

	/**
	 * \brief Sets the id of the domain.
	 * @param val The id of the domain.
	 */
	void id(int val)
	{
		_id=val;
	}

	/**
	 * \brief Returns the length of the domain.
	 * @return Domain length.
	 */
	size_t seq_length() const
	{
		return _seq_end - _seq_start + 1;
	}


	void seq_length(size_t length)
	{
		_seq_end  = _seq_start + length -1;
	}

	/**
	 * \brief Returns the length of the evelope.
	 * @return The envelope length.
	 */
	size_t env_length() const
	{
		return _env_end - _env_start + 1;
	}

	/**
	 * \brief Returns the length of the HMM match.
	 * @return Length of the HMM match.
	 */
	size_t hmm_length() const
	{
		return _hmm_end - _hmm_start +1;
	}

	/**
	 * \brief The original position of the domain in a domain architecture.
	 * @return The original position.
	 */
	int ori_pos() const
	{
		return _original_position;
	}

	/**
	 * \brief Sets the original position of the domain in a domain architecture.
	 * @param pos The position.
	 */
	void ori_pos(int pos)
	{
		_original_position = pos;
	}
	/**@}*/

};






}


#endif /* DOMAIN_HPP_ */
