/*
 * Domain.cpp
 *
 *  Created on: Jun 2, 2013
 *      Author: Carsten Kemena
 */

#include "DomainArchitecture.hpp"


using namespace std;

namespace MDAT
{



/***********************************************************************
 * 						Constructors & Destructors					   *
 ***********************************************************************/


DomainArchitecture::DomainArchitecture():_id(-1), _n_members(0), _domains()
{}

DomainArchitecture::~DomainArchitecture()
{}

bool domain_start_sort(const Domain &dom1, const Domain &dom2)
{
	return (dom1.seq_start() < dom2.seq_start());
}


/**
 * \brief Sorts the Domain architectures by domain start
 */
void
DomainArchitecture::sort()
{
	std::sort(_domains.begin(), _domains.end(), domain_start_sort);
}


/**
 * \brief Function to solve conflicting domains
 *
 * This function can solve several problems which appear after domain annotation, this includes nested domains, overlapping domains and split domains
 * @param options Determines which problem(s) have to be solved.
 */
void
DomainArchitecture::clean_up(unsigned char options)
{

	size_t n_domains = _domains.size();
	size_t i;
	n_domains = _domains.size();
	if (options & CleanNested)
	{
		int last;
		vector<size_t> to_delete;
		last=0;
		for (i=1; i<n_domains; ++i)
		{
			if ((_domains[i].seq_start() >= _domains[last].seq_start()) && (_domains[i].seq_end() <= _domains[last].seq_end()))
			{
				to_delete.push_back(i);
				--last;
			}
			else if ((_domains[last].seq_start() == _domains[i].seq_start()) && (_domains[last].seq_end() <= _domains[i].seq_end()))
				to_delete.push_back(i-1);
			++last;
		}
		std::sort(to_delete.begin(), to_delete.end(), std::greater<size_t>());
		for (i=0; i<to_delete.size(); ++i)
			_domains.erase(_domains.begin()+to_delete[i]);
		n_domains = _domains.size();
	}
	if (options & CleanOverlap)
	{
		for (i=1; i<n_domains; ++i)
		{
			if (_domains[i-1].env_end() >= _domains[i].env_start())
			{
				if (_domains[i-1].evalue() < _domains[i].evalue())
				{
					_domains[i].env_start(_domains[i-1].env_end()+1);
					if (_domains[i-1].seq_end() >= _domains[i].seq_start())
						_domains[i].seq_start(_domains[i-1].seq_end()+1);
				}
				else
				{
					_domains[i-1].env_end(_domains[i].env_start()-1);
					if (_domains[i-1].seq_end() >= _domains[i].seq_start())
						_domains[i-1].seq_end(_domains[i].seq_start()-1);
				}
			}
		}
	}
	if (options & CleanMerged)
	{
		vector<size_t> to_delete;
		for (i=1; i<n_domains; ++i)
		{
			if ((_domains[i].id() == _domains[i-1].id()) && (_domains[i-1].hmm_end() < _domains[i].hmm_start()))
			{
				_domains[i].seq_start(_domains[i-1].seq_start());
				_domains[i].env_start(_domains[i-1].env_start());
				_domains[i].hmm_start(_domains[i-1].hmm_start());
				to_delete.push_back(i-1);
			}
		}
		std::sort(to_delete.begin(), to_delete.end(), std::greater<size_t>());
		for (i=0; i<to_delete.size(); ++i)
			_domains.erase(_domains.begin()+to_delete[i]);
	}
	n_domains = _domains.size();
}


std::ostream& operator<< (std::ostream &out, const DomainArchitecture &archi)
{
	if (archi[0].id() == -1)
		out << "-------   ";
	else
		out << archi[0].accession() << "*" << std::setw(2) << archi[0].repeated();
	for (size_t i=1; i<archi.length(); ++i)
	{
		if (archi[i].id() == -1)
			out << " " << "-------   ";
		else
			out << " " << archi[i].accession() << "*" << std::setw(2) << archi[i].repeated();
	}
	out << endl;
	return out;
}


void
DomainArchitecture::insert_gaps(std::string edit_string)
{
	Domain gap_domain("", "", 0, 0);
	size_t pos=_domains.size();
	size_t new_length=edit_string.size();
	_domains.resize(new_length);
	std::vector<Domain>::iterator it = _domains.end();
	for (size_t i=0; i<new_length; ++i)
	{
		if (edit_string[i]=='-')
			(*(--it)) = gap_domain;
		else
			(*(--it)) = _domains[--pos];
	}
}

void
DomainArchitecture::collapse_repeats()
{
	size_t n_domains = _domains.size();
	if (n_domains < 2)
		return;
	size_t pos =0;
	for (size_t i=1; i<n_domains; ++i)
	{
		if (_domains[pos].id()==_domains[i].id())
		{
			_domains[pos].seq_end(_domains[i].seq_end());
			_domains[pos].env_end(_domains[i].env_end());
			_domains[pos].hmm_end(_domains[i].hmm_end());
			_domains[pos].repeated(_domains[pos].repeated()+1);
		}
		else
		{
			++pos;
			if (pos!=i)
				_domains[pos]=_domains[i];
		}
	}
	_domains.resize(pos+1);
}

DomainArchitectureSet::DomainArchitectureSet(): _archis()
{

}


DomainArchitectureSet::~DomainArchitectureSet()
{

}


std::ostream& operator<< (std::ostream &out, const DomainArchitectureSet &archi)
{
	for (size_t i=0; i<archi._archis.size(); ++i)
		out << std::right << std::setw(2)<< archi._archis[i].id() << " " << archi._archis[i];
	return out;
}


bool gap_pattern_sort (const DomainArchitecture &arch1, const DomainArchitecture &arch2)
{
	size_t size_1 = arch1.size();
	size_t size_2 = arch2.size();
	int first_gap1=-1;
	int first_gap2=-1;
	int n_doms1=0,n_doms2=0;
	size_t i;
	for (i=0; i<size_1; ++i)
	{
		if (arch1[i].id() != -1)
			++n_doms1;
		else if (first_gap1==-1)
			first_gap1=i;
	}
	for (i=0; i<size_2; ++i)
	{
		if (arch2[i].id() != -1)
			++n_doms2;
		else if (first_gap2==-1)
			first_gap2=i;
	}
	if (n_doms1 != n_doms2)
		return n_doms1 > n_doms2;
	else
		return first_gap1 > first_gap2;
}


bool member_sort (const DomainArchitecture &arch1, const DomainArchitecture &arch2)
{
	return arch1.n_members() > arch2.n_members();
}

bool id_sort (const DomainArchitecture &arch1, const DomainArchitecture &arch2)
{
	return arch1.id() < arch2.id();
}


void
DomainArchitectureSet::sort(const std::string sort_type)
{
	if (sort_type=="gap_pattern")
	{
		std::sort(_archis.begin(), _archis.end(), gap_pattern_sort);
	}
	else if (sort_type=="members")
	{
		std::sort(_archis.begin(), _archis.end(), member_sort);
	}
	else if (sort_type=="id")
	{
		std::sort(_archis.begin(), _archis.end(), id_sort);
	}
}


/*
void
removeGaps(DomainArchitectureSet &set)
{

	size_t nArchis=set.size();
	size_t length = set[0].length();
	size_t j;
	std::vector<int> nonGapCounter(length, 0);
	for (size_t i=0; i<nArchis; ++i)
	{
		DomainArchitecture &arch = set[i];
		for (j=0; j<length; ++j)
		{
			if (arch[j].id() == -1)
				++nonGapCounter[j];
		}
	}

	int pos;
	for (size_t i=0; i<nArchis; ++i)
	{
		DomainArchitecture &arch = set[i];
		pos = -1;
		for (j=0; j<length; ++j)
		{
			if (nonGapCounter[j]!=0)
				arch[++pos]=arch[j];
		}
		arch.resize(++pos);
	}
}
*/



} /* namespace MDAT */


