/*
 * Domain.h
 *
 *  Created on: Jun 2, 2013
 *      Author: Carsten Kemena
 */

#ifndef DomainArchitecture_HPP_
#define DomainArchitecture_HPP_


// C++ header
#include <algorithm>
#include <memory>
#include <string>
#include <vector>
#include <iostream>
#include <iomanip>

// MDAT header
#include "Domain.hpp"

namespace MDAT
{


/**
 * \brief Domain cleanup options.
 */
enum CleanupOptions {
  CleanNested    = 0x01,//!< CleanNested Cleans nested domains
  CleanOverlap   = 0x02,//!< CleanOverlap Cleans overlapping domains
  CleanMerged    = 0x04//!< CleanMerged Cleans splitted domains
  //CleanRepeat    = 0x08,//!< CleanRepeat
};


/**
 * \brief Represents a set of domains
 */
class DomainArchitecture
{
private:
	int _id;
	size_t _n_members;
	std::vector<Domain> _domains;

public:

	/***********************************************************************
	 * 						Constructors & Destructors					   *
	 ***********************************************************************/

	/**@{*/
	//! \name Constructores & Destructors
	/**
	 * \brief Creates an empty DomainArchitecture
	 */
	DomainArchitecture();

	/**
	 * \brief Destructor
	 */
	virtual ~DomainArchitecture();

	/**@}*/

	/**
	 * \brief Appends a Domain to the architecture
	 * @param domain The domain
	 */
	void
	add_domain(const Domain &domain)
	{
		_domains.push_back(domain);
	}

	/**
	 * Length of the architecture.
	 * @return The number of domains in the architecture.
	 */
	size_t
	size() const
	{
		return _domains.size();
	}

	void
	resize(size_t new_size)
	{
		_domains.resize(new_size);
	}

	/**
	 * Length of the architecture.
	 * @return The number of domains in the architecture.
	 */
	size_t
	length() const
	{
		return _domains.size();
	}


	Domain &operator[](unsigned int index)
	{
		return _domains[index];
	}

	/**
	 * \overload
	 */
	const Domain &operator[](unsigned int index) const
	{
		return _domains[index];
	}


	/**
	 * \brief Checks if to architectures are the same.
	 * @param dom1 The first architecture.
	 * @param dom2 The second architecture.
	 * @return True if the two architectures are the same else false.
	 */
    friend bool operator== (const DomainArchitecture &dom1, const DomainArchitecture &dom2)
	{
    	if (dom1.length() != dom2.length())
    		return false;
    	size_t length = std::min(dom1.length(), dom2.length());
    	for (size_t i=0; i<length; ++i)
    	{
    		if ((dom1[i].id() != dom2[i].id()) || (dom1[i].repeated() != dom2[i].repeated()))
    			return false;
    	}
    	return true;
	}


    /**
     * \brief Checks if to architectures are different.
     * @param dom1 The first architecture.
     * @param dom2 The second architecture.
     * @return True if the two architectures are different else false.
     */
    friend bool operator!= (DomainArchitecture &dom1, DomainArchitecture &dom2)
	{
    	if (dom1.length() != dom2.length())
    		return true;
    	size_t length = std::min(dom1.length(), dom2.length());
    	for (size_t i=0; i<length; ++i)
    	{
    		if ((dom1[i].id() != dom2[i].id()) || (dom1[i].repeated() != dom2[i].repeated()))
    			return true;
    	}
    	return false;
	}


    friend std::ostream& operator<< (std::ostream &out, const DomainArchitecture &archi);

    /**
     * \brief Returns the id architecture.
     * @return The id of the architecture.
     */
	int id() const
	{
		return _id;
	}

	/**
	 * \brief Sets the id of the architecture.
	 * @param id The id of the architecture.
	 */
	void id(int id)
	{
		_id=id;
	}

	/**
	 * \brief Returns the number of sequences with this architecture.
	 * @return The number of sequences with this architecture.
	 */
	size_t n_members() const
	{
		return _n_members;
	}

	/**
	 * \brief Sets the number of sequences having this architecture.
	 * @param value The number of sequences having this architecture.
	 */
	void n_members(size_t value)
	{
		_n_members=value;
	}

	/**
	 * \brief solves several problems of a domain architecture.
	 * \details Several problems can appear when
	 *
	 * @param options The options to be used (\sa CleanupOptions);
	 *
	 */
	void clean_up(unsigned char options);


	/**
	 * \brief Sorts the domains in the architecture according to their start.
	 */
	void sort();

	/**
	 * \brief Inserts a GapDomain at the given positions.
	 * \details The edit_string has to be of the form "m-mm-" where '-' denote the gaps to introduce.
	 * @param edit_string String noting gap positions.
	 */
	void insert_gaps(std::string edit_string);

	/**
	 * \brief Merges repeats to together.
	 */
	void
	collapse_repeats();

	void
	removeGaps()
	{
		size_t len=_domains.size();
		int pos=-1;
		for (size_t i=0; i<len; ++i)
		{
			if (_domains[i].id()!=-1)
				_domains[++pos]=_domains[i];
		}
	}
};

/*
void
print_archis(std::vector<DomainArchitecture> vec);
*/

/**
 * \brief Class to store several Architectures.
 */
class DomainArchitectureSet
{
private:
	std::vector<DomainArchitecture > _archis;

public:

	/**@{*/
	//! \name Constructors & Destructors
	/**
	 * \brief Constructor
	 */
	DomainArchitectureSet();

	/**
	 * \brief Destructor
	 */
	~DomainArchitectureSet();
	/**@}*/

	/**@{*/
	//! \name Operators

	/**
	 * \brief Accesses a DomainArchitecture
	 * @param index The index of the architecture.
	 * @return Reference to the domain architecture.
	 */
	DomainArchitecture &operator[](unsigned int index)
	{
		return _archis[index];
	}

	/**
	 * \overload
	 */
	const DomainArchitecture &operator[](unsigned int index) const
	{
		return _archis[index];
	}
	/**@}*/


	/**@{*/
	//! \name Basic functions

	/**
	 * \brief Adds a domain architecture to the set
	 * @param archi The domain architecture.
	 */
	void
	push_back(const DomainArchitecture &archi)
	{
		_archis.push_back(archi);
	}

	/**
	 * \brief Returns the number of architectures.
	 * @return The number of architectures.
	 */
	size_t size() const
	{
		return _archis.size();
	}

	size_t
	max_length() const
	{
		size_t n_archis=_archis.size();
		size_t min=0;
		for (size_t i=0; i<n_archis; ++i)
		{
			if (_archis[i].length()>min)
				min=_archis[i].length();
		}

		return min;
	}


	/**
	 * Sorts the DomainArchitecture set.
	 * @param stort_type The way to sort the sequences.
	 */
	void
	sort(const std::string stort_type="gap_pattern");
	/**@}*/

	/**
	 * \brief Prints the domain architecture set.
	 * @param out The output stream.
	 * @param archi The architecture set.
	 * @return The output stream.
	 */
	friend std::ostream& operator<< (std::ostream &out, const DomainArchitectureSet &archi);

	/**
	 * \brief Collapses all repeated domains in each architecture into a single domains.
	 */
	void
	collapse_repeats();

};

void
removeGaps(DomainArchitectureSet &set);


} /* namespace Sequence */
#endif /* DOMAIN_H_ */


