/*
 * SequenceFeatures.h
 *
 *  Created on: Jun 8, 2013
 *      Author: ck
 */

#ifndef GENOMEFEATURES_H_
#define GENOMEFEATURES_H_

// C++ header
#include <algorithm>
#include <fstream>
#include <limits>
#include <map>
#include <string>
#include <vector>


// Boost header
#include <boost/tokenizer.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/iterator.hpp>


#include "../Basics/basics.hpp"
#include "../utils/MDAT_Exceptions.hpp"

namespace MDAT
{

struct List_;
typedef Tag<List_> List;


struct
SingleSequenceFeature
{
	std::string seq_id;
	std::string source;
	std::string type;
	size_t start;
	size_t end;
	double score;
	char strand;
	short phase;

	std::map<std::string, std::string> attributes;

	size_t length() const
	{
		return end-start+1;
	}


};


typedef  std::vector<SingleSequenceFeature>::iterator SequenceFeatures_iter;


template<typename FeatureType>
class SequenceFeatures
{
private:
	typedef std::map<std::string, std::vector<SingleSequenceFeature> >Feature_container;
	Feature_container _feature_list;

	struct sort_features
	{
	    inline bool operator() (const SingleSequenceFeature& feature1, const SingleSequenceFeature& feature2)
	    {
	    	if ((feature1.type!="gene") && (feature2.type!="gene") && (feature1.type!="contig") && (feature2.type!="contig"))
	    	{
	    		if (feature1.attributes.at("Parent") != feature2.attributes.at("Parent"))
	    			return feature1.attributes.at("Parent") < feature2.attributes.at("Parent");
	    	}
	    	if (feature1.start < feature2.start)
	    		return true;
	    	else if (feature1.start == feature2.start)
	    		return (feature1.length()>feature2.length());
	    	else
	    		return false;
	    }
	};


public:

	SequenceFeatures();

	virtual ~SequenceFeatures();

	/**
	 * \brief Reads a GFF file.
	 * @param gff_f The file to read in gff format.
	 */
	void
	read_gff(const std::string &gff_f);

	/**
	 * \brief Write data to disk.
	 * @param out_f The output file to write to.
	 */
	void
	write_gff(const std::string &out_f);

	SequenceFeatures_iter
	begin()
	{
		return _feature_list[_feature_list.begin()->first].begin();
	}

	SequenceFeatures_iter
	end()
	{
		return _feature_list[_feature_list.begin()->first].end();
	}


	/**
	 * \brief Iterator to the first feature of a given sequence.
	 * @param seq_id The sequence name.
	 * @return Iterator to the first feature of a sequence.
	 */
	SequenceFeatures_iter
	begin(const std::string &seq_name)
	{
		return _feature_list[seq_name].begin();
	}

	/**
	 * \brief Iterator to the element behind the last element of a given sequence.
	 * @param seq_id The sequence name.
	 * @return Iterator to just behind the last feature of a sequence.
	 */
	SequenceFeatures_iter
	end(const std::string &seq_name)
	{
		return _feature_list[seq_name].end();
	}


/*	void
	sort()
	{
		std::map<std::string, std::vector<SingleSequenceFeature> >::iterator it,it_end=_feature_list.end();
		for (it=_feature_list.begin(); it!=it_end; ++it)
			std::sort(it->second.begin(), it->second.end(), sort_features());
	}*/

	void
	append(const std::string &seq_id, const SingleSequenceFeature &feature)
	{
		_feature_list[seq_id].push_back(feature);
	}

	/**
	 * \brief Add introns to each gene.
	 */
	void
	add_introns();

	/**
	 * Checks if any features have been found.
	 * @return True if nothing found, else false.
	 */
	bool
	empty() const
	{
		return _feature_list.empty();
	}
/*
	void
	add_promoter(size_t length);
*/
};

template<typename FeatureType>
SequenceFeatures<FeatureType>::SequenceFeatures()
{
	// TODO Auto-generated constructor stub

}

template<typename FeatureType>
SequenceFeatures<FeatureType>::~SequenceFeatures()
{
	// TODO Auto-generated destructor stub
}



template<typename FeatureType>
void
SequenceFeatures<FeatureType>::read_gff(const std::string &gff_f)
{
	Feature_container::iterator cont_it, cont_it_end;
	std::ifstream gff_F(gff_f);
	if(!gff_F) throw MDAT_IO_Exception("Could not open file\n");
	std::string line;
	size_t pos=0;
	typedef boost::tokenizer< boost::char_separator<char> > t_tokenizer;
	boost::char_separator<char> sep("\t");
	t_tokenizer tok(std::string(" "), sep);
	t_tokenizer::iterator it;
	std::string attribute, tmp_string;
	std::vector<std::string> attribute_list;
	std::vector<SingleSequenceFeature>::iterator feature_it;
	int n_elements;
	int i;
	std::string seq_id;
	//for(boost::tokenizer<>::iterator beg= tok.begin(); beg!=tok.end(); ++beg)
	//scf7180000350264	pbar_OGSv1.2	gene	77142	87005	.	+	.	ID=PB18755;Name=PB18755;Alias=PB18755;
	std::string last_id ="";
	while (!gff_F.eof())
	{
		getline(gff_F, line);
		if (line.empty() || (line[0] == '#'))
			continue;

		tok.assign(line);
		it = tok.begin();
		seq_id = *it;
		if (seq_id!=last_id)
		{
			cont_it = _feature_list.find(seq_id);
			if (cont_it==_feature_list.end())
			{
				cont_it=_feature_list.insert(std::pair<std::string, std::vector<SingleSequenceFeature> >(seq_id, std::vector<SingleSequenceFeature>())).first;
			}
			last_id=seq_id;
		}

		cont_it->second.push_back(SingleSequenceFeature());
		feature_it=cont_it->second.end()-1;
		feature_it->seq_id=seq_id;
		feature_it->source = *(++it);
		feature_it->type = *(++it);
		feature_it->start = boost::lexical_cast<size_t>(*(++it))-1; // gff starts at 1 in file
		feature_it->end = boost::lexical_cast<size_t>(*(++it))-1;
		if (*(++it) == ".")
			feature_it->score = std::numeric_limits<double>::min();
		else
			feature_it->score = boost::lexical_cast<double>(*it);
		feature_it->strand=(*(++it))[0];
		if (*(++it) == ".")
			feature_it->phase = std::numeric_limits<short>::min();
		else
			feature_it->phase = boost::lexical_cast<short>(*it);

		tmp_string =  *(++it);
		boost::split(attribute_list, tmp_string, boost::is_any_of(";="), boost::algorithm::token_compress_on);
		n_elements = attribute_list.size();
		for (i=0; i<n_elements; i+=2)
		{
			if (!attribute_list[i].empty())
				feature_it->attributes[attribute_list[i]] = attribute_list[i+1];
			else
				break;
		}
		++pos;
	}
}


template<typename FeatureType>
void
SequenceFeatures<FeatureType>::write_gff(const std::string &out_f)
{
	std::ofstream out_F(out_f);
	out_F << "##gff-version" << std::endl;
	std::map<std::string, std::vector<SingleSequenceFeature> >::iterator it,it_end=_feature_list.end();
	std::vector<SingleSequenceFeature>::iterator source_it, source_it_end;
	std::map<std::string, std::string>::iterator attr_it,attr_it_end;
	for (it=_feature_list.begin(); it!=it_end; ++it)
	{
		source_it_end=it->second.end();
		for (source_it=it->second.begin(); source_it!=source_it_end; ++source_it)
		{
			out_F << source_it->seq_id << "\t" << source_it->source << "\t" << source_it->type << "\t"<< source_it->start+1 << "\t"<< source_it->end+1 << "\t";
			if (source_it->score==std::numeric_limits<double>::min())
				out_F << ".";
			else
				out_F << source_it->score;
			out_F << "\t"<< source_it->strand<< "\t";
			if (source_it->phase==std::numeric_limits<short>::min())
				out_F<< ".\t";
			else
				out_F<< source_it->phase << "\t";
			attr_it_end=source_it->attributes.end();
			for (attr_it=source_it->attributes.begin(); attr_it!=attr_it_end; ++attr_it)
				out_F << attr_it->first << "=" <<attr_it->second<<";";

			out_F << std::endl;

		}
	}
	out_F.close();
}

template<typename FeatureType>
void
SequenceFeatures<FeatureType>::add_introns()
{
	std::map<std::string, std::vector<SingleSequenceFeature> >::iterator it,it_end=_feature_list.end();
	size_t last;
	size_t i;
	bool new_mRNA=false;
	SingleSequenceFeature new_feat;

	for (it=_feature_list.begin(); it!=it_end; ++it)
	{
		std::vector<SingleSequenceFeature> &vec = it->second;
		last=0;
		for (i=0; i<vec.size(); ++i)
		{
			if (vec[i].type=="mRNA")
				new_mRNA = true;
			else if (vec[i].type=="exon")
			{
				if (new_mRNA)
					new_mRNA=false;
				else
				{
					new_feat=vec[i];

					if (vec[i].start>vec[last].start)
					{
						new_feat.start=vec[last].end+1;
						new_feat.end = vec[i].start-1;
						new_feat.type="intron";
					}
					else
					{
						new_feat.start=vec[i].end+1;
						new_feat.end=vec[last].start-1;
						new_feat.type="intron";
					}
					new_feat.attributes["ID"]=new_feat.attributes["Parent"]+":intron";
					vec.insert(vec.begin()+i,new_feat);
					++i;
				}
				last=i;
			}

		}
	}
	//sort();
}



/*
template<typename FeatureType>
void
SequenceFeatures<FeatureType>::add_promoter(size_t length)
{
	std::map<std::string, std::vector<SingleSequenceFeature> >::iterator it,it_end=_feature_list.end();
	bool first_CDS=false;
	SingleSequenceFeature promoter_feat;
	size_t i;

	for (it=_feature_list.begin(); it!=it_end; ++it)
	{
		std::vector<SingleSequenceFeature> &vec = it->second;
		for (i=0; i<it->second.size(); ++i)
		{
			if (vec[i].type=="mRNA")
				first_CDS=true;
			if ((first_CDS) && (vec[i].type=="CDS"))
			{

				promoter_feat=vec[i];
				if (promoter_feat.start>length)
				{
					promoter_feat.length=length;
					promoter_feat.start-=length;
				}
				else
				{
					promoter_feat.length=promoter_feat.start;
					promoter_feat.start=0;
				}
				promoter_feat.attributes["ID"]=promoter_feat.attributes["Parent"]+":promoter";
				promoter_feat.type="promoter";
				it->second.insert(it->second.begin()+i,promoter_feat);
				first_CDS=false;
			}
		}
	}
	//sort();
}
*/



} /* namespace Sequence */
#endif /* GENOMEFEATURES_H_ */
