/*
 * ProteinSequence.h
 *
 *  Created on: Jun 10, 2013
 *      Author: Carsten Kemena
 *
 * This file is part of MDAT.
 *
 * MDAT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MDAT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with MDAT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * \file ProteinSequence.hpp
 * \brief File containing the ProteinSequence class
 */
#ifndef PROTEINSEQUENCE_H_
#define PROTEINSEQUENCE_H_


// boost header
#include <boost/lexical_cast.hpp>

#include "Sequence.hpp"
#include "../Annotation/DomainArchitecture.hpp"


namespace MDAT
{

/** \addtogroup Sequence
 *  @{
 */

/**
 * \brief A class to represent a protein sequence
 * \details Beside the normal information it allows to store domain information.
 */
class ProteinSequence: public MDAT::Sequence
{
private:
	DomainArchitecture _dom_architecture;

public:
	/**@{*/
	//! \name Constructors & Destructors

	/**
	 *  Full initialization.
	 */
	ProteinSequence(const std::string &seq_name, const std::string &seq, const std::string &comment_="", size_t seq_id=0);
	/**
	 * Reserves the necessary space
	 */
	ProteinSequence(const std::string &seq_name, const std::string &comment_, unsigned int seq_length, size_t seq_id=0);
	/**
	 * Copy constructor
	 */
	ProteinSequence(const Sequence &seq);


	virtual ~ProteinSequence();
	/**@}*/



	/**@{*/
	//! \name Domain related methods

	/**
	 * \brief Append a domain to the existing set.
	 * @param dom The domain to append.
	 */
	void add_domain(const Domain &dom)
	{
		_dom_architecture.add_domain(dom);
	}


	/**
	 * \brief Returns the DomainArchitecture of the sequence.
	 * @return The architecture.
	 */
	DomainArchitecture &
	dom_archi()
	{
		return _dom_architecture;
	}

	/**
	 * \overload
	 */
	const DomainArchitecture &
	dom_archi() const
	{
		return _dom_architecture;
	}


	/**
	 * \brief Returns a domain at a specific position.
	 * @param index The index.
	 * @return The domain.
	 */
	Domain &
	domain(size_t index)
	{
		return _dom_architecture[index];
	}

	/**
	 * \brief The number of domains of the sequence.
	 * @return The number of domains.
	 */
	size_t
	n_domains() const
	{
		return _dom_architecture.length();
	}
	/**@}*/
};


/**
 *  @}
 */



} /* namespace Sequence */
#endif /* PROTEINSEQUENCE_H_ */
