/*
 * ProteinSequenceSet.hpp
 *
 *  Created on: 19 Jul 2013
 *      Author: Carsten Kemena
 *
 * This file is part of MDAT.
 *
 * MDAT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MDAT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with MDAT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef PROTEINSEQUENCESET_HPP_
#define PROTEINSEQUENCESET_HPP_


// C++ header
#include <fstream>

// Boost header
#include <boost/algorithm/string.hpp>

// BioSeqoolBox header
#include "SequenceSet.hpp"
#include "Seq_functs.hpp"
#include "SplitSet.hpp"
#include "ProteinSequence.hpp"


namespace MDAT
{


/** \addtogroup SequenceSet
 *  @{
 */


/**
 * \brief A SequenceSet object for sequences of type ProteinSequence.
 * \details It provides additional functions to read domain files of various formats and connects them to the sequences.
 */
template<typename MemoryType> class ProteinSequenceSet : public SequenceSetBase<ProteinSequence, MemoryType>
{

private:
	typedef std::shared_ptr<ProteinSequence> Seq_ptr;
	DomainArchitectureSet _dom_archis;
	std::vector<std::string> _domain_accessions;

	std::string _identify_domain_file_format(std::ifstream &domain_F);


	void _read_hmmscan_pfam(std::ifstream &domain_F);
	void _read_pfamscan(std::ifstream &domain_F);
	void _read_xdom(std::ifstream &domain_F);
	void _read_sff(std::ifstream &domain_F);

public:

	ProteinSequenceSet() : _dom_archis(), _domain_accessions()
	{}
	ProteinSequenceSet(size_t id_val): SequenceSetBase<ProteinSequence, MemoryType>(id_val), _dom_archis(), _domain_accessions()
	{}

	ProteinSequenceSet(const ProteinSequenceSet&) = delete;
	ProteinSequenceSet(ProteinSequenceSet&&) = default;

	/**@{*/
	//! \name Domain functions
	/**
	 * \brief Add domain information to the sequence set.
	 * The input file can be either the hmmer domain tabular output or pfamScan.pl output.
	 * @param domain_f A file containing the domains.
	 */
	void add_pfam_domains(const std::string &domain_f);

	/**
	 * \brief Solves overlaps and nested domains.
	 */
	void clean_up_domains(unsigned char options);


	/**
	 * \brief Refines the domain boundaries by using the envelope and hmm information.
	 */
	void refine_boundaries();

	/**
	 * \brief Calculates the architectures from the domains in the set.
	 */
	void extract_architectures();

	/**
	 * \brief Returns the number of architectures of the set.
	 * @return Number of architectures.
	 */
	size_t n_architectures() const
	{
		return _dom_archis.size();
	}

	/**
	 * \brief Returns the number of different domains.
	 * @return Number of different domains.
	 */
	size_t n_domains() const
	{
		return _domain_accessions.size();
	}

	/**
	 * \brief Returns the domain accession name
	 * \param i The id of the domain.
         */
	const std::string&
	domain_name(size_t i) const
	{
		return _domain_accessions[i];
	}


	/**
	 * Returns the DomainArchitectureSet belonging to the sequences.
	 * @return
	 */
	DomainArchitectureSet &dom_archis()
	{
		return _dom_archis;
	}

	/**
	 * \overload
	 */
	const DomainArchitectureSet &dom_archis() const
	{
		return _dom_archis;
	}


	void
	dom_archis(DomainArchitectureSet dom_arch)
	{
		_dom_archis=dom_arch;
	}

	/**
	 * Writes the domain Architectures to a file.
	 * @param out_f The file to write the architectures to.
	 */
	void
	write_domArchitecture(const std::string &out_f) const;

	/**@}*/
};

/**
 *  @}
 */



template<typename MemoryType>
void
ProteinSequenceSet<MemoryType>::extract_architectures()
{
	size_t j,k;
	size_t n_seqs = this->n_seqs();
	std::vector<std::vector<size_t> > lengths;
	size_t archlen;
	for (size_t i = 0; i<n_seqs; ++i)
	{
		for (j=0; j<_dom_archis.size(); ++j)
		{
			if ((*this)[i].dom_archi() == _dom_archis[j])
			{
				const DomainArchitecture &arch =(*this)[i].dom_archi();
				archlen = arch.size();
				for (k=0; k<archlen; ++k)
					lengths[j][k] += arch[k].seq_length();
				break;
			}
		}
		(*this)[i].dom_archi().id(j);
		if (j==_dom_archis.size())
		{
			_dom_archis.push_back((*this)[i].dom_archi());
			lengths.push_back(std::vector<size_t>());
			for (k=0; k<_dom_archis[j].size(); ++k)
			{
				_dom_archis[j][k].ori_pos(k);
				lengths[j].push_back(_dom_archis[j][k].seq_length());
			}
		}
		_dom_archis[j].n_members(_dom_archis[j].n_members()+1);
	}
	for (j=0; j<_dom_archis.size(); ++j)
	{
		DomainArchitecture &arch = _dom_archis[j];
		archlen=arch.size();
		for (k=0; k<archlen; ++k)
		{
			arch[k].seq_start(0);
			arch[k].seq_end((lengths[j][k]/arch.n_members())-1);
		}
	}
}


template<typename MemoryType>
void
ProteinSequenceSet<MemoryType>::add_pfam_domains(const std::string &domain_f)
{
	 std::ifstream domain_F;
	 domain_F.exceptions( std::ifstream::failbit | std::ifstream::badbit);
	 domain_F.open(domain_f.c_str(), std::ifstream::in);
	 domain_F.exceptions(std::ifstream::badbit);
	 std::string line;
	 getline(domain_F, line);
	 if (line.substr(0,14)== "# pfam_scan.pl")
		 _read_pfamscan(domain_F);
	 else
	 {
		if (line[0] == '>')
		{
			domain_F.seekg(0);
			_read_xdom(domain_F);
		}
		else
		{
			getline(domain_F, line);
			domain_F.seekg(0);
			if (line.substr(0,13)=="# target name")
				_read_hmmscan_pfam(domain_F);
			else
				_read_sff(domain_F);
		}
	 }
	 size_t n_seqs=this->n_seqs();
	 for (size_t i=0; i<n_seqs; ++i)
		 (*this)[i].dom_archi().sort();
	 domain_F.close();
}



/*
#                                                                            --- full sequence --- -------------- this domain -------------   hmm coord   ali coord   env coord
# target name        accession   tlen query name           accession   qlen   E-value  score  bias   #  of  c-Evalue  i-Evalue  score  bias  from    to  from    to  from    to  acc description of target
#------------------- ---------- ----- -------------------- ---------- ----- --------- ------ ----- --- --- --------- --------- ------ ----- ----- ----- ----- ----- ----- ----- ---- ---------------------
HIPIP                PF01355.12    64 A0YD43_9GAMM/35-102  -             68   2.5e-21   75.5   0.6   1   1   1.9e-25   2.8e-21   75.4   0.6     2    64     2    68     1    68 0.94 High potential iron-sulfur protein
*/
template<typename MemoryType>
void
ProteinSequenceSet<MemoryType>::_read_hmmscan_pfam(std::ifstream &domain_F)
{
	std::string line;
	Seq_ptr tmp;
	size_t aln_start, env_start, hmm_start;
	int max_id=-1;
	std::map<std::string, int> accession2id;
	std::map<std::string, int>::iterator it;
	std::string accession;

	while (getline(domain_F, line))
	{
		if ((line[0] == '#') || (line[0] == '\0'))
			continue;
		std::vector<std::string> splitted;
		boost::split(splitted, line, boost::is_any_of(" "), boost::token_compress_on);
		if (!this->contains(splitted[3]))
			continue;
		ProteinSequence &seq = (*this)[splitted[3]];

		aln_start=boost::lexical_cast<size_t>(splitted[17])-1;
		env_start=boost::lexical_cast<size_t>(splitted[19])-1;
		hmm_start=boost::lexical_cast<size_t>(splitted[15])-1;
		accession=splitted[1].substr(0,7);
		it=accession2id.find(accession);
		if (it != accession2id.end())
			seq.add_domain(Domain(splitted[0], accession, aln_start, boost::lexical_cast<size_t>(splitted[18])-1, env_start, boost::lexical_cast<size_t>(splitted[20])-1, hmm_start, boost::lexical_cast<size_t>(splitted[16])-1, boost::lexical_cast<double>(splitted[17]),it->second));
		else
		{
			accession2id[accession] = ++max_id;
			_domain_accessions.push_back(accession);
			seq.add_domain(Domain(splitted[0], accession, aln_start, boost::lexical_cast<size_t>(splitted[18])-1, env_start, boost::lexical_cast<size_t>(splitted[20])-1, hmm_start, boost::lexical_cast<size_t>(splitted[16])-1, boost::lexical_cast<double>(splitted[17]),max_id));
		}
	}
}


//# <seq id> <alignment start> <alignment end> <envelope start> <envelope end> <hmm acc> <hmm name> <type> <hmm start> <hmm end> <hmm length> <bit score> <E-value> <significance> <clan>
//
//1aab_       5     77      5     78 PF09011.5   HMG_box_2         Domain     1    72    73     94.7     3e-27   1 CL0114
template<typename MemoryType>
void
ProteinSequenceSet<MemoryType>::_read_pfamscan(std::ifstream &domain_F)
{
	std::string line;
	Seq_ptr tmp;
	size_t aln_start, env_start, hmm_start;
	int max_id=-1;
	std::map<std::string, int> name2id;
	std::map<std::string, int>::iterator it;
	std::string name, accession;
	while (getline(domain_F, line))
	{

		if ((line[0] == '#') || (line[0] == '\0'))
			continue;
		std::vector<std::string> splitted;
		boost::split(splitted, line, boost::is_any_of(" "), boost::token_compress_on);
		if (!this->contains(splitted[0]))
			continue;

		ProteinSequence &seq = (*this)[splitted[0]];
		aln_start=boost::lexical_cast<size_t>(splitted[1])-1;
		env_start=boost::lexical_cast<size_t>(splitted[3])-1;
		hmm_start=boost::lexical_cast<size_t>(splitted[8])-1;
		accession=splitted[5].substr(0,7);
		name=splitted[6];

		it=name2id.find(accession);
		if (it != name2id.end())
			seq.add_domain(Domain(name, accession, aln_start, boost::lexical_cast<size_t>(splitted[2])-1, env_start, boost::lexical_cast<size_t>(splitted[4])-1, hmm_start, boost::lexical_cast<size_t>(splitted[9])-1, boost::lexical_cast<double>(splitted[12]),it->second ));
		else
		{
			name2id[accession] = ++max_id;
			_domain_accessions.push_back(accession);
			seq.add_domain(Domain(name, accession, aln_start, boost::lexical_cast<size_t>(splitted[2])-1, env_start, boost::lexical_cast<size_t>(splitted[4])-1, hmm_start, boost::lexical_cast<size_t>(splitted[9])-1, boost::lexical_cast<double>(splitted[12]),max_id ));
		}
	}
}

template<typename MemoryType>
void
ProteinSequenceSet<MemoryType>::_read_xdom(std::ifstream &domain_F)
{
	/*
	>A0YZ02	719
	42	357	CHASE2	1.7E-72
	431	539	PAS_4	4.5E-10
	563	719	GGDEF	1.5E-50
	*/
	std::string accession;
	std::string line;
	std::vector<std::string> splitted;
	std::string pfamNamesf = getenv("HOME");
	pfamNamesf.append("/.mda/Pfam-A.clans.tsv");
	std::map<std::string, std::string> name2acc;
	std::ifstream pfamNamesF(pfamNamesf);
	std::map<std::string, int> name2id;
	std::map<std::string, int>::iterator it;
	size_t start, end;
	int max_id=-1;
	while (getline(pfamNamesF, line))
	{
		splitted.clear();
		boost::split(splitted, line, boost::is_any_of("\t"), boost::token_compress_on);
		name2acc[splitted[3]] = splitted[0];
	}
	ProteinSequence *seq = nullptr;
	std::string seqName;
	while (getline(domain_F, line))
	{
		if (line[0] == '>')
		{
			seqName = substr(line, '\t', 1);
			if (!this->contains(seqName))
			{
				seq=nullptr;
				continue;
			}
			else
				seq = &(((*this)[seqName]));
			continue;
		}
		if (seq == nullptr)
			continue;

		splitted.clear();
		boost::split(splitted, line, boost::is_any_of("\t"), boost::token_compress_on);
		accession = name2acc[splitted[2]];
		start=boost::lexical_cast<size_t>(splitted[0])-1;
		end=boost::lexical_cast<size_t>(splitted[1])-1;
		it=name2id.find(accession);

		if (it != name2id.end())
			seq->add_domain(Domain(splitted[2], accession, start, end, start, end, start, end, stod(splitted[3]),it->second ));
		else
		{
			name2id[accession] = ++max_id;
			_domain_accessions.push_back(accession);
			seq->add_domain(Domain(splitted[2], accession, start, end, start, end, start, end, stod(splitted[3]),max_id ));
		}
	}
}

template<typename MemoryType>
void
ProteinSequenceSet<MemoryType>::_read_sff(std::ifstream &domain_F)
{
	/*
	 SSF Format Description
	#---------------------------------------------------------------------------------
	# FILE FORMAT: Sam Summary File (SSF) Format 2.1
	#
	# Column 1:   Sequence accession code e.g. gi|159385
	# Column 2:   HMM model name (not taken from stat or align file; set externally by user)
	# Column 3:   Total sequence length
	# Column 4:   Total HMM Length
	# Column 5:   Length of alignment (upper case letters AND gaps that appear within segments)
	# Column 6:   Number of matched residues (upper case letters)
	# Column 7:   First residue in sequence to match model (first upper case letter)
	# Column 8:   Last residue in sequence to match model (last upper case letter)
	# Column 9:   First matched position in model
	# Column 10:  Last matched position in model
	# Column 11:  Evalue
	# Column 12:  Simple score
	# Column 13:  Reverse score
	# Column 14:  Number of matched sequence segments (2+: discontinuous match)
	# Column 15:  Colon-separated segment boundaries e.g. 1:51:99:163
	# Column 16 (Extended only):  1 or 0 indicating DomainFinder 3 mode (Cliquer or BMS).
	#==============================================================================*/

	std::string line;
	Seq_ptr tmp;
	size_t aln_start, hmm_start;
	int max_id=-1;
	std::map<std::string, int> name2id;
	std::map<std::string, int>::iterator it;
	std::string name, accession;
	while (getline(domain_F, line))
	{
		if ((line[0] == '#') || (line[0] == '\0'))
			continue;
		std::vector<std::string> splitted;
		boost::split(splitted, line, boost::is_any_of(" "), boost::token_compress_on);
		if (!this->contains(splitted[0]))
			continue;
		ProteinSequence &seq = (*this)[splitted[0]];
		aln_start=boost::lexical_cast<size_t>(splitted[6])-1;
		hmm_start=boost::lexical_cast<size_t>(splitted[8])-1;

		name=splitted[7];

		it=name2id.find(accession);
		if (it != name2id.end())
			seq.add_domain(Domain(name, accession, aln_start, boost::lexical_cast<size_t>(splitted[7])-1, 0, 0, hmm_start, boost::lexical_cast<size_t>(splitted[9])-1, boost::lexical_cast<double>(splitted[10]),it->second ));
		else
		{
			name2id[accession] = ++max_id;
			_domain_accessions.push_back(accession);
			seq.add_domain(Domain(name, accession, aln_start, boost::lexical_cast<size_t>(splitted[7])-1, 0, 0, hmm_start, boost::lexical_cast<size_t>(splitted[9])-1, boost::lexical_cast<double>(splitted[10]),max_id ));
		}
	}
}

/**
 * \brief Solves several problems given the used domains.
 * @param options The cleaning options to be performed
 */
template<typename MemoryType>
void
ProteinSequenceSet<MemoryType>::clean_up_domains(unsigned char options)
{
	size_t n_seqs = this->n_seqs();
	for (size_t i=0; i<n_seqs; ++i)
		(*this)[i].dom_archi().clean_up(options);
}

template<typename MemoryType>
void
ProteinSequenceSet<MemoryType>::refine_boundaries()
{

}


template<typename MemoryType>
void
ProteinSequenceSet<MemoryType>::write_domArchitecture(const std::string &out_f) const
{
	std::ofstream out_F(out_f);
	out_F << "## MDA FORMAT 1" << std::endl << std::endl;
	out_F << "#MDA" << std::endl;
	int n_archis = _dom_archis.size();
	for (int i=0; i<n_archis; ++i)
		out_F << std::right << std::setw(2)<< i+1 << " " << _dom_archis[i];

	out_F << std::endl << "# Arch-ID: Sequences" << std::endl;
	size_t j;
	int id;
	size_t n_seqs=this->n_seqs();
	for (int i=0; i<n_archis; ++i)
	{
		id = _dom_archis[i].id();
		out_F << i+1 << ":";
		for (j=0; j<n_seqs; ++j)
		{
			if ((*this)[j].dom_archi().id() == id)
				out_F << " " << (*this)[j].name();
		}
		out_F << std::endl;
	}
}



/**
 * \brief Splits a ProteinSequenceSet into columns according to its domains.
 *
 * The sequences of the set are split into domain and non-domain columns according to the domain architecture set. In case of gaps in the
 * domain architecture the domain column and the following non_domain column will contain an empty string.
 * \note The size of each occurring architectures needs to be the same.
 * \tparam The Memory type of The SequenceSet
 * @param set[in] The ProteinSequenceSet to split.
 * @param splitSet[out] The set to which the single columns will be added.
 * \relates ProteinSequenceSet
 */
template<typename MemoryType>
void
domain_column_split(const ProteinSequenceSet<MemoryType> &set, SplitSet<ProteinSequenceSet<Default> > &splitSet)
{
	size_t n_seqs=set.n_seqs();
	const DomainArchitectureSet &dom_arch_set= set.dom_archis();
	size_t n_domains=dom_arch_set[0].length();
	size_t n_columns = n_domains*2+1;
	size_t i,j;
	std::vector<int> pattern(n_columns, 0);
	for (j=0; j<n_domains; ++j)
	{
		if (dom_arch_set[0][j].id() != -1)
			pattern[j*2+1]=dom_arch_set[0][j].id()+1;
	}
	splitSet.pattern(pattern);

	for (i=0; i<n_columns; ++i)
		splitSet.push_back(new ProteinSequenceSet<Default>());

	size_t start;
	int id =0;
	int seq_id, arch_id;
	for (i=0; i<n_seqs; ++i)
	{
		start =0;
		const DomainArchitecture &dom_arch = dom_arch_set[0];//set[i].dom_archi().id()];
		arch_id=dom_arch.id();
		seq_id=set[i].id();
		for (j=0; j<n_domains; ++j)
		{
			if (dom_arch[j].id() != -1)
			{
				const Domain &dom = set[i].dom_archi()[dom_arch[j].ori_pos()];
				if (dom.seq_start()>0)
					splitSet[j*2].add_seq(substr_end_no_renaming(set[i], start, dom.seq_start()-1));
				else
					splitSet[j*2].add_seq(new ProteinSequence(set[i].name(), "", "", seq_id));
				splitSet[j*2+1].add_seq(substr_end_no_renaming(set[i], dom.seq_start(), dom.seq_end()));
				start=dom.seq_end()+1;
				splitSet[j*2+1][id].dom_archi().id(arch_id);
				splitSet[j*2][id].dom_archi().id(arch_id);
			}
		}

		// the last column
		if (start!=set[i].size())
			splitSet[j*2].add_seq(substr_end_no_renaming(set[i], start, set[i].size()-1));
		else
			splitSet[j*2].add_seq(new ProteinSequence(set[i].name(), "", "", set[i].id()));
		splitSet[j*2][id].id(seq_id);
		splitSet[j*2][id].dom_archi().id(arch_id);
		++id;
	}
}


/**
 * \brief Splits a set according to the domain architecture of the sequences.
 * @param set The sequence set.
 * @param architectureSplits The resulting split
 * \relates ProteinSequenceSet
 */
template<typename MemoryType>
void
splitByArchitecture(const ProteinSequenceSet<MemoryType> &set, std::vector<ProteinSequenceSet<MemoryType> > &architectureSplits)
{
	size_t nArchis = set.dom_archis().size();
	size_t nSeqs = set.size();
	architectureSplits.clear();
	architectureSplits.reserve(nArchis);

	size_t i;
	for (i=0; i<nArchis; ++i)
		architectureSplits.push_back(ProteinSequenceSet<MemoryType>(i));

	for (size_t i=0; i<nSeqs; ++i)
		architectureSplits[set[i].dom_archi().id()].share(set, i);

	for (size_t i=0; i<nArchis; ++i)
	{
		DomainArchitectureSet arch;
		arch.push_back(set.dom_archis()[i]);
		architectureSplits[i].dom_archis(arch);
	}
}

}

#endif /* PROTEINSEQUENCESET_HPP_ */
