/*
 * Seq_functs.hpp
 *
 *  Created on: Jun 26, 2013
 *      Author: Carsten Kemena
 *
 * This file is part of MDAT.
 *
 * MDAT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MDAT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with MDAT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * \file Seq_functs.hpp
 * \brief Contains several functions related to Sequence modification
 */
#ifndef SEQ_FUNCTS_HPP_
#define SEQ_FUNCTS_HPP_


#include <boost/lexical_cast.hpp>

#include <algorithm>
#include <memory>
#include <vector>


namespace MDAT
{


/**
 * \brief returns a new Sequence that is a substring of an existing one.
 * @param seq The sequence to produce the new one from.
 * @param start The start position of the substring.
 * @param length The lengths of the substr.
 * @return The new sequence.
 * \relatesalso Sequence
 */
template<typename SequenceType>
SequenceType
substr(const SequenceType &seq, size_t start, size_t length)
{
	return SequenceType(seq.name()+"_"+boost::lexical_cast<std::string>(start+1)+"-"+boost::lexical_cast<std::string>(start+length), seq.sequence().substr(start, length), seq.comment(), seq.id());
}

/**
 * \brief returns a new Sequence that is a substring of an existing one.
 * @param seq The sequence to produce the new one from.
 * @param start The start position of the substring.
 * @param length The lengths of the substr.
 * @return The new sequence.
 * \relatesalso Sequence
 */
template<typename SequenceType>
SequenceType
substr_no_renaming(const SequenceType &seq, size_t start, size_t length)
{
	return SequenceType(seq.name(), seq.sequence().substr(start, length), seq.comment(), seq.id());
}


template<typename SequenceType>
SequenceType*
substr_no_renaming2(const SequenceType &seq, size_t start, size_t length)
{
	return new SequenceType(seq.name(), seq.sequence().substr(start, length), seq.comment(), seq.id());
}

/**
 * \brief returns a new Sequence that is a substring of an existing one.
 * @param seq The sequence to produce the new one from.
 * @param start The start position of the substring.
 * @param end The last position of the substring in sequence.
 * @return The new sequence.
 * \relatesalso Sequence
 */
template<typename SequenceType>
SequenceType*
substr_end(const SequenceType &seq, size_t start, size_t end)
{
	return new SequenceType(seq.name()+"_"+boost::lexical_cast<std::string>(start+1)+"-"+boost::lexical_cast<std::string>(end+1), seq.sequence().substr(start, end-start+1), seq.comment(), seq.id());
}


template<typename SequenceType>
SequenceType*
substr_end_no_renaming(const SequenceType &seq, size_t start, size_t end)
{
	return new SequenceType(seq.name(), seq.sequence().substr(start, end-start+1), seq.comment(), seq.id());
}



void
reverse_complement(std::string &seq);


/**
 * \brief Computes the reverse complement of a DNA sequence
 * @param seq The sequence to change.
 * \relatesalso Sequence
 */
template<typename SequenceType>
void
reverse_complement(SequenceType &seq)
{
	size_t length =seq.size();
	seq.reverse();
	std::vector<char>transformation(128,'!');
	transformation['A']='T';
	transformation['a']='t';
	transformation['C']='G';
	transformation['c']='g';
	transformation['T']='A';
	transformation['t']='a';
	transformation['G']='C';
	transformation['g']='c';
	transformation['R']='Y';
	transformation['r']='y';
	transformation['Y']='R';
	transformation['y']='r';
	transformation['M']='K';
	transformation['m']='k';
	transformation['K']='M';
	transformation['k']='m';
	transformation['W']='W';
	transformation['w']='w';
	transformation['S']='S';
	transformation['s']='s';
	transformation['B']='V';
	transformation['b']='v';
	transformation['V']='B';
	transformation['v']='b';
	transformation['D']='H';
	transformation['d']='h';
	transformation['H']='D';
	transformation['h']='d';
	transformation['N']='N';
	transformation['n']='n';
	transformation['-']='-';

	for (size_t i=0; i<length; ++i)
		seq[i] = transformation[seq[i]];

}


}


#endif /* SEQ_FUNCTS_HPP_ */
