/*
 * Sequence.hpp
 *
 *  Created on: Jun 2, 2013
 *      Author: Carsten Kemena
 *
 * This file is part of MDAT.
 *
 * MDAT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MDAT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with MDAT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/**
 * \file Sequence.hpp
 *
 * \brief Contains the Sequence class.
 */
#ifndef SEQUENCE_HPP_
#define SEQUENCE_HPP_

// C++ header
#include <algorithm>
#include <map>
#include <memory>
#include <string>
#include <utility>
#include <vector>
#include <iostream>

// Boost header
#include <boost/lexical_cast.hpp>

// MDAT header
#include "Sequence_Interface.hpp"


namespace MDAT
{


/** \addtogroup Sequence
 *  @{
 */

/**
 * \brief A generic class to represent a plain sequence.
 *
 */
class Sequence
{
private:
	std::string _name;
	std::string _sequence;
	std::string _comment;
	size_t _id;

public:

	/******************************************************
	 *					Constructors & Destructors   	  *
	 ******************************************************/

	/**@{*/
	//! \name Constructors & Destructors
	/**
	 * \brief A simple sequence constuctor.
	 * @param seq_name The sequence name.
	 * @param comment_ A sequence comment.
	 * @param seq The sequence.
	 * @param seq_id The sequence id to give to the sequence.
	 */
	Sequence(const std::string &seq_name, const std::string &seq,  const std::string &comment_="", size_t seq_id=0);
	/**
	 * \brief A sequence constructor which allows the reservation of memory.
	 *
	 * The memory for the actual sequence will be reserved but the actual sequence can be larger without causing problems.
	 * @param seq_name The sequence name.
	 * @param comment_ A sequence comment.
	 * @param seq_length The sequence size for which memory should be reserved.
	 * @param seq_id The sequence id to give to the sequence.
	 */
	Sequence(const std::string &seq_name, const std::string &comment_, unsigned int seq_length, size_t seq_id=0);

	/**
	 * \brief The copy constructor.
	 *
	 * @param seq The sequence of wich a copy should be made.
	 */
	Sequence(const Sequence &seq);
	//Sequence(Sequence &&seq);

	/**
	 * \brief Destructor
	 */
	virtual ~Sequence();
	/**@}*/



	/**@{*/
	//! \name %Operators

	/**
	 * \brief Access operator.
	 * \param index The index of to access.
	 */
	char &operator[](unsigned int index)
	{
		return _sequence[index];
	}

	/**
	 * \overload
	 */
	const char &operator[](unsigned int index) const
	{
		return _sequence[index];
	}

	/**
	 * \brief Comparison operators.
	 * \param a The first sequence.
	 * \param b The second sequence.
	 */
	friend bool operator ==(const Sequence& a, const Sequence& b)
	{
		return(a.sequence() == b.sequence());
	}

	/**
	 * \brief Comparison operators.
	 * \param a The first sequence.
	 * \param b The second sequence.
	 */
	friend bool operator !=(const Sequence& a, const Sequence& b)
	{
		return(a.sequence() != b.sequence());
	}

	/**
	 * \brief Lesser operator.
	 * \param a The first sequence.
	 * \param b The second sequence.
	 */
	friend bool operator <(const Sequence& a, const Sequence& b)
	{
		return(a.sequence() < b.sequence());
	}

	/**
	 * \brief Greater operator.
	 * \param a The first sequence.
	 * \param b The second sequence.
	 */
	friend bool operator >(const Sequence& a, const Sequence& b)
	{
		return(a.sequence() > b.sequence());
	}

	/**
	 * \brief Lesser or equal operator.
	 * \param a The first sequence.
	 * \param b The second sequence.
	 */
	friend bool operator <=(const Sequence& a, const Sequence& b)
	{
		return(a.sequence() <= b.sequence());
	}

	/**
	 * \brief Greater or equal operator
	 * \param a The first sequence.
	 * \param b The second sequence.
	 */
	friend bool operator >=(const Sequence& a, const Sequence& b)
	{
		return(a.sequence()==b.sequence());
	}



	   Sequence & operator= (const Sequence & other)
	    {
	        if (this != &other) // protect against invalid self-assignment
	        {
	        	_name=other._name;
	        	_comment=other._comment;
	        	_sequence=other._sequence;
	        	_id=other._id;
	        }
	        // by convention, always return *this
	        return *this;
	    }

	   friend std::ostream& operator<< (std::ostream &out, const Sequence &seq);
	   /**@}*/
	//Functions

	/**@{*/
	//! \name Basic methods

	/**
	 * \brief Returns the name of the sequence
	 * \return The name.
	 */
	const
	std::string	& name() const
	{
		return _name;
	}

	void name(const std::string &na)
	{
		_name=na;
	}


	/**
	 * \brief Returns the sequence as a string.
	 * \return The sequence.
	 */
	const
	std::string	& sequence() const
	{
		return _sequence;
	}


	/**
	 * \brief Sets the sequence.
	 * @param seq The sequence
	 */
	void
	sequence(const std::string &seq)
	{
		_sequence=seq;
	}

	/**
	 * \brief Returns the comment of a string.
	 * \return The comment.
	 */
	const
	std::string	& comment() const
	{
		return _comment;
	}


	/**
	 * \brief Sets the comment.
	 * @param com The comment.
	 */
	void
	comment(const std::string &com)
	{
		_comment=com;
	}

	/**
	 * \brief The length of the sequence.
	 * \return Length of the sequence.
	 */
	size_t size() const
	{
		return _sequence.size();
	}

	/**
	 * \brief The length of the sequence.
	 * @return Length of the sequence.
	 */
	size_t length() const
	{
		return _sequence.size();
	}

	/**
	 * \brief The size of the sequence without gaps.
	 * \return The size of the ungapped sequence.
	 */
	size_t ungapped_size() const;

	/**
	 * \brief Returns the sequence id of the sequence.
	 *
	 * It represents the order in which the sequences were read.
	 * \return The id of the sequence.
	 */
	size_t id() const
	{
		return _id;
	}

	/**
	 * \brief Sets the id of the sequence.
	 * @param val The new value.
	 */
	void id(size_t val)
	{
		_id=val;
	}
	/**@}*/



	/**@{*/
	//! \name Manipulation methods


	/**
	 * \brief Appends a string to the sequence.
	 */
	template<class T>
	void append(const T &seq)
	{
		_sequence.append(seq);
	}

	/**
	 * \brief Appends a string to the sequence.
	 */
	void append(const Sequence &seq)
	{
		_sequence.append(seq.sequence());
	}


	/**
	 * \brief Appends a char to the sequence.
	 */
	void append(char c)
	{
		_sequence.push_back(c);
	}

	void append(size_t n, char c)
	{
		_sequence.append(n, c);
	}

	/**
	 * \brief Resizes the sequence.
	 * \param new_length The new_length.
	 */
	void
	resize(unsigned int new_length)
	{
		_sequence.resize(new_length);
	}

	/**
	 * \brief Turns all charachters of the sequence to uppercase.
	 */
	void
	to_upper()
	{
		std::string::iterator it, it_end =_sequence.end();
		for (it = _sequence.begin(); it != it_end; ++it)
			*it = toupper(*it);
	}

	/**
	 * \brief Turns all characters of the sequence to lowercase.
	 */
	void
	to_lower()
	{
		std::string::iterator it, it_end =_sequence.end();
		for (it = _sequence.begin(); it != it_end; ++it)
			*it = tolower(*it);
	}

	/**
	 * \brief Reverses the order of characters in the sequence.
	 */
	void
	reverse()
	{
		std::reverse(_sequence.begin(), _sequence.end());
	}

	/**
	 * \brief Inserts gaps into the sequence.
	 * \param vec A vector of pairs. The first value of a pair gives the position of the gap, the second one the length of it.
	 */
	void
	insert_gaps(const std::vector<std::pair<unsigned int, unsigned int> > vec);
	void
	insert_gaps(const std::string &edit_string);

	/**@}*/

};

/**
 *  @}
 */




template<typename SeqType>
char
identify_seq_type(const SeqType &seq)
{
	size_t seq_len = seq.size();
	char c;
	for (unsigned int i = 0; i < seq_len; ++i)
	{
		c = tolower(seq[i]);
		if ((c != 'a') && (c != 'c') && (c != 'g') && (c != 't') && (c != 'u'))
			return 'P';
	}
	return 'N';
}


template<typename SeqType>
std::pair<size_t, size_t>
coverage(const SeqType &seq1, const SeqType &seq2)
{
	size_t len=seq1.size();
	if (len != seq2.size())
		return std::pair<size_t, size_t>(-1,-1);
	size_t pair_len=0, pos=0;
	for (size_t i=0; i<len; ++i)
	{
		if ((seq1[i]!='-') || (seq2[i]!='-'))
			++pair_len;
		if ((seq1[i]!='-') && (seq2[i]!='-'))
			++pos;
	}

	return std::pair<size_t, size_t>(pos, pair_len);
}

template<typename SeqType>
std::pair<size_t, size_t>
id(const SeqType &seq1, const SeqType &seq2)
{
	size_t len=seq1.size();
	if (len != seq2.size())
		return std::pair<size_t, size_t>(-1,-1);
	size_t pair_len=0, pos=0;
	for (size_t i=0; i<len; ++i)
	{
		if ((seq1[i]!='-') || (seq2[i]!='-'))
			++pair_len;
		if (seq1[i] == seq2[i])
			++pos;
	}

	return std::pair<size_t, size_t>(pos, pair_len);
}


template<typename SeqType>
bool
seq_check(const SeqType &seq1, const SeqType &seq2)
{
	size_t seq1_l = seq1.size();
	size_t seq2_l = seq2.size();
	size_t i,j=0;
	for (i=0; i < seq1_l; ++i)
	{
		if (seq1[i] == '-')
			continue;
		while ((j<seq2_l) && (seq2[j] == '-'))
			++j;

		if ((j==seq2_l) || (tolower(seq1[i]) != tolower(seq2[j])))
			return 0;
		++j;
	}
	while (j!=seq2_l)
	{
		if (seq2[j] != '-')
			return 0;
		++j;
	}
	return 1;
}

template<typename SeqType>
void
removeGaps(SeqType &seq)
{
	int pos=-1;
	size_t len=seq.size();
	for (size_t i=0; i<len; ++i)
	{
		if (seq[i] != '-')
			seq[++pos] = seq[i];
	}
	seq.resize(++pos);
}

template<typename SeqType>
bool
bio_seq(const SeqType &seq)
{
	size_t len = seq.size();
	int c;
	for (size_t i=0; i<len; ++i)
	{
		c = tolower(seq[i]);
		if ((c!=45) && ((c<97) || (c>122)))
			return false;
	}

	return true;
}




/**
 * \brief Turns a DNA sequence into a ProteinSequence.
 * \tparam SequenceTypeIn The type of the sequence input
 * \tparam SequenceTypeOut The outputType of the sequence
 * @param seq The sequence to translate.
 * @return The new Sequence.
 * \relates Sequence
 */
template<typename SequenceTypeIn, typename SequenceTypeOut>
SequenceTypeOut
dna2prot(const SequenceTypeIn &seq)
{
	SequenceTypeOut new_seq(seq.name(), seq.comment(), seq.size()/3, seq.id());
	const std::string &seq_string = seq.sequence();
	new_seq.append(dna2prot<std::string,std::string>(seq_string));
	return new_seq;
}


template<>
std::string
dna2prot<std::string, std::string>(const std::string &seq);


} /* namespace Sequence */
#endif /* SEQUENCE_HPP_ */
