/*
 * SequenceSetBase.hpp
 *
 *  Created on: Jun 6, 2013
 *      Author: Carsten Kemena
 *
 * This file is part of MDAT.
 *
 * MDAT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MDAT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with MDAT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#ifndef SequenceSetBase_HPP_
#define SequenceSetBase_HPP_

// C header
#include <cstdlib>
#include <cstring>
#include <cstdio>
#include <iostream>
// C++ header
#include <algorithm>
#include <fstream>
#include <map>
#include <memory>
#include <string>
#include <vector>
#include <utility>


// MDAT header
#include "../Basics/basics.hpp"
#include "../Basics/utils.hpp"
#include "../utils/MDAT_Exceptions.hpp"
#include "../utils/filesystem.h"


// Boost header

namespace MDAT
{




/** \addtogroup SequenceSet
 *  @{
 */


/**
 * \brief The base class for a SequenceSet.
 * \tparam SequenceType The type of Sequence to be stored inside (Sequence, ProteinSequence or DNASequence).
 * \tparam MemoryType The memory mode to use for the sequence set (Default or MemSafe).
 */
template<typename SequenceType, typename MemoryType>
class SequenceSetBase
{


typedef std::shared_ptr<SequenceType> Seq_ptr;

private:
	std::vector<Seq_ptr> _seqs;
	char _seq_type;				// Sequence type
	std::string _file;			// File the alignment was read from
	size_t _id;
	mutable std::map<std::string, size_t> _name2index;

	// read multiple sequence alignment file
	short _identify_aln_format(FILE *aln_F);
	void _read_fasta_f(FILE *aln_F, const std::map<std::string, short> &seq_names);
	void _read_clustalw_f(FILE *aln_F, const std::map<std::string, short> &seq_names);
	void _read_msf_f(FILE *aln_F, const std::map<std::string, short> &seq_names);
	void _read_stockholm_f(FILE *aln_F, const std::map<std::string, short> &seq_names);
	void _read_codata_f(FILE *aln_F, const std::map<std::string, short> &seq_names);
	void _read_amps_f(FILE *aln_F, const std::map<std::string, short> &seq_names);
	void _read_phylip_f(FILE *aln_F, const std::map<std::string, short> &seq_names);


	// write alignment to file
	void _write_fasta(FILE *aln_f, unsigned int line_break=99999999) const;
	void _gap_replace();
	bool _seq_check();


public:
	/**
	 * \brief The sequence type used in the object.
	 */
	typedef SequenceType value_type;

	/***********************************************************************
	 * 						Constructors & Destructors					   *
	 ***********************************************************************/

	/**@{*/
	//! \name Constructors & Destructors


	/**
	 * \brief Constructor
	 */
	SequenceSetBase();
	SequenceSetBase(size_t value);
	//SequenceSetBase(SequenceSetBase &seqSet);
	//SequenceSetBase(SequenceSetBase &&seqSet);
	/**
	 * \brief Destructor
	 */
	virtual ~SequenceSetBase();
	/**@}*/

	/***********************************************************************
	 * 						Simple Access functions 					   *
	 ***********************************************************************/

	// Operators
	/**@{*/
	//! \name Operators

	/**
	 * \brief Operator to access the sequence
	 * @param index The sequence position to return.
	 * @return Pointer to the sequence.
	 */
	SequenceType &operator[](unsigned int index)
	{
		return *_seqs[index];
	}

	/**
	 * \overload
	 */
	const SequenceType &operator[](unsigned int index) const
	{
		return *_seqs[index];
	}


	/**
	 * \brief Access a function by name.
	 * @param seq_name The name of the sequence
	 * @return The Sequence
	 */
	SequenceType &operator[](const std::string &seq_name)
	{
		if (_name2index.size() != _seqs.size())
		{
			size_t n_seqs = _seqs.size();
			for (size_t i=0; i<n_seqs; ++i)
				_name2index[_seqs[i]->name()]=i;
		}
		return *_seqs[_name2index[seq_name]];
	}

	/**
	 * \overload
	 */
	const SequenceType &operator[](const std::string &seq_name) const
	{
		if (_name2index.size() != _seqs.size())
		{
			size_t n_seqs = _seqs.size();
			for (size_t i=0; i<n_seqs; ++i)
				_name2index[_seqs[i]->name()]=i;
		}
		return *_seqs[_name2index[seq_name]];
	}

	bool
	contains(const std::string &seq_name) const
	{
		if (_name2index.size() != _seqs.size())
		{
			size_t n_seqs = _seqs.size();
			for (size_t i=0; i<n_seqs; ++i)
				_name2index[_seqs[i]->name()]=i;
		}
		return _name2index.find(seq_name) != _name2index.end();

	}

	/**
	 * Simple print of a SequenceSet in fasta format.
	 * @param out[in|out] The output stream.
	 * @param seqSet[in] The sequence set
	 * @return The output stream
	 */
	template<typename SeqType, typename MemType>
	friend std::ostream& operator<< (std::ostream &out, const SequenceSetBase<SeqType, MemType> &seqSet);


	/**@}*/


	/**@{*/
	//! \name Basic methods

	/**
	 * \brief Returns a sequence.
	 * @param index Index of the sequence.
	 * @return const reference to the sequence.
	 */
	const SequenceType* seq(unsigned int index) const
	{
		return &(*(_seqs[index]));
	}


	/**
	 * \brief returns the number of sequences.
	 * \return The number of sequences.
	 */
	size_t n_seqs() const
	{
		return _seqs.size();
	}

	/**
	 * \brief returns the number of sequences.
	 * @return The number of sequences
	 */
	size_t size() const
	{
		return _seqs.size();
	}


	/**
	 * \brief Returns the length of the sequence inside
	 * \details Returns the length of the first sequence or 0 if not existant.
	 * @return The length
	 */
	size_t
	length() const
	{
		if (_seqs.empty())
			return 0;
		else
			return _seqs[0]->size();
	}

	/**
	 * \brief The average size of the sequence set.
	 * \return The average size.
	 */
	double avg_size() const;

	/**
	 * \brief Returns true if no sequences are contained in this object.
	 */
	bool empty() const
	{
		return (_seqs.empty());
	}

	/**
	 * \brief Returns the file the sequences were read from.
	 */
	std::string file() const
	{
		return _file;
	}

	/**
	 * \brief Returns type.
	 * @return The sequence type.
	 */
	char seq_type() const throw()
	{
		return _seq_type;
	}

	/**
	 * \brief Sets the sequence type.
	 * @param seq_type_ The sequence type.
	 */
	void seq_type(char seq_type_) throw()
	{
		_seq_type = seq_type_;
	}


	/**
	 * \brief Returns the id of the set.
	 * @return The id
	 */
	int id() const throw()
	{
		return _id;
	}

	/**
	 * \brief Sets the id of the sequence set
	 * @param val The id
	 */
	void id(int val)
	{
		_id=val;
	}

	/**
	 * \brief Sets everything to 0.
	 */
	void clear()
	{
		_seqs.clear();
		_file.clear();
	}
	/**@}*/


	//******   read/write SequenceSetBases   ******
	/**@{*/
	//! \name Input & Output

	/**
	 * \brief Extracts a subalignment from the sequence set.
	 *
	 * Only sequences which are denoted in seq_names are extracted. Names in seq_names not occurring in the alignment are ignored. Columns consisting of gaps only are removed.
	 * @param seq_f The file of the sequences to read.
	 * @param seq_names The names to read.
	 * @param format The format of the alignment. (-1 enables automatic format detection)
	 * @param check Checks if the sequence is a proper biological sequence
	 */
	virtual void read(const std::string &seq_f, const std::vector<std::string> &seq_names, bool check=false, short format =-1);

	/**
	 * \brief Reads a set of sequences.
	 *
	 * This function can read unaligned sequences in FASTA format as well as aligned sequences in several formats.
	 * @param seq_f The file with the sequences to read.
	 * @param format The format of the alignment. (-1 enables automatic format detection)
	 * @param check Checks if the sequence is a proper biological sequence
	 */
	virtual void read(const std::string &seq_f, bool check=false, short format =-1)
	{
		read(seq_f, std::vector<std::string>(), check, format);
	}


	/**
	 * \brief Writes the sequences into a file.
	 *
	 * This function supports the following formats: FASTA, MSF.
	 * @param seq_f The file to write the alignment to
	 * @param format The format to use (fasta, clustalw, msf, phylip_i, phylip_s)
	 */
	virtual void write(const std::string &seq_f, const std::string format) const;


	/**
	 * \brief Append a sequence to a set.
	 * @param seq A pointer to the new sequence.
	 */
	void add_seq(SequenceType *seq)
	{
		_seqs.push_back(std::shared_ptr<SequenceType> (seq));
	}


	/**@}*/


	//******   Manipulation methods   ******
	/**@{*/
	//! \name Manipulation methods


	/**
	 * \brief Turns all characters to uppercase.
	 */
	void to_upper()
	{
		for (size_t i = 0; i<_seqs.size(); ++i)
			_seqs[i]->to_upper();
	}

	/**
	 * \brief Turns all characters to lowercase.
	 */
	void to_lower()
	{
		for (size_t i = 0; i<_seqs.size(); ++i)
			_seqs[i]->to_lower();
	}



	/**
	 * \brief Deletes sequences from the alignment
	 * \param names The names of the sequences to delete
	 */
	virtual void delete_seqs(const std::map<std::string,bool> &names);

	/**
	 * \brief Deletes sequences from the alignment
	 * \param indices The indices of the sequences to delete.
	 */
	virtual void delete_seqs(std::vector<size_t> &indices);

	/**
	 * \brief Deletes sequences if they are not in the given list.
	 * \param indices The indices of the sequences to keep.
	 */
	virtual void keep_seqs(std::vector<size_t> &indices);


	/**
	 * \brief Shares a sequence between two sets
	 * @param set The set to take the sequence from.
	 * @param id The index of the sequence.
	 */
	void
	share(const SequenceSetBase<SequenceType, MemoryType> &set, size_t id)
	{
		_seqs.push_back(Seq_ptr());
		_seqs[_seqs.size()-1] = set._seqs[id];
	}


	/**
	 * \brief Transfer a sequence from one set to another.
	 * @param set The set to take the sequence from.
	 * @param id The index of the sequence.
	 */
	void
	transfer(SequenceSetBase<SequenceType, MemoryType> &set, size_t id)
	{
		_seqs.push_back(Seq_ptr(set._seqs[id]));
		set._seqs.erase(set._seqs.begin()+id);
	}

	/**
	 * \brief Transfers all sequences from one set to another.
	 * @param set The set to take the sequence from.
	 */
	void
	transfer(SequenceSetBase<SequenceType, MemoryType> &set)
	{
		size_t n_seqs=set.n_seqs();
		_seqs.reserve(this->n_seqs()+n_seqs);
		for (size_t i=0; i<n_seqs; ++i)
			_seqs.push_back(Seq_ptr(set._seqs[i]));
		set.clear();
	}


	/**
	 * \brief Sorts the sequences.
	 * \param type "input" sorts the sequences by order of the input. "name" sorts by sequence name. "seq" sorts the sequences by alphabetical order.
	 */
	void sort(std::string type);


	/**
	 * \brief Inserts gaps into each sequence.
	 * @param edit_string The matter of matches and gaps in reverse order
	 */
	void
	insert_gaps(const std::string &edit_string);

	/**@}*/





};

/**
 *  @}
 */


/***********************************************************************
 * 						Constructors & Destructors					   *
 ***********************************************************************/

template<typename SequenceType, typename MemoryType>
SequenceSetBase<SequenceType, MemoryType>::SequenceSetBase():_seqs(), _seq_type('x'), _file(""), _id(-1), _name2index()
{}

template<typename SequenceType, typename MemoryType>
SequenceSetBase<SequenceType, MemoryType>::SequenceSetBase(size_t value):_seqs(), _seq_type('x'), _file(""), _id(value), _name2index()
{}

/*template<typename SequenceType, typename MemoryType>
SequenceSetBase<SequenceType, MemoryType>::SequenceSetBase(SequenceSetBase<SequenceType, MemoryType> &&seqSet)
{
	*this=std::move(seqSet);
}*/

template<typename SequenceType, typename MemoryType>
SequenceSetBase<SequenceType, MemoryType>::~SequenceSetBase()
{

}



template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::_gap_replace()
{
	size_t num_seqs = this->n_seqs();
	size_t n_cols;
	size_t j;

	for (size_t i=0; i<num_seqs; ++i)
	{
		SequenceType &tmp_seq = *_seqs[i];
		n_cols = tmp_seq.size();
		for (j=0; j<n_cols; ++j)
		{
			if (!isalpha(tmp_seq[j]))
					tmp_seq[j]='-';
		}
	}
}


template<typename SequenceType, typename MemoryType>
short
SequenceSetBase<SequenceType, MemoryType>::_identify_aln_format(FILE *aln_F)
{
	const unsigned int LINE_LENGTH = 501;
	char line[LINE_LENGTH];
	char *tmp, *pos;
	while (fgets(line, LINE_LENGTH, aln_F) != nullptr)
	{
		if (line[0] == '>')
		{
			if (fgets(line, LINE_LENGTH, aln_F)==nullptr)
			{
				fprintf(stderr, "Error occurred when trying to identify format!\n");
				exit(EXIT_FAILURE);
			}
			if (line[0] == '>')
				return 6; // AMPS
			else
				return 1; // Fasta
		}
		if (strstr(line, "CLUSTAL ") != nullptr)
			return 2; // Clustal
		if (strstr(line, "MSF:") != nullptr)
			return 3; // MFS
		if (strstr(line, "# STOCKHOLM 1.0") != nullptr)
			return 4; // Stockholm
		if (strstr(line, "ENTRY") != nullptr )
			return 5; // Codata

		//check Phylip
		StrTok strTok(line);
		tmp = strTok.next(" \n\t");
		if (tmp == nullptr)
			continue;
		pos=tmp;
		while (*pos != '\0')
			if (!isdigit(*(pos++)))
				continue;
		tmp = strTok.next(" \n\t");
		if (tmp == nullptr)
			continue;
		pos=tmp;
		while (*pos != '\0')
			if (!isdigit(*(pos++)))
				continue;
		tmp = strTok.next(" \n\t");
		if (tmp == nullptr)
			return 7; // Phylip
	}
	return 0;
}


template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::read(const std::string &seq_f, const std::vector<std::string> &seq_names, bool check, short format)
{
	FILE *aln_F = my_fopen(seq_f.c_str(), "r");

	_file = seq_f;
	if (format <0)
		format = _identify_aln_format(aln_F);
	std::map<std::string, short> extract_only;
	size_t n_extract_seqs = seq_names.size();
	for (unsigned int i = 0; i < n_extract_seqs; ++i)
		extract_only.insert(std::pair<std::string,short>(seq_names[i],1));

	fseek ( aln_F , 0 , SEEK_SET );
	switch (format)
	{
		case 1:
			_read_fasta_f(aln_F, extract_only);
			break;
		case 2:
			_read_clustalw_f(aln_F, extract_only);
			break;
		case 3:
			_read_msf_f(aln_F, extract_only);
			break;
		case 4:
			_read_stockholm_f(aln_F, extract_only);
			break;
		case 5:
			_read_codata_f(aln_F, extract_only);
			break;
		case 6:
			_read_amps_f(aln_F, extract_only);
			break;
		case 7:
			_read_phylip_f(aln_F, extract_only);
			break;
		default:
			throw Alignment_Exception("Format could not be identified");
	}
	fclose(aln_F);

	if ((check) && (!check_set(*this)))
	{
		throw Alignment_Exception("Sequence contains bad character");
	}
	this->seq_type(identify_seq_type(*(this->_seqs)[0]));
}


template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::_read_fasta_f(FILE *aln_F, const std::map<std::string, short> &seq_names)
{
	const unsigned int LINE_LENGTH = 500;
	char line[LINE_LENGTH];
	SequenceType *tmp_seq = nullptr;
	char *comment = nullptr, *name = nullptr;
	bool read_sequence = true;
	size_t id=0;
	size_t seq_length=0;
	while (fgets(line, LINE_LENGTH, aln_F))
	{
		if (line[0] == '>')
		{
			if (this->n_seqs())
				seq_length = tmp_seq->size();
			StrTok tokenizer(&line[1]);
			name = tokenizer.next(" \n");
			comment = tokenizer.next("\n");
			if (seq_names.empty() || (seq_names.count(name)>0))
			{
				read_sequence = true;
				if (comment == nullptr)
					tmp_seq = new SequenceType(name, "", seq_length, id++);
				else
					tmp_seq = new SequenceType(name, comment, seq_length, id++);
				_seqs.push_back(Seq_ptr(tmp_seq));
			}
			else
				read_sequence = false;
		}
		else if (line[0] == '/')
			break;
		else
		{
			if (read_sequence)
			{
				tmp_seq->append(line);
				if ((*tmp_seq)[tmp_seq->size()-1] == '\n')
					tmp_seq->resize(tmp_seq->size()-1);
			}
		}
	}
}


template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::_read_clustalw_f(FILE *aln_F, const std::map<std::string, short> &seq_names)
{
	const unsigned int LINE_LENGTH = 200;
	char line[LINE_LENGTH];
	if (fgets(line, LINE_LENGTH, aln_F)==nullptr)
	{
		fprintf(stderr, "Error when reading file.");
		exit(EXIT_FAILURE);
	}
	long t_pos = 0;
	size_t id = 0;
	while ((fgets(line, LINE_LENGTH, aln_F) != nullptr) && (line[0] == '\n'))
		t_pos = ftell(aln_F);
	fseek(aln_F, t_pos, SEEK_SET);

	//read first block in format
	SequenceType *seq_p = nullptr;
	char *name = nullptr, *tmp_seq = nullptr;
	std::vector<short> read_sequence;
	StrTok tokenizer;
	while (fgets(line, LINE_LENGTH, aln_F) != nullptr)
	{
		if ((line[0] == ' ') || (line[0] == '\n'))
			break;
		else
		{
			tokenizer.set(line);
			name = tokenizer.next(" \n");
			if (seq_names.empty() || seq_names.count(name) > 0)
			{
				read_sequence.push_back(1);
				tmp_seq = tokenizer.next(" \n");
				seq_p = new SequenceType(name, "", tmp_seq, id++);
				_seqs.push_back(Seq_ptr(seq_p));
			}
			else
				read_sequence.push_back(0);
		}
	}

	// Read rest of the alignment
	unsigned int counter = 0;
	unsigned int seq_id = 0;
	while (fgets(line, LINE_LENGTH, aln_F) != nullptr)
	{
		if ((line[0] == ' ') || (line[0] == '\n'))
		{
			seq_id = 0;
			counter = 0;
		}
		else
		{
			if (read_sequence[counter])
			{
				tokenizer.set(line);
				tokenizer.next(" \n");
				tmp_seq = tokenizer.next(" \n");
				_seqs[seq_id++]->append(tmp_seq);
			}
			++counter;
		}
	}
}


template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::_read_msf_f(FILE *aln_F, const std::map<std::string, short> &seq_names)
{
	char *msf;
	const unsigned int LINE_LENGTH = 501;
	char line[LINE_LENGTH];
	while ((fgets(line, LINE_LENGTH, aln_F) != nullptr) && ((msf=strstr(line, "MSF:")) == nullptr));
	msf +=4;
	size_t seq_length=atoi(msf);
	char *seq_name;
	SequenceType *tmp_seq;
	StrTok tokenizer;
	std::vector<short> read_sequence;
	std::size_t id = 0;

	//read comment block
	while (fgets(line, LINE_LENGTH, aln_F) != nullptr)
	{
		if ((line[0] == '/') && (line[1] == '/'))
			break;
		if (((seq_name=strstr(line, "Name:")) != nullptr) || ((seq_name=strstr(line, "NAME:")) != nullptr))
		{
			tokenizer.set(seq_name+5);
			seq_name = tokenizer.next(" \n");
			if (seq_names.empty() || (seq_names.count(seq_name)>0))
			{
				read_sequence.push_back(1);
				tmp_seq = new SequenceType(seq_name, "", seq_length, id++);
				_seqs.push_back(Seq_ptr(tmp_seq));

			}
			else
				read_sequence.push_back(0);
		}
	}

	//read alignment
	char *pos;
	char c;
	unsigned int counter = 0;
	unsigned int seq_id = 0;
	while (fgets(line, LINE_LENGTH, aln_F) != nullptr)
	{
		if (line[0] == '\n')
		{
			counter = 0;
			seq_id = 0;
		}
		else
		{
			if (read_sequence[counter]==1)
			{
				pos = line;
				while (*(++pos) != ' ');
				while (1)
				{
					while ((c=*pos) != '\0')
					{
						if ((c=='-') || (c=='.') || (isalpha(c)))
							_seqs[seq_id]->append(c);
						++pos;
					}
					if (*(pos-1)=='\n')
						break;
					else
					{
						if(fgets(line, LINE_LENGTH, aln_F)==nullptr)
						{
							fprintf(stderr, "Erorr occured\n");
							exit(EXIT_FAILURE);
						}
						pos = &line[0];
					}
				}
				++seq_id;
			}
			else
			{
				while (1)
				{
					pos = line;
					while (*pos != '\0')
						++pos;
					if (*(pos-1)=='\n')
						break;
					if (fgets(line, LINE_LENGTH, aln_F)==nullptr)
					{
						fprintf(stderr, "Erorr occured\n");
						exit(EXIT_FAILURE);
					}
				}
			}
			++counter;
		}
	}
	_gap_replace();
}


template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::_read_stockholm_f(FILE *aln_F, const std::map<std::string, short> &seq_names)
{
	const unsigned int LINE_LENGTH = 501;
	char line[LINE_LENGTH];
	char *part1_p = nullptr, *part2_p =nullptr;
	StrTok tokenizer;
	SequenceType *seq_p = nullptr;
	char *end=&line[LINE_LENGTH-1];
	bool use = false;
	bool append = false;
	size_t seq_id = 0;
	std::vector<short> read_seq;
	*end=8;
	size_t id = 0;
	while (fgets(line, LINE_LENGTH, aln_F))
	{
		if (line[0] == '\n')
		{
			append = true;
			seq_id = 0;
		}
		else if (line[0] == '#')
		{
			while ((*end != 8) && (*end != '\n'))
			{
				*end = 8;
				if (!fgets(line, LINE_LENGTH, aln_F))
				{
					fprintf(stderr, "Error occurred\n");
					exit(EXIT_FAILURE);
				}
			}
		}
		else if (line[0] != '/')
		{
			tokenizer.set(line);
			part1_p = tokenizer.next(" \n");
			part2_p = tokenizer.next(" \n");
			if (part2_p != nullptr)
			{
				if (seq_names.empty() || (seq_names.count(part1_p)))
				{
					use=true;
					if (append)
					{
						seq_p = &(*_seqs[seq_id]);
						seq_p->append(part2_p);
					} else {
						seq_p = new SequenceType(part1_p, part2_p, " ", id++);
						_seqs.push_back(Seq_ptr(seq_p));
					}
					++seq_id;
				}
				else
					use=false;
			}
			else
			{
				if (use)
					seq_p->append(part1_p);
			}
		} else
			use=false;
		*end=8;
	}

	//Turn "." into "-"
	_gap_replace();
}


template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::_read_codata_f(FILE *aln_F, const std::map<std::string, short> &seq_names)
{
	const unsigned int LINE_LENGTH = 501;
	char line[LINE_LENGTH];
	char *name=nullptr;
	size_t seq_length = 0;
	StrTok tokenizer;
	SequenceType *seq_p = nullptr;
	size_t id = 0;
	char c;
	unsigned int pos1 = 0, pos2 = 1;
	bool extract_seq = false;
	while (fgets(line, LINE_LENGTH, aln_F))
	{
		if (!strncmp(line, "ENTRY", 5))
		{
			tokenizer.set(&line[6]);

			name = tokenizer.next(" \n");
		}
		else if (!strncmp(line, "SEQUENCE", 8))
		{
			if (seq_names.empty() || seq_names.count(name))
			{
				seq_p = new SequenceType(name, "", seq_length, id++);
				_seqs.push_back(Seq_ptr(seq_p));
				if (!fgets(line, LINE_LENGTH, aln_F))
				{
					fprintf(stderr, "Error occurred\n");
					exit(EXIT_FAILURE);
				}
				extract_seq = true;
			}
			else
				extract_seq = false;
		}
		else if (strncmp(line, "///", 3) == 0)
		{
			if (extract_seq)
				seq_length = seq_p->size();
		}
		else
		{
			if (extract_seq)
			{
				pos1 = 0;
				pos2 = 0;
				while ((c=line[pos1++]) != '\0')
				{
					if (isalpha(c) || (c == '-'))
						line[pos2++] = c;
				}
				line[pos2] = '\0';
				seq_p->append(line);
			}
		}
	}
}


template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::_read_amps_f(FILE *aln_F, const std::map<std::string, short> &seq_names)
{
	const unsigned int LINE_LENGTH = 501;
	char line[LINE_LENGTH];
	std::vector<short> extract_seq;
	StrTok tokenizer;
	SequenceType *seq_p;
	size_t id = 0;
	char *name = nullptr;
	unsigned int num_seqs = 0;
	while (fgets(line, LINE_LENGTH, aln_F))
	{
		if (line[0] == '*')
			break;
		else if (line[0] == '>')
		{
			++num_seqs;
			tokenizer.set(&line[1]);
			name = tokenizer.next("\n");
			if (seq_names.empty() || seq_names.count(name))
			{
				seq_p = new SequenceType(name, "", 0, id++);
				_seqs.push_back(Seq_ptr(seq_p));
				extract_seq.push_back(1);
			}
			else
				extract_seq.push_back(0);
		}
	}
	char c;
	char *seq_line = new char[num_seqs+2];
	unsigned int n_seqs2 = num_seqs+2;
	unsigned int i;
	unsigned int seq_id = 0;
	while (fgets(seq_line, n_seqs2, aln_F))
	{
		if (seq_line[0] == '*')
			break;
		else
		{
			seq_id = 0;
			for (i=0; i<num_seqs; ++i)
			{
				if (extract_seq[i])
				{
					if ((c = seq_line[i]) == ' ')
						_seqs[seq_id]->append('-');
					else
						_seqs[seq_id]->append(c);
					++seq_id;
				}
			}
		}
	}
	delete[] seq_line;
}


// reads interleaved and sequential Phylip format
template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::_read_phylip_f(FILE *aln_F, const std::map<std::string, short> &seq_names)
{
	const unsigned int LINE_LENGTH = 500;
	char line[LINE_LENGTH];
	std::vector<int> use_seq;
	size_t num_seqs, seq_length;
	if (fgets(line, LINE_LENGTH, aln_F))
	{
		fprintf(stderr, "Error when reading file.");
		exit(EXIT_FAILURE);
	}
	sscanf(line, "%lu %lu", &num_seqs, &seq_length);
	char name[11];
	name[10] = '\0';
	char *pos;
	SequenceType *seq_p;
	size_t seq_num = num_seqs;
	size_t id = 0;
	char c;
	size_t line_num = 0;
	int x =-1;
	while (fgets(line, LINE_LENGTH, aln_F))
	{
		if (line[0] == '\n')
			continue;

		//Read sequence name
		if (line_num < num_seqs)
		{
			strncpy(name, line, 10);
			pos=name;
			while ((*pos != '\0') && (*pos != ' '))
				++pos;
			*pos = '\0';
			if ((seq_names.empty()) || (seq_names.count(name)))
			{
				use_seq.push_back(++x);
				seq_p = new SequenceType(name, "", seq_length, id++);
				_seqs.push_back(Seq_ptr(seq_p));
			}
			else
			{
				use_seq.push_back(-1);
			}
			pos=&line[10];
			++line_num;
		}
		else
			pos=&line[0];

		//read_sequence
		++seq_num;
		if (seq_num >= num_seqs)
			seq_num=0;
		if (use_seq[seq_num] >= 0)
		{

			seq_p=&(*_seqs[use_seq[seq_num]]);
			while (1)
			{
				while ((c=*pos) != '\0')
				{
					if ((isalpha(c)) || (c=='-'))
						seq_p->append(c);
					++pos;
				}

				if (*(pos-1) != '\n')
				{
					if (!fgets(line, LINE_LENGTH, aln_F))
						break;
					pos=&line[0];
				}
				else
					break;
			}
		}
		else
		{
			while (line[strlen(line)-1] != '\n')
			{
				if (!fgets(line, LINE_LENGTH, aln_F))
					break;
			}
		}

	}
}



// Write functions

template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::write(const std::string &aln_f, const std::string format) const
{
	std::string format_lc;
	format_lc.reserve(format.size());
	for (size_t i= 0; i<format.size(); ++i)
		format_lc.push_back(tolower(format[i]));

	FILE *aln_F;
	if ((format_lc == "fasta"))
		aln_F = my_fopen(aln_f.c_str(), "w");
	else
		throw(Alignment_Exception("Unknown alignment format!"));

	if (format_lc == "fasta")
		_write_fasta(aln_F);

	fclose(aln_F);
}


template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::_write_fasta(FILE *aln_F, unsigned int line_break) const
{
	unsigned int num_seqs = this->n_seqs();
	size_t current_length, seq_length;
	std::string tmp_seq;
// 	char symbol='\0';

	for (unsigned int i = 0; i < num_seqs; ++i)
	{
		if (_seqs[i]->comment().empty())
			fprintf(aln_F, ">%s\n", _seqs[i]->name().c_str());
		else
			fprintf(aln_F, ">%s %s\n", _seqs[i]->name().c_str(), _seqs[i]->comment().c_str());
		current_length = 0;
		tmp_seq = _seqs[i]->sequence();
		seq_length=_seqs[i]->size();
		while (current_length < seq_length)
		{
			fprintf(aln_F, "%.*s\n", line_break, &tmp_seq.c_str()[current_length]);
			current_length += line_break;
		}

	}
}





template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::delete_seqs(std::vector<size_t> &indices)
{
	std::sort(indices.begin(), indices.end());
	size_t n_dels = indices.size();
	size_t num_seqs = this->n_seqs();
	size_t index_pos = 0;
	size_t seq_pos = 0;
	for (size_t i = 0; i < num_seqs; ++i)
	{
		if ((index_pos==n_dels) || ((index_pos<n_dels) && (indices[index_pos] != i)))
			_seqs[seq_pos++] = _seqs[i];
		else
			++index_pos;
	}
	_seqs.resize(seq_pos);
}

template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::keep_seqs(std::vector<size_t> &indices)
{
	std::sort(indices.begin(), indices.end());
	size_t n_dels = indices.size();
	size_t num_seqs = this->n_seqs();
	size_t index_pos = 0;
	size_t seq_pos = 0;
	for (size_t i = 0; i < num_seqs; ++i)
	{
		if (indices[index_pos] == i)
		{
			_seqs[seq_pos++] = _seqs[i];
			++index_pos;
		}
		if (index_pos==n_dels)
			break;
	}
	_seqs.resize(seq_pos);
}

template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::delete_seqs(const std::map<std::string,bool> &names)
{
	size_t num_seqs = this->n_seqs();
	size_t seq_pos = 0;
	for (size_t i = 0; i < num_seqs; ++i)
	{
		if (names.count(_seqs[i]->name()))
			_seqs[seq_pos++] = _seqs[i];
	}
	_seqs.resize(seq_pos);
}




template <typename T>
struct Input_Sort : public std::binary_function<T,T,bool>
{
    bool operator()(const T& o1, const T& o2)
    {
        return o1->id() < o2->id();
    }
};


template <typename T>
struct Seq_Sort : public std::binary_function<T,T,bool>
{
    bool operator()(const T& o1, const T& o2)
    {
        return *o1<*o2;
    }
};


template <typename T>
struct Name_Sort : public std::binary_function<T,T,bool>
{
    bool operator()(const T& o1, const T& o2)
    {
        return o1->name() < o2->name();
    }
};

template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::sort(std::string type)
{
	if (type == "id")
	{
		std::sort(_seqs.begin(), _seqs.end(), Input_Sort<std::shared_ptr<SequenceType> >());
	}
	else if (type == "seq")
	{
		std::sort(_seqs.begin(), _seqs.end(), Seq_Sort<std::shared_ptr<SequenceType> >());
	}
	else if (type == "name")
	{
		std::sort(_seqs.begin(), _seqs.end(), Name_Sort<std::shared_ptr<SequenceType> >());
	}
}


template<typename SequenceType, typename MemoryType>
void
SequenceSetBase<SequenceType, MemoryType>::insert_gaps(const std::string &edit_string)
{
	size_t n_seqs=_seqs.size();
	for (size_t i=0; i<n_seqs; ++i)
		_seqs[i]->insert_gaps(edit_string);
}


template<typename SequenceType, typename MemoryType>
bool
check_set(const SequenceSetBase<SequenceType, MemoryType> &set)
{
	size_t n_seqs = set.n_seqs();
	size_t j, len;
	size_t *val_counting = new size_t[256];
	for (j=0; j<256; ++j)
		val_counting[j] = 0;
	// counting occurrences of characters
	for (size_t i=0; i<n_seqs; ++i)
	{

		const SequenceType &seq = set[i];

		len=seq.size();
		for (j=0; j<len; ++j)
			++val_counting[static_cast<int>(seq[j])];
	}

	// see if strange character has been found
	for (j=0; j<45; ++j)
	{
		if (val_counting[j] != 0)
			return false;
	}
	for (j=46; j<65; ++j)
	{
		if (val_counting[j] != 0)
			return false;
	}
	for (j=91; j<97; ++j)
	{
		if (val_counting[j] != 0)
			return false;
	}
	for (j=123; j<256; ++j)
	{
		if (val_counting[j] != 0)
			return false;
	}
	delete[] val_counting;
	return true;
}

template<typename SeqType, typename MemType>
std::ostream& operator<< (std::ostream &out, const SequenceSetBase<SeqType, MemType> &seqSet)
{
	size_t n_seqs = seqSet.n_seqs();
	for (size_t i=0; i<n_seqs; ++i)
		out << seqSet[i] << std::endl;
	return out;
}


}

#endif /* SequenceSetBase_HPP_ */
