/*
 * SequenceSet_memsafe.hpp
 *
 *  Created on: Jun 2, 2013
 *      Author: Carsten Kemena
 *
 * This file is part of MDAT.
 *
 * MDAT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MDAT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with MDAT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef SequenceSetBase_IO_MEMSAFE_HPP_
#define SequenceSetBase_IO_MEMSAFE_HPP_


// C header
#include <cstdlib>
#include <cstring>
#include <cstdio>

// C++ header
#include <algorithm>
#include <fstream>
#include <string>
#include <vector>


// MDAT header
#include "../Basics/basics.hpp"
#include "../Basics/utils.hpp"
#include "../utils/filesystem.h"


// Boost header
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>



using boost::bad_lexical_cast;

namespace MDAT
{
/** \addtogroup SequenceSet
 *  @{
 */

/**
 * \brief This SequenceSet class uses less memory than the normal one.
 * \details This memory uses less memory than the default class. This achieved by reading a sequence only when it is needed. The functionality
 * is reduced compared to the default one.
 * \warning Sequences are read when accessed. They only stay available until a new one is accessed.
 */
template<typename SequenceType>
class SequenceSetBase<SequenceType, MemSafe>
{

typedef std::shared_ptr<SequenceType> Seq_ptr;

private:
	Seq_ptr _seq;

	mutable std::vector<std::streampos> _id_index;
	mutable std::map<std::string, size_t> _name_index;
	mutable bool _complete;
	mutable bool _index_changed;
	mutable size_t _current_id;
	mutable std::string _current_name;
	mutable std::ifstream _seq_F;
	mutable size_t _n_seqs;

	char _seq_type;				// Sequence type
	std::string _file;			// File the alignment was read from

	void
	_read_single_seq(const std::streampos &pos) const
	{
		std::string line;
		_seq_F.seekg(pos);
		getline(_seq_F, line);
		size_t length=line.size();
		size_t i;
		for (i=0; i<length; ++i)
		{
			if (line[i]==' ')
				break;
		}
		StrTok tokenizer(&line[1]);
		_seq->name(tokenizer.next(" \n"));
		char *comment = tokenizer.next("\n");
		_seq->comment((comment != NULL)?comment: "" );

		_seq->sequence("");
		while (std::getline(_seq_F, line))
		{
			if (line[0] == '>')
				break;
			_seq->append(line);
		}
		if (!_seq_F.good())
		{
			_complete=true;
			_seq_F.clear();
		}
	}

	Seq_ptr
	_access_value(size_t index_val) const
	{

		if (index_val!=_current_id)
		{
			if (index_val >= _n_seqs)
			{
				std::string line;
				size_t i;
				size_t length;
				_seq_F.seekg(_id_index[_n_seqs-1]);
				std::getline(_seq_F, line);
				std::streampos pos=0;
				while (std::getline(_seq_F, line))
				{
					if (line[0] =='>')
					{
						length=line.size();
						for (i=0; i<length; ++i)
						{
							if (line[i]==' ')
								break;
						}

						std::string seq_name = line.substr(1,i-1);
						_id_index.push_back(pos);
						_name_index[seq_name]=_n_seqs;
						++_n_seqs;
						if ((_n_seqs) == index_val+1)
							break;
					}
					pos=_seq_F.tellg();
				}
				_index_changed=true;
			}
			_current_id=index_val;
			_current_name=_seq->name();
			_read_single_seq(_id_index[index_val]);
		}
		return _seq;
	}

	Seq_ptr
	_access_value(const std::string seq_id) const
	{
		std::map<std::string, size_t>::iterator it_end= _name_index.end();
		if (_name_index.find(seq_id)==it_end)
		{
			std::string line;
			size_t i;
			size_t length;
			_seq_F.seekg(_id_index[_n_seqs-1]);
			std::getline(_seq_F, line);
			std::streampos pos=0;
			while (std::getline(_seq_F, line))
			{
				if (line[0] =='>')
				{
					length=line.size();
					for (i=0; i<length; ++i)
					{
						if (line[i]==' ')
							break;
					}

					std::string seq_name = line.substr(1,i-1);
					_id_index.push_back(pos);
					_name_index[seq_name]=pos;
					++_n_seqs;
					if (seq_name==seq_id)
						break;
				}
				pos=_seq_F.tellg();
			}
			_read_single_seq(_id_index[_n_seqs-1]);
			_index_changed=true;
		}
		else
			_read_single_seq(_id_index[_name_index[seq_id]]);
		return _seq;
	}


public:

	/***********************************************************************
	 * 						Constructors & Destructors					   *
	 ***********************************************************************/


	SequenceSetBase():_complete(false),_index_changed(false),_current_id(-1),_n_seqs(0),_seq_type('x')
	{
		SequenceType *seq = new SequenceType("","",0);
		_seq.reset(seq);
	}

	SequenceSetBase(const std::string &seq_f)
	{
		SequenceType *seq = new SequenceType("","",0);
		_seq.reset(seq);
		set_file(seq_f);
	}

	SequenceSetBase(const std::string &seq_f, const std::string &index_f)
	{
		SequenceType *seq = new SequenceType("","",0);
		_seq.reset(seq);
		set_file(seq_f, index_f);
	}

	virtual ~SequenceSetBase()
	{
		_seq_F.close();
	}


	/***********************************************************************
	 * 						Simple Access functions 					   *
	 ***********************************************************************/

	// Operators
	/**@{*/
	//! \name Operators

	Seq_ptr
	next()
	{
		if (((_current_id+1)==_n_seqs) && (_complete))
			return Seq_ptr(NULL);
		else
			return _access_value(_current_id+1);
	}


	/**
	 * \brief Operator to access the sequence
	 * @param index The sequence position to return.
	 * @return Pointer to the sequence.
	 */
	SequenceType &operator[](unsigned int index)
	{
		return *_access_value(index);
	}
	SequenceType &operator[](std::string &seq_id)
	{
		return *_access_value(seq_id);
	}

	/**
	 * \overload
	 */
	const SequenceType &operator[](unsigned int index) const
	{
		return *_access_value(index);
	}

	const SequenceType &operator[](const std::string &seq_id) const
	{
		return *_access_value(seq_id);
	}
	/**@}*/


	/**@{*/
	//! \name Basic methods

	void
	build_index() const
	{
		std::streampos pos;
		std::string line;
		_seq_F.seekg(_id_index[_n_seqs-1]);
		std::getline(_seq_F, line);
		size_t i, length;
		while (!_complete)
		{
			while (std::getline(_seq_F, line))
			{
				if (line[0] == '>')
					break;
				pos=_seq_F.tellg();
			}
			if (!_seq_F.good())
			{
				_complete=true;
				break;
			}
			length=line.size();
			for (i=0; i<length; ++i)
			{
				if (line[i]==' ')
					break;
			}
			_id_index.push_back(pos);
			_name_index[line.substr(1,i-1)] = _n_seqs;
			++_n_seqs;
			_index_changed=true;
		}
		_seq_F.clear();
	}

	size_t
	n_seqs() const
	{
		if (!_complete)
			build_index();
		return _n_seqs;
	}

	void
	read(const std::string &seq_f)
	{
		set_file(seq_f);
	}

	void
	set_file(const std::string &seq_f)
	{
		_current_id=0;
		_complete=false;
		_seq_F.open(seq_f);
		_id_index.clear();
		_name_index.clear();

		std::string line;
		std::streampos pos=_seq_F.tellg();
		while (std::getline(_seq_F, line))
		{
			if (line[0] == '>')
				break;
			pos=_seq_F.tellg();

		}
		_n_seqs=1;
		size_t i;
		size_t length=line.size();
		for (i=0; i<length; ++i)
		{
			if (line[i]==' ')
				break;
		}

		std::string seq_name = line.substr(1,i-1);
		_id_index.push_back(pos);
		_name_index[seq_name]=pos;
		_read_single_seq(pos);
		_index_changed=true;
	}

	void
	set_file(const std::string &seq_f, const std::string &index_f)
	{
		_current_id=0;
		_seq_F.open(seq_f);
		_id_index.clear();
		_name_index.clear();
		read_index(index_f);
		_index_changed=false;
		_read_single_seq(_id_index[0]);
	}

	const SequenceType* seq(unsigned int index) const
	{
		return &(*(_access_value(index)));
	}

	void
	read_index(const std::string &index_f)
	{
		std::ifstream index_F(index_f);
		std::string line;
		std::vector<std::string> list;
		size_t seq_id;
		getline(index_F, line); //Version check
		if (line=="#MDAT - SeqIndex v1.0")
		{
			getline(index_F, line); // is completed
			_n_seqs=boost::lexical_cast<size_t>(line);
			_id_index.resize(_n_seqs);
			getline(index_F, line);
			if (line=="complete")
				_complete=true;
			else
				_complete=false;
			while(getline(index_F, line))
			{
				split(list, line, boost::is_any_of(" "));
				std::streampos pos = boost::lexical_cast<size_t>(list[2]);
				seq_id = boost::lexical_cast<size_t>(list[1]);
				_id_index[seq_id] = pos;
				_name_index[list[0]] = seq_id;
			}
		}
		else
		{
			std::cerr << "Problem reading the index file" << std::endl;
		}
		index_F.close();
	}


	void
	write_index(const std::string &index_f) const
	{
		std::ofstream index_F(index_f);
		std::map<std::string, size_t>::const_iterator it,it_end=_name_index.end();
		index_F << "#MDAT - SeqIndex v1.0" << std::endl;
		index_F << _n_seqs << std::endl;
		if (_complete)
			index_F << "complete" << std::endl;
		else
			index_F << "incomplete" << std::endl;
		for (it=_name_index.begin(); it!=it_end; ++it)
		{
			index_F << it->first << " " << it->second << " " << _id_index[it->second] << std::endl;
		}
	}

	bool
	index_changed() const
	{
		return _index_changed;
	}

};

/**
 *  @}
 */


}

#endif
