/*
 * Library.hpp
 *
 *  Created on: Apr 6, 2012
 *      Author: Carsten Kemena
 *
 * This file is part of MDAT++.
 *
 * MDAT++ is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MDAT++ is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with MDAT++.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


/**
 * \file Library.hpp
 * \brief Contains the Library class to handle pairwise matches.
 */
#ifndef LIBRARY_H_
#define LIBRARY_H_

// C header
#include <cstdlib>

// C++ header
#include <algorithm>
#include <limits>
#include <map>
#include <memory>
#include <utility>
#include <vector>


#include <iostream>
// MDAT header
#include "../utils/Matrix.hpp"
#include "../utils/ThreadPool.hpp"

namespace MDAT {


/**
 * \brief A match.
 */

typedef std::pair<unsigned int, unsigned int> Match;


/**
 * \brief Simple struct to save a match and its score.
 */
struct Match_point
{
	/**
	 * \brief Constructor which sets every value to 0.
	 */
	Match_point():x(0),y(0),score(0.0)
	{}

	/**
	 * \brief Constructor initializing values to user given values.
	 * \param x_ The first position.
	 * \param y_ The second position.
	 * \param score_ The score of the match.
	 */
	Match_point(unsigned int x_, unsigned int y_, float score_):x(x_),y(y_),score(score_)
	{}

	/**
	 * \brief Lesser operator.
	 * 
	 * First the first coordinate is compared then the second.
	 * \param a The first point.
	 * \param b The second point.
	 */
	friend bool operator <(const Match_point &a, const Match_point &b)
	{
		if (a.x != b.x)
			return (a.x < b.x);
		else
			return (a.y < b.y);
	}

	unsigned int x;
	unsigned int y;
	float score;
};


/**
 * \brief A Library object stores pairwise matches and is able to do the consistency extension.
 *
 * \tparam DataType DataType specifies the kind of data to be used as a base (e.g. SequenceSet). It has to support the functions size() and the access operator.
 * The subelements have to support the size() function as well.
 *
 * The Library concept has been introduced by C. Notredame et al. ("T-Coffee: A novel method for fast and accurate multiple sequence alignment", J Mol Biol 302, 2000).
 * It stores the matches produced by an alignment method and can produce extension (reweighting scores according to consistency using a third sequence).
 *
 *
 */
template <typename DataType>
class Library
{

private:

	const DataType &_data;
	size_t _n_elems;
	size_t _max_length;
	size_t _n_pairs;

	std::auto_ptr<std::vector<std::vector<Match_point> > > _pairs;

	struct RelaxHelper
	{
		Matrix<float> final_mat;
		Matrix<float> tmp_mat;
		std::vector<std::pair<int,int> > used;
		std::vector<int> n_hits;

		RelaxHelper(size_t max_length): final_mat(Matrix<float>(max_length, max_length)), tmp_mat(Matrix<float>(max_length, max_length*2)), used(max_length*max_length), n_hits(std::vector<int>(max_length*2, 0))
		{
		}

	};

	template <typename RelaxFunc>
	void
	_relax_pair(size_t i, std::vector<std::vector<Match_point> > *new_vals, RelaxFunc relax_function, RelaxHelper &helper);

public:
	// Constructors & Destructors

	/**@{*/
	//! \name Constructors & Destructors

	/**
	 * \brief Constructor
	 * \param data The basic data the Library is constructed for.
	 */
	Library(const DataType &data);


	/**
	 * \brief Destructor
	 */
	virtual ~Library();
	/**@}*/


	/**@{*/
	//! \name Input & Output


	/**
	 * \brief Adds a matching pare to the library.
	 *
	 * If the pair does not exist yet in the library, the score will be set to 1. If the pair already exists, the score is increased by 1.
	 * \param id1 The sequence id of the first sequence.
	 * \param id2 The sequence id of the second sequence.
	 * \param pos1 The residue of sequence 1.
	 * \param pos2 The residue of sequence 2.
	 * \param score The score of the pair.
	 */
	void add(unsigned int id1, unsigned int id2, unsigned int pos1, unsigned int pos2, float score=1.0);

	/**
	 * \brief Calculates the scores for each residue pair.
	 * \param[in] id1 The id of sequence 1
	 * \param[in] id2 The id of sequence 2
	 * \param[out] match_points In this structure, the pairs with the scores will be saved.
	 */
	void get(size_t id1, size_t id2, std::map<Match, int> &match_points) const;

	/**
	 * \brief Calculates the scores for each residue pair.
	 *
	 * This function compares the matching scores for two sets. The scores between every pair of elements from the different sets are computed
	 * and converted to fit the aligned sequences.
	 * \tparam GapFunc A function pointer or a functor.
	 * \param[in] ids1 The first alignment.
	 * \param[in] ids2 The second alignment.
	 * \param[in] data The data base data_set.
	 * \param[out] match_points In this structure, the pairs with the scores will be saved.
	 * \param[in] gap_func A simple function returning true if the element is a gap else false.
	 *
	 * \sa get
	 */
	template<typename GapFunc>
	void get(const std::vector<size_t> &ids1, const std::vector<size_t> &ids2, std::map<Match, int> &match_points, const GapFunc gap_func) const;


	/**
	 * \brief Reserves additional memory for a certain pair of elements.
	 * @param id1 The id of the first element.
	 * @param id2 The id of the second element.
	 * @param n_entries The number of additional matches that will be added.
	 */
	void reserve_add_memory(unsigned int id1, unsigned int id2, size_t n_entries);
	/**@}*/


	/**@{*/
	//! \name Modify

	/**
	 * \brief Reweights the scores of each match according to its consistency with other sequences.
	 *
	 * This function checks the consistency of each pairwise match with every other sequence in the set. The higher the consistency the higher will be
	 * the final score of the match.
	 * \attention The original scores will be replaced, so that this function should only be called once every matches have been added.
	 * @param relax_function The function to be used for the reweighting.
	 */
	template <typename RelaxFunc>
	void
	relax(RelaxFunc relax_function);

	template<typename RelaxFunc>
	void
	relax(RelaxFunc relax_function, int n_threads);

/*
	void
	sort();*/

	void
	print()
	{
		int i,j,k;
		int pos;
		for (i=0; i<_n_elems; ++i)
		{
			for (j=i+1; j<_n_elems; ++j)
			{
				printf("%i %i\n", i, j);
				pos= i*_n_elems-(i*i+i)/2 + j -i-1;
				std::vector<Match_point> &matches=(*_pairs)[pos];
				for (k=0; k<matches.size(); ++k)
					printf("(%i %i %f) ", matches[k].x, matches[k].y, matches[k].score);
				printf("\n");
			}
		}
	}
	/**@}*/
};




template<typename DataType>
Library<DataType>::Library(const DataType &set):_data(set), _n_elems(set.size()), _max_length(0),_n_pairs((_n_elems*(_n_elems-1))/2), _pairs(new std::vector<std::vector<Match_point> >(_n_pairs))
{
	for (size_t i = 0; i<_n_elems; ++i)
	{
		if (set[i].length() > _max_length)
			_max_length = set[i].length();
	}
}


template<typename DataType>
Library<DataType>::~Library() {
	// TODO Auto-generated destructor stub
}


template<typename DataType>
void
Library<DataType>::add(unsigned int id1, unsigned int id2, unsigned int pos1, unsigned int pos2, float score)
{
	// the index of a pair i/j (with i<j) is calculated as following: i*n_seq - (i*(i+1)/2) +j-i-1
	// i*n_seq - (i*(i+1)/2) gives the first index having i at the beginning
	// j-i-1 adds the shift necessary for the pair i/j
	score=100;
	unsigned int id;
	if (id1 < id2)
	{
		id = id1*_n_elems - (id1*(id1+1)/2) +id2-id1-1;
		(*_pairs)[id].push_back(Match_point(pos1, pos2, score));
	}
	else
	{
		id = id2*_n_elems - (id2*(id2+1)/2) +id1-id2-1;
		(*_pairs)[id].push_back(Match_point(pos2, pos1, score));
	}
}

template<typename DataType>
void
Library<DataType>::reserve_add_memory(unsigned int id1, unsigned int id2, size_t n_entries)
{
	unsigned int id;
	if (id1 < id2)
		id = id1*_n_elems - (id1*(id1+1)/2) +id2-id1-1;
	else
		id = id2*_n_elems - (id2*(id2+1)/2) +id1-id2-1;
	(*_pairs)[id].reserve((*_pairs)[id].size()+n_entries);
}

template<typename DataType>
void
Library<DataType>::get(size_t id1, size_t id2, std::map<Match, int> &match_points) const
{
	unsigned int id;
	if (id1 < id2)
	{
		id = id1*_n_elems - (id1*(id1+1)/2) +id2-id1-1;
		std::vector<Match_point> &matches=(*_pairs)[id];
		size_t n_matches=matches.size();
		for (size_t i=0; i<n_matches; ++i)
			match_points[Match(matches[i].x, matches[i].y)] = matches[i].score;
	}
	else
	{
		id = id2*_n_elems - (id2*(id2+1)/2) +id1-id2-1;
		std::vector<Match_point> &matches=(*_pairs)[id];
		size_t n_matches=matches.size();
		for (size_t i=0; i<n_matches; ++i)
			match_points[Match(matches[i].y, matches[i].x)] = matches[i].score;
	}


}

template<typename DataType>
template<typename GapFunc>
void Library<DataType>::get(const std::vector<size_t> &ids1, const std::vector<size_t> &ids2, std::map<Match, int> &match_points, const GapFunc gap_func) const
{
	match_points.clear();
	size_t n_ids1=ids1.size();
	size_t n_ids2=ids2.size();
	size_t n_elems1=_data[ids1[0]].length();
	size_t n_elems2=_data[ids2[0]].length();
	size_t i,j,k,pos;
	std::map<Match, int> tmp_match_points;
	std::map<Match, int>::iterator it, it_end;
	std::vector<unsigned int> convert1, convert2;
	convert1.resize(n_elems1);
	convert2.resize(n_elems2);
	std::pair<std::map<Match, int>::iterator,bool> ret;
	for (i=0; i<n_ids1; ++i)
	{
		pos=0;
		for (k=0; k<n_elems1; ++k)
		{
			if (!gap_func(_data[ids1[i]],k))
				convert1[pos++] = k;
		}

		for (j=0; j<n_ids2; ++j)
		{
			pos=0;
			for (k=0; k<n_elems2; ++k)
			{
				if (!gap_func(_data[ids2[j]],k))
					convert2[pos++] = k;
			}

			get(_data[ids1[i]].id(), _data[ids2[j]].id(), tmp_match_points);
			it_end=tmp_match_points.end();
			for (it=tmp_match_points.begin(); it!=it_end; ++it)
			{
				Match tmp_m(convert1[it->first.first], convert2[it->first.second]);
				ret=match_points.insert(std::pair<Match, int>(tmp_m, it->second));
				if (!ret.second)
					ret.first->second+=it->second;

			}
		}
	}
	it_end = match_points.end();
	size_t overall=n_ids1*n_ids2;
	for (it = match_points.begin(); it != it_end; ++it )
		it->second/=overall;

}



template<typename DataType>
template<typename RelaxFunc>
void
Library<DataType>::relax(RelaxFunc relax_function, int n_threads)
{
	if (n_threads <=1)
	{
		relax(relax_function);
		return;
	}
	size_t i;
	std::vector<std::vector<Match_point> > *new_vals = new std::vector<std::vector<Match_point> >();
	new_vals->resize(_n_pairs);
	ThreadPool<RelaxHelper> pool(n_threads, RelaxHelper(_max_length));
	for (i=0; i<_n_elems; ++i)
		pool.addTask(&Library<DataType>::_relax_pair<RelaxFunc>, this, i, new_vals, relax_function);
	pool.stop();
	_pairs.reset(new_vals);
}



template<typename DataType>
template<typename RelaxFunc>
void
Library<DataType>::relax(RelaxFunc relax_function)
{
	std::vector<std::vector<Match_point> > *new_vals = new std::vector<std::vector<Match_point> >();
	new_vals->resize(_n_pairs);
	Matrix<float> final_mat(_max_length, _max_length);
	Matrix<float> tmp_mat(_max_length, _max_length*2);
	std::vector<std::pair<int,int> > used;
	used.resize(_max_length*_max_length);
	std::vector<int> n_hits(_max_length*2, 0);
	size_t use=0;

	std::vector<Match_point>::const_iterator it,it_end;

	size_t i,j,k,m,id, current_pair_id, pos;
	int l;
	for (i=0; i<_n_elems; ++i)
	{
		for (j=i+1; j<_n_elems; ++j)
		{
			current_pair_id=i*_n_elems - (i*(i+1)/2) +j-i-1;
			it_end=(*_pairs)[current_pair_id].end();
			use=0;
			for (it=(*_pairs)[current_pair_id].begin(); it!=it_end; ++it)
			{
				final_mat[it->x][it->y]+=2*it->score;
				used[use].first=it->x;
				used[use].second=it->y;
				++use;
			}

			for (k=0; k<_n_elems; ++k)
			{
				if ((k==i) || (k==j))
					continue;
//					tmp_mat.fill(0);
				if (k<i)
				{

					id=k*_n_elems - (k*(k+1)/2) +i-k-1;
					it=(*_pairs)[id].begin();
					it_end=(*_pairs)[id].end();
					for (it=(*_pairs)[id].begin(); it!=it_end; ++it)
					{
						pos=n_hits[it->x]*2;
						++n_hits[it->x];
						tmp_mat[it->x][pos]=it->y;
						tmp_mat[it->x][++pos]=it->score;
					}
				}
				else
				{
					id=i*_n_elems - (i*(i+1)/2) +k-i-1;
					it=(*_pairs)[id].begin();
					it_end=(*_pairs)[id].end();
					for (it=(*_pairs)[id].begin(); it!=it_end; ++it)
					{
						pos=n_hits[it->y]*2;
						++n_hits[it->y];
						tmp_mat[it->y][pos]=it->x;
						tmp_mat[it->y][++pos]=it->score;
					}
				}
				if (k<j)
				{
					id=k*_n_elems - (k*(k+1)/2) +j-k-1;
					it=(*_pairs)[id].begin();
					it_end=(*_pairs)[id].end();
					for (it=(*_pairs)[id].begin(); it!=it_end; ++it)
						for(l=0; l<n_hits[it->x]; ++l)
							if (final_mat[tmp_mat[it->x][l*2]][it->y])
								final_mat[tmp_mat[it->x][l*2]][it->y]+= relax_function(tmp_mat[it->x][l*2+1], it->score);
							/*else
							{
								used[use].first=tmp_mat[it->x][l*2];
								used[use].second=it->y;
								++use;
								final_mat[tmp_mat[it->x][l*2]][it->y] = relax_function(tmp_mat[it->x][l*2+1], it->score);
							}*/
				}
				else
				{
					id=j*_n_elems - (j*(j+1)/2) +k-j-1;
					it=(*_pairs)[id].begin();
					it_end=(*_pairs)[id].end();
					for (it=(*_pairs)[id].begin(); it!=it_end; ++it)
						for(l=0; l<n_hits[it->y]; ++l)
							if (final_mat[tmp_mat[it->y][l*2]][it->x])
								final_mat[tmp_mat[it->y][l*2]][it->x]+= relax_function(tmp_mat[it->y][l*2+1], it->score);
							/*else
							{
								used[use].first=tmp_mat[it->y][l*2];
								used[use].second=it->x;
								++use;
								final_mat[tmp_mat[it->y][l*2]][it->x]= relax_function(tmp_mat[it->y][l*2+1], it->score);
							}*/

				}
				// reset tmp_mat to 0
				if (k<i)
				{
					id=k*_n_elems - (k*(k+1)/2) +i-k-1;
					it=(*_pairs)[id].begin();
					it_end=(*_pairs)[id].end();
					for (it=(*_pairs)[id].begin(); it!=it_end; ++it)
					{
						pos=n_hits[it->x]*2;
						++n_hits[it->x];
						tmp_mat[it->x][pos]=it->y;
						tmp_mat[it->x][++pos]=0;
					}
				}
				else
				{
					id=i*_n_elems - (i*(i+1)/2) +k-i-1;
					it=(*_pairs)[id].begin();
					it_end=(*_pairs)[id].end();
					for (it=(*_pairs)[id].begin(); it!=it_end; ++it)
					{
						pos=n_hits[it->y]*2;
						++n_hits[it->y];
						tmp_mat[it->y][pos]=it->x;
						tmp_mat[it->y][++pos]=0;
					}
				}
				for (m=0; m<2*_max_length; ++m)
					n_hits[m]=0;
			}

			for (k=0; k<use; ++k)
			{
				(*new_vals)[current_pair_id].push_back(Match_point(used[k].first, used[k].second, (final_mat[used[k].first][used[k].second])));
				final_mat[used[k].first][used[k].second] = 0;
			}
			use=0;
		}
	}
	_pairs.reset(new_vals);
}


template<typename DataType>
template<typename RelaxFunc>
void
Library<DataType>::_relax_pair(size_t i, std::vector<std::vector<Match_point> > *new_vals, RelaxFunc relax_function, RelaxHelper &helper)
{
	Matrix<float> &final_mat = helper.final_mat;
	Matrix<float> &tmp_mat = helper.tmp_mat;
	std::vector<std::pair<int,int> > used = helper.used;
	std::vector<int> &n_hits = helper.n_hits;
	size_t use=0;
	std::vector<Match_point>::const_iterator it,it_end;
	size_t k,m,id, current_pair_id, pos;
	int l;
	for (size_t j=i+1; j<_n_elems; ++j)
	{
		current_pair_id=i*_n_elems - (i*(i+1)/2) +j-i-1;
		it_end=(*_pairs)[current_pair_id].end();
		use=0;
		for (it=(*_pairs)[current_pair_id].begin(); it!=it_end; ++it)
		{
			final_mat[it->x][it->y]+=2*it->score;
			used[use].first=it->x;
			used[use].second=it->y;
			++use;
		}

		for (k=0; k<_n_elems; ++k)
		{
			if ((k==i) || (k==j))
				continue;
	//		tmp_mat.fill(0);
			if (k<i)
			{

				id=k*_n_elems - (k*(k+1)/2) +i-k-1;
				it=(*_pairs)[id].begin();
				it_end=(*_pairs)[id].end();
				for (it=(*_pairs)[id].begin(); it!=it_end; ++it)
				{
					pos=n_hits[it->x]*2;
					++n_hits[it->x];
					tmp_mat[it->x][pos]=it->y;
					tmp_mat[it->x][++pos]=it->score;
				}
			}
			else
			{
				id=i*_n_elems - (i*(i+1)/2) +k-i-1;
				it=(*_pairs)[id].begin();
				it_end=(*_pairs)[id].end();
				for (it=(*_pairs)[id].begin(); it!=it_end; ++it)
				{
					pos=n_hits[it->y]*2;
					++n_hits[it->y];
					tmp_mat[it->y][pos]=it->x;
					tmp_mat[it->y][++pos]=it->score;
				}
			}
			if (k<j)
			{
				id=k*_n_elems - (k*(k+1)/2) +j-k-1;
				it=(*_pairs)[id].begin();
				it_end=(*_pairs)[id].end();
				for (it=(*_pairs)[id].begin(); it!=it_end; ++it)
					for(l=0; l<n_hits[it->x]; ++l)
						if (final_mat[tmp_mat[it->x][l*2]][it->y])
							final_mat[tmp_mat[it->x][l*2]][it->y]+= relax_function(tmp_mat[it->x][l*2+1], it->score);
						/*else
						{
							used[use].first=tmp_mat[it->x][l*2];
							used[use].second=it->y;
							++use;
							final_mat[tmp_mat[it->x][l*2]][it->y] = relax_function(tmp_mat[it->x][l*2+1], it->score);
						}*/
			}
			else
			{
				id=j*_n_elems - (j*(j+1)/2) +k-j-1;
				it=(*_pairs)[id].begin();
				it_end=(*_pairs)[id].end();
				for (it=(*_pairs)[id].begin(); it!=it_end; ++it)
					for(l=0; l<n_hits[it->y]; ++l)
						if (final_mat[tmp_mat[it->y][l*2]][it->x])
							final_mat[tmp_mat[it->y][l*2]][it->x]+= relax_function(tmp_mat[it->y][l*2+1], it->score);
						/*else
						{
							used[use].first=tmp_mat[it->y][l*2];
							used[use].second=it->x;
							++use;
							final_mat[tmp_mat[it->y][l*2]][it->x]= relax_function(tmp_mat[it->y][l*2+1], it->score);
						}*/

			}
			if (k<i)
			{

				id=k*_n_elems - (k*(k+1)/2) +i-k-1;
				it=(*_pairs)[id].begin();
				it_end=(*_pairs)[id].end();
				for (it=(*_pairs)[id].begin(); it!=it_end; ++it)
				{
					pos=n_hits[it->x]*2;
					++n_hits[it->x];
					tmp_mat[it->x][pos]=it->y;
					tmp_mat[it->x][++pos]=0;
				}
			}
			else
			{
				id=i*_n_elems - (i*(i+1)/2) +k-i-1;
				it=(*_pairs)[id].begin();
				it_end=(*_pairs)[id].end();
				for (it=(*_pairs)[id].begin(); it!=it_end; ++it)
				{
					pos=n_hits[it->y]*2;
					++n_hits[it->y];
					tmp_mat[it->y][pos]=it->x;
					tmp_mat[it->y][++pos]=0;
				}
			}
			for (m=0; m<2*_max_length; ++m)
				n_hits[m]=0;
		}

		for (k=0; k<use; ++k)
		{
			(*new_vals)[current_pair_id].push_back(Match_point(used[k].first, used[k].second, (final_mat[used[k].first][used[k].second])));
			final_mat[used[k].first][used[k].second] = 0;
		}
	}
}



/*
void
Library::print(const MDAT::SequenceSet &set, std::string out_f)
{
	FILE *out_F = my_fopen(out_f, "w");
	vector<Match_point>::const_iterator it, it_end;
	size_t i, j;

	// add information from direct pairwise alignment
	// the index of a pair i/j (with i<j) is calculated as following: i*n_seq - (i*(i+1)/2) +j-i-1
	// i*n_seq - (i*(i+1)/2) gives the first index having i at the beginning
	// j-i-1 adds the shift necessary for the pair i/j
	unsigned int id;
	for (i=0; i<_n_elems; ++i)
	{
		for (j=i+1; j<_n_elems; ++j)
		{
			fprintf(out_F, "#%li %li\n", i, j);
			id = i*_n_elems - (i*(i+1)/2) +j-i-1;
			it     = (*_pairs)[id].begin();
			it_end = (*_pairs)[id].end();

			while (it != it_end)
			{
				fprintf(out_F, "%u %u %f\n", it->x, it->y, it->score);
				++it;
			}

		}
	}
	fclose(out_F);
}
*/






} /* namespace MDAT */
#endif /* LIBRARY_H_ */
