/*
 * consistency_aln.hpp
 *
 *  Created on: Sep 12, 2013
 *      Author: ckeme_01
 */

/**
 * \file consistency_aln.hpp
 *
 * \brief Alignment functions to be used in the consistency approach.
 */
#ifndef CONSISTENCY_ALN_HPP_
#define CONSISTENCY_ALN_HPP_

// C++ header
#include <string>
#include <algorithm>
#include <utility>

// MDAT header
#include "Library.hpp"
#include "../utils/MatrixStack.hpp"
#include "../clustering/Tree.hpp"

namespace MDAT{


/**
 * \brief This is a special version of the Needleman-Wunsch algorithm for consistency chaining.
 * 
 * This algorithm does not include gap costs and and allows matches only when there is a positive value. The reason is that the score are coming from the consistency approach
 * and thus only matches should occur that have been seen in a pairwise alignment.
 * \param dim1 The first dimension.
 * \param dim2 The second dimension.
 * \param matrix The dynamic programming matrix.
 */
void
nw_dyn_consistency(size_t dim1, size_t dim2, Matrix<std::pair<int, char> > &matrix);

/**
 * \brief Traceback for the nw_dyn_consistency algorithm.
 * @param i The size of the first dimension.
 * @param j The size of the second dimension.
 * @param matrix The matrix filled by nw_dyn_consistency function.
 * @param edit_string1 The first edit_string.
 * @param edit_string2 The second edit_string.
 */
void
nw_dyn_consistency_traceback(size_t i, size_t j, const Matrix<std::pair<int, char> > &matrix, std::string &edit_string1, std::string &edit_string2);


/**
 * \brief Calculation of pairwise alignment.
 * 
 * This function uses the gotoh algorithm for the consistency algorithm. Only matches with a score != 0 are allowed.
 * \param dim1 First dimension.
 * \param dim2 Second dimension.
 * \param matrices The dynamic programming matrices.
 * \param gop Gap opening costs.
 * \param gep Gap extension costs.
 */
void
gotoh_dyn_consistency(int dim1, int dim2, MatrixStack<3, std::pair<int, char> > &matrices, int gop, int gep);

/**
 * \brief Traceback for the gotoh_dyn_consistency algorithm.
 * @param i The size of the first dimension.
 * @param j The size of the second dimension.
 * @param matrix The matrix filled by nw_dyn_consistency function.
 * @param edit_string1 The first edit_string.
 * @param edit_string2 The second edit_string.
 */
void
gotoh_dyn_consistency_traceback(size_t i, size_t j, MatrixStack<3,std::pair<int, char> > &matrices, std::string &edit_string1, std::string &edit_string2);


/**
 * Fills the dynamic programming matrix with values from the Library.
 * @param matrices The dynamic programming matrix.
 * @param dim1 The size of the first dimension.
 * @param dim2 The size of the second dimension.
 * @param match_points The scores of the matches.
 */
/*template<typename DataType>
void
enterDataIntoMatrix(MatrixStack<3, std::pair<DataType, char> > &matrices, size_t dim1, size_t dim2, const std::map<Match, DataType> &match_points)
{
	std::cout << "AHA" << std::endl;
	matrices.resize(dim1+1, dim2+1);
	matrices.fill(std::pair<DataType, char>(0,'-'));
	typename std::map<Match, DataType>::const_iterator it, it_end=match_points.end();
	Matrix<std::pair<DataType, char> > &mat = matrices[0];
	for (it=match_points.begin(); it!=it_end; ++it)
	{

		mat[it->first.first+1][it->first.second+1].first=it->second;
	}
}
*/

/**
 * Fills the dynamic programming matrix with values from the Library.
 * @param matrices The dynamic programming matrix.
 * @param dim1 The size of the first dimension.
 * @param dim2 The size of the second dimension.
 * @param match_points The scores of the matches.
 */
template<typename DataType>
void
enterDataIntoMatrix(Matrix<std::pair<DataType, char> > &matrix, size_t dim1, size_t dim2, const std::map<Match, DataType> &match_points)
{
	matrix.resize(dim1+1, dim2+1);
	matrix.fill(std::pair<DataType, char>(0,'-'));
	typename std::map<Match, DataType>::const_iterator it, it_end=match_points.end();
	for (it=match_points.begin(); it!=it_end; ++it)
		matrix[it->first.first+1][it->first.second+1].first=it->second;
}



/**
 * \brief Produces a progressive alignment of a dataset.
 * \tparam The data to be used.
 * @param lib The library containing the pairwise matches.
 * @param guide_tree The guide tree to be used during the process.
 * @param set The data set to use.
 */
template<typename DataType, typename GapFunction>
void
progressive_consistency_align(const Library<DataType > &lib, Tree &guide_tree, DataType &set, GapFunction gap_func)
{
	size_t n_elems = set.size();
	std::stack<std::pair<TreeNode*, unsigned int> > to_do;
	to_do.push(std::pair<TreeNode*, int>(guide_tree.root(), 0));
	std::map<Match, int> match_points;

	Matrix<std::pair<int, char> > matrix(10,10);
	std::vector<std::pair<unsigned int, unsigned int> > gap1, gap2;

	Matrix<size_t> ids(n_elems, 1);
	for (size_t i=0; i<n_elems; ++i)
		ids[i][0]=i;

	size_t dim1, dim2, i;
	std::string edit_string1, edit_string2;
	while (!to_do.empty())
	{
		std::pair<TreeNode*, unsigned int> &current = to_do.top();
		if (current.first->children.empty())
			to_do.pop();
		else if (current.second == current.first->children.size())
		{
			std::vector<size_t> &ids1 = ids[current.first->children[0]->id];
			dim1=set[ids1[0]].length();
			std::vector<size_t> &ids2 = ids[current.first->children[1]->id];
			dim2=set[ids2[0]].length();
			lib.get(ids1, ids2, match_points, gap_func);
			enterDataIntoMatrix(matrix, dim1, dim2, match_points);
			nw_dyn_consistency(dim1, dim2, matrix);
			nw_dyn_consistency_traceback(dim1, dim2, matrix, edit_string1, edit_string2);
			for (i=0; i<ids1.size(); ++i)
				set[ids1[i]].insert_gaps(edit_string1);
			for (i=0; i<ids2.size(); ++i)
				set[ids2[i]].insert_gaps(edit_string2);
			ids1.reserve( ids1.size() + ids2.size() );
			ids1.insert(ids1.end(), ids2.begin(), ids2.end());
			current.first->id = current.first->children[0]->id;
			to_do.pop();
		}
		else
		{
			to_do.push(std::pair<TreeNode*, int>(&(*current.first->children[current.second]), 0));
			++current.second;
		}
	}
}




}


#endif /* CONSISTENCY_ALN_HPP_ */
