/*
 * domain_aln.cpp
 *
 *  Created on: Sep 20, 2013
 *      Author: Carsten Kemena
 */

#include "domain_aln.hpp"

using namespace std;

namespace MDAT
{

/**
 * \brief Calculates the score of a match of two Domains.
 *
 * This function has been taken from the RADS algorithm
 * @param opts The options containing the scores.
 * @param dm1 The first domain
 * @param dm2 The second domain
 * @return The score of the domain match
 */
int
get_score(const Opt_DA &opts, const Domain &dm1, const Domain &dm2)
{
	int score ;
	int l, lm, la, lb ;
	if(opts.matches[dm1.id()][dm2.id()]>=30)  //|| ((opts->matches[dm2->id]) && (opts->matches[dm1->id][dm2->id])) )
	{
		score=opts.matches[dm1.id()][dm2.id()];
		/* scaling the score according to domain length */
		if( opts.l_prop )
		{
			la = dm1.seq_length()/dm1.repeated() ;
			lb = dm2.seq_length()/dm2.repeated() ;
			if( la > lb )
			{
				l  = lb ;
				lm = la ;
			}
			else
			{
				l  = la ;
				lm = lb ;
			}

			if (1.0 * opts.scale * lm != 0)
				score =  score *  l * l / ( 1.0*opts.scale * lm ) ;
			if(score==0)
				score=1;
			/* The ratio score/scale is the points/bits to attribute to one expectied conserved position (if score=100 it is 1), it is then multipled by the length of the shortest domains, and finally scaled by the ratio of domain sizes */
		}
	}
	else
	{
		score = opts.mismatch ;
		/* scaling the score according to domain length */
		if( opts.l_prop )
		{
			la = dm1.seq_length()/dm1.repeated();
			lb = dm2.seq_length()/dm2.repeated();
			if( la > lb )
			{
				l  = lb ;
				lm = la ;
			}
			else
			{
				l  = la ;
				lm = lb ;
			}
			if (1.0 * opts.scale * lm != 0)
				score = score *  l * l / ( 1.0*opts.scale * lm ) ;

			if(score==0)
				score=-20;
		  /* The ratio score/scale is the points/bits to attribute to one expected conserved position (if score=100 it is 1), it is then multipled by the length of the shortest domains, and finally scaled by the ratio of domain sizes */
		}
	}
	return score ;
}


void
rads_align_archi (const Opt_DA &opts, const DomainArchitecture &doms1, const DomainArchitecture &doms2, MatrixStack<3,std::pair<int, char> > &matrices)
{
	// initialization

	Matrix<std::pair<int, char> > &mat_m = matrices[0];
	Matrix<std::pair<int, char> > &mat_h = matrices[1];
	Matrix<std::pair<int, char> > &mat_v= matrices[2];
	size_t n_doms1 = doms1.length()+1;
	size_t n_doms2 = doms2.length()+1;
	int score;
	// 0=match, 1=insert, 2=deletion
	int gapopen = opts.terminal_gop;
	int gapext = opts.terminal_gep;


	mat_m[0][0].first= 0;
	mat_h[0][0].first= gapopen;
	mat_v[0][0].first= gapext;
	size_t i,j;
	for (i=1; i<n_doms1; ++i)
	{
		mat_v[i][0].first=INT_MIN;
		mat_m[i][0].first=mat_h[i][0].first=mat_h[i-1][0].first+gapext*doms1[i-1].repeated();
	}
	for (j=1; j<n_doms2; ++j)
	{
		mat_m[0][j].first=mat_v[0][j].first=mat_v[0][j-1].first+gapext*doms2[j-1].repeated();
		mat_h[0][j].first=INT_MIN;
	}


	for (i=1; i<n_doms1; ++i)
	{
		for (j=1; j<n_doms2; ++j)
		{
			//calculate insert value
			if (j==(n_doms2-1))
			{
				gapopen = opts.terminal_gop;
				gapext  = opts.terminal_gep;
			} else
			{
				gapopen = opts.gop;
				gapext = opts.gep;
			}

			if (mat_v[i-1][j].first > mat_m[i-1][j].first +gapopen)
			{
				mat_v[i][j].second = 'i';
				mat_v[i][j].first = mat_v[i-1][j].first;
			}else
			{
				mat_v[i][j].second = 'm';
				mat_v[i][j].first = mat_m[i-1][j].first+gapopen;
			}
			mat_v[i][j].first += gapext*doms1[i-1].repeated();

			//calculate deletion value
			if (i==(n_doms1-1))
			{
				gapopen = opts.terminal_gop;
				gapext  = opts.terminal_gep;
			} else
			{
				gapopen = opts.gop;
				gapext = opts.gep;
			}
			if (mat_h[i][j-1].first > mat_m[i][j-1].first +gapopen)
			{
				mat_h[i][j].second = 'd';
				mat_h[i][j].first = mat_h[i][j-1].first;
			} else
			{
				mat_h[i][j].second = 'm';
				mat_h[i][j].first = mat_m[i][j-1].first+gapopen;
			}
			mat_h[i][j].first += gapext*doms2[j-1].repeated();

			//calculate match value
			if (mat_v[i][j].first > mat_h[i][j].first)
			{
				mat_m[i][j].second = 'i';
				mat_m[i][j].first = mat_v[i][j].first;
			}
			else
			{
				mat_m[i][j].second = 'd';
				mat_m[i][j].first = mat_h[i][j].first;
			}
			score = get_score(opts, doms1[i-1], doms2[j-1]) *std::min(doms1[i-1].repeated(), doms2[j-1].repeated())+abs(doms1[i-1].repeated()-doms2[j-1].repeated())*opts.gep;
			if (mat_m[i-1][j-1].first + score >= mat_m[i][j].first)
			{
				mat_m[i][j].second = 'm';
				mat_m[i][j].first = mat_m[i-1][j-1].first + score;
			}
		}
	}
}



void
rads_traceback_archi(Library<DomainArchitectureSet > &lib, const DomainArchitecture &arch1, const DomainArchitecture &arch2, const MatrixStack<3,std::pair<int, char> > &matrices, int aln_score)//const Opt_DA &opts)
{
	size_t i = arch1.length();
	size_t j = arch2.length();

	char state='c';
	int mat = 0;
	while ((i!=0) && (j!=0))
	{
		state = matrices[mat][i][j].second;
		if (mat==0)
		{
			if (state=='m')
			{
				--i;
				--j;
				lib.add(arch1.id(), arch2.id(), i, j, aln_score);//get_score(opts, arch1[i], arch2[j]));
			}
			else
			{
				if (state=='i')
					mat=2;
				else
					mat=1;
			}
		}
		else
		{
			if (mat==2)
				--i;
			else
				--j;
			if (state=='m')
				mat = 0;
		}
	}
}


void
rads_traceback_archi(size_t dim1, size_t dim2, const MatrixStack<3,std::pair<int, char> > &matrices, std::string &editString1, std::string &editString2)
{
	size_t i = dim1;
	size_t j = dim2;

	char state='c';
	int mat = 0;
	while ((i!=0) && (j!=0))
	{
		state = matrices[mat][i][j].second;
		if (mat==0)
		{
			if (state=='m')
			{
				--i;
				--j;
				editString1.push_back('m');
				editString2.push_back('m');
			}
			else
			{
				if (state=='i')
					mat=2;
				else
					mat=1;
			}
		}
		else
		{
			if (mat==2)
			{
				--i;
				editString1.push_back('m');
				editString2.push_back('-');
			}
			else
			{
				--j;
				editString1.push_back('-');
				editString2.push_back('m');
			}
			if (state=='m')
				mat = 0;
		}
	}
}



Matrix<float> *
all_rads_pair_align_archi(Library<DomainArchitectureSet > &lib, const ProteinSequenceSet<Default> &seqSet, const Opt_DA &opts)
{
	size_t i,j;
	size_t n_seqs = seqSet.n_architectures();
	int max_n_domains= seqSet.dom_archis().max_length()+1;
	MatrixStack<3,std::pair<int,char> > matrices(max_n_domains, max_n_domains);
	Matrix<float> *sim_mat = new Matrix<float>(n_seqs, n_seqs);
	for (i=0; i<n_seqs; ++i)
	{
		const DomainArchitecture &arch1=seqSet.dom_archis()[i];

		for (j=i+1; j<n_seqs; ++j)
		{
			const DomainArchitecture &arch2=seqSet.dom_archis()[j];
			rads_align_archi (opts, arch1, arch2, matrices);
			(*sim_mat)[i][j] = (*sim_mat)[j][i] =matrices[0][arch1.length()][arch2.length()].first;
			rads_traceback_archi(lib, arch1, arch2, matrices, (*sim_mat)[i][j]);// opts);
		}
	}
	return sim_mat;
}

void
run_single_rads(size_t i, Library<DomainArchitectureSet > &lib, const DomainArchitectureSet &archs, Matrix<float> *sim_mat, const Opt_DA &opts, MatrixStack<3,std::pair<int,char> > &matrices)
{
	const DomainArchitecture &arch1=archs[i];
	size_t n_archis = archs.size();
	size_t j;
	for (j=i+1; j<n_archis; ++j)
	{
		const DomainArchitecture &arch2=archs[j];
		rads_align_archi (opts, arch1, arch2, matrices);
		(*sim_mat)[i][j] = (*sim_mat)[j][i] =matrices[0][arch1.length()][arch2.length()].first;
		rads_traceback_archi(lib, arch1, arch2, matrices, (*sim_mat)[i][j]);
	}
}


Matrix<float> *
all_rads_pair_align_archi(Library<DomainArchitectureSet > &lib, const ProteinSequenceSet<Default> &seqSet, const Opt_DA &opts, unsigned int n_threads)
{
	if (n_threads <= 1)
		return all_rads_pair_align_archi(lib, seqSet, opts);

	size_t i;
	size_t n_archis = seqSet.n_architectures();
	int max_n_domains= seqSet.dom_archis().max_length()+1;
	Matrix<float> *sim_mat = new Matrix<float>(n_archis, n_archis);
	ThreadPool<MatrixStack<3,std::pair<int,char> > > pool(n_threads, MatrixStack<3,std::pair<int,char> >(max_n_domains, max_n_domains));
	for (i=0; i<n_archis; ++i)
		pool.addTask(run_single_rads,i, std::ref(lib), std::cref(seqSet.dom_archis()), sim_mat, std::ref(opts));
	pool.stop();
	return sim_mat;
}


int
moveCheck(DomainArchitectureSet &set, int domainId, int columnFrom, int columnTo)
{
	size_t nArchis = set.size();
	int movable=0;
	bool movableDomain;
	int j;
	for (size_t i=0; i<nArchis; ++i)
	{
		DomainArchitecture &arch = set[i];
		if (arch[columnFrom].id() == domainId)
		{
			movableDomain=true;
			if (columnFrom<columnTo)
			{
				for (j=columnFrom+1; j<=columnTo; ++j)
				{
					if (arch[j].id()!=-1)
					{
						movableDomain=false;
						break;
					}
				}
			}
			else
			{
				for (j=columnTo; j<columnFrom; ++j)
				{
					if (arch[j].id()!=-1)
					{
						movableDomain=false;
						break;
					}
				}
			}
			if (movableDomain)
				++movable;
		}
	}
	return movable;
}

float
columnEntropy(const map<int, int> &idCounter, int nonGap)
{
	float entropy=0;
	for (const auto &value : idCounter)
	{
		entropy += (value.second*log(1.0*value.second/nonGap));
	}
	return (-1)*entropy;
}




void
move(DomainArchitectureSet &set, int domainId, int columnFrom, int columnTo, vector<map<int, int> > &idCounter, vector<int> &nonGapCounter, vector<float> &entropy)
{
	size_t nArchis = set.size();
	bool movableDomain;
	int j;
	for (size_t i=0; i<nArchis; ++i)
	{
		DomainArchitecture &arch = set[i];
		if (arch[columnFrom].id() == domainId)
		{
			movableDomain=true;
			if (columnFrom<columnTo)
			{
				for (j=columnFrom+1; j<=columnTo; ++j)
				{
					if (arch[j].id()!=-1)
					{
						movableDomain=false;
						break;
					}
				}
			}
			else
			{
				for (j=columnTo; j<columnFrom; ++j)
				{
					if (arch[j].id()!=-1)
					{
						movableDomain=false;
						break;
					}
				}
			}
			if (movableDomain)
			{
				set[i][columnTo] = set[i][columnFrom];
				set[i][columnFrom].id(-1);
				++idCounter[columnTo][domainId];
				--nonGapCounter[columnFrom];
				++nonGapCounter[columnTo];
				--idCounter[columnFrom][domainId];
			}
		}
	}
	if (idCounter[columnFrom][domainId] == 0)
		idCounter[columnFrom].erase(domainId);
	entropy[columnFrom] = columnEntropy(idCounter[columnFrom], nonGapCounter[columnFrom]);
	entropy[columnTo] = columnEntropy(idCounter[columnTo], nonGapCounter[columnTo]);
}


void
move(DomainArchitectureSet &set, int domainId, int columnFrom, int columnTo, vector<map<int, int> > &idCounter, vector<int> &nonGapCounter)
{
	size_t nArchis = set.size();
	bool movableDomain;
	int j;
	for (size_t i=0; i<nArchis; ++i)
	{
		DomainArchitecture &arch = set[i];
		if (arch[columnFrom].id() == domainId)
		{
			movableDomain=true;
			if (columnFrom<columnTo)
			{
				for (j=columnFrom+1; j<=columnTo; ++j)
				{
					if (arch[j].id()!=-1)
					{
						movableDomain=false;
						break;
					}
				}
			}
			else
			{
				for (j=columnTo; j<columnFrom; ++j)
				{
					if (arch[j].id()!=-1)
					{
						movableDomain=false;
						break;
					}
				}
			}
			if (movableDomain)
			{
				set[i][columnTo] = set[i][columnFrom];
				set[i][columnFrom].id(-1);
				++idCounter[columnTo][domainId];
				--nonGapCounter[columnFrom];
				++nonGapCounter[columnTo];
				--idCounter[columnFrom][domainId];
			}
		}
	}
	if (idCounter[columnFrom][domainId] == 0)
		idCounter[columnFrom].erase(domainId);

}


void
refinePercentageColumn(DomainArchitectureSet &set, vector<map<int, int> > &idCounter, vector<int> &nonGapCounter, int columnId, bool left)
{
	size_t n_archis = set.size();
	int k;
	int id;
	int best_pos=-3;
	float best_val;
	float diffPercentageOld, diffPercentageNew, diffAbsoluteOld, diffAbsoluteNew;

	int length=set[0].size();
	vector<int> movable;
	int n_movable;
//	int nonGap;
	for (size_t i=0; i<n_archis; ++i)
	{
		DomainArchitecture &arch = set[i];
		if (arch[columnId].id()!=-1)
		{
			best_pos=k=columnId;
			id = arch[columnId].id();
			best_val = 0;
			if (left)
			{
				while ((--k>=0) && (arch[k].id()==-1))
				{
					if (idCounter[k].find(id)!=idCounter[k].end())
					{
						n_movable = moveCheck(set, id, columnId, k);
						diffPercentageOld =  abs(1.0*idCounter[k][id]/nonGapCounter[k]-1.0*idCounter[columnId][id]/nonGapCounter[columnId]);
						if (nonGapCounter[columnId]-n_movable != 0)
							diffPercentageNew = abs((1.0*idCounter[k][id]+n_movable)/(nonGapCounter[k]+n_movable)-(1.0*idCounter[columnId][id]-n_movable)/(nonGapCounter[columnId]-n_movable));
						else
							diffPercentageNew = (1.0*idCounter[k][id]+n_movable)/(nonGapCounter[k]+n_movable);
						diffAbsoluteOld = abs(idCounter[k][id]-idCounter[columnId][id]);
						diffAbsoluteNew = abs(idCounter[k][id]+n_movable-(idCounter[columnId][id]-n_movable));
						//nonGap = nonGapCounter[k]+n_movable;
						if ((diffPercentageNew-diffPercentageOld>=best_val) && (diffAbsoluteNew >= diffAbsoluteOld))
						{
							best_val=diffPercentageNew-diffPercentageOld;
							best_pos=k;
						}
					}
				}
			}
			//k=columnId;
			else
			{
				while ((++k<length) && (arch[k].id()==-1))
				{
					if (idCounter[k].find(id)!=idCounter[k].end())
					{
						n_movable = moveCheck(set, id, columnId, k);
						diffPercentageOld =  abs(1.0*idCounter[k][id]/nonGapCounter[k]-1.0*idCounter[columnId][id]/nonGapCounter[columnId]);
						if (nonGapCounter[columnId]-n_movable != 0)
							diffPercentageNew = abs((1.0*idCounter[k][id]+n_movable)/(nonGapCounter[k]+n_movable)-(1.0*idCounter[columnId][id]-n_movable)/(nonGapCounter[columnId]-n_movable));
						else
							diffPercentageNew = (1.0*idCounter[k][id]+n_movable)/(nonGapCounter[k]+n_movable);
						diffAbsoluteOld = abs(idCounter[k][id]-idCounter[columnId][id]);
						diffAbsoluteNew = abs(idCounter[k][id]+n_movable-(idCounter[columnId][id]+n_movable));
						//nonGap = nonGapCounter[k]+n_movable;

						if ((diffPercentageNew-diffPercentageOld>=best_val) && (diffAbsoluteNew >= diffAbsoluteOld))
						{
							best_val=diffPercentageNew-diffPercentageOld;
							best_pos=k;
						}
					}
				}
			}

			if (best_pos!=columnId)
				move(set, id, columnId, best_pos, idCounter, nonGapCounter);
		}
	}
}

void
refineMinimumShannonEntropyColumn(DomainArchitectureSet &set, vector<map<int, int> > &idCounter, vector<int> &nonGapCounter, vector<float> &entropy, int columnId, bool left)
{
	size_t n_archis = set.size();
	int k;
	int id;
	int best_pos=-3;
	float test_val;
	float best_val;
	int length=set[0].size();
	vector<int> movable;
	int n_movable;
	int nonGap;
	for (size_t i=0; i<n_archis; ++i)
	{
		DomainArchitecture &arch = set[i];
		if (arch[columnId].id()!=-1)
		{
			best_pos=k=columnId;
			id = arch[columnId].id();
			best_val = 0;
			if (left)
			{
				while ((--k>=0) && (arch[k].id()==-1))
				{
					if (idCounter[k].find(id)!=idCounter[k].end())
					{
						n_movable = moveCheck(set, id, columnId, k);
						test_val=0;
						nonGap = nonGapCounter[k]+n_movable;
						for (const auto &value : idCounter[k])
						{
							if (value.first==id)
								test_val += (1.0*(value.second+n_movable)*log(1.0*(value.second+n_movable)/nonGap));
							else
								test_val += (1.0*value.second*log(1.0*value.second/nonGap));
						}
						nonGap = nonGapCounter[columnId]-n_movable;
						for (const auto &value : idCounter[columnId])
						{
							if (value.first==id)
								test_val += (1.0*(value.second-n_movable)*log(1.0*(value.second-n_movable)/nonGap));
							else
								test_val += (1.0*value.second*log(1.0*value.second/nonGap));
						}
						test_val *= -1;
						if (test_val-entropy[k]-entropy[columnId]<best_val)
						{
							best_val=test_val;
							best_pos=k;
						}
					}
				}
			}
			else
			{
				while ((++k<length) && (arch[k].id()==-1))
				{
					if (idCounter[k].find(id)!=idCounter[k].end())
					{
						n_movable = moveCheck(set, id, columnId, k);
						test_val=0;
						nonGap = nonGapCounter[k]+n_movable;
						for (const auto &value : idCounter[k])
						{
							if (value.first==id)
								test_val += (1.0*(value.second+n_movable)*log(1.0*(value.second+n_movable)/nonGap));
							else
								test_val += (1.0*value.second*log(1.0*value.second/nonGap));
						}
						nonGap = nonGapCounter[columnId]-n_movable;
						for (const auto &value : idCounter[columnId])
						{
							if (value.first==id)
								test_val += (1.0*(value.second-n_movable)*log(1.0*(value.second-n_movable)/nonGap));
							else
								test_val += (1.0*value.second*log(1.0*value.second/nonGap));
						}
						test_val *= -1;
						if (test_val-entropy[k]-entropy[columnId]<best_val)
						{
							best_val=test_val;
							best_pos=k;
						}
					}
				}
			}

			if (best_pos!=columnId)
				move(set, id, columnId, best_pos, idCounter, nonGapCounter, entropy);
		}
	}
}


void
refineMinimumShannonEntropy(DomainArchitectureSet &set)
{
	int length = set.max_length();
	size_t n_architectures = set.size();
	vector<map<int, int> > idCounter(length);
	vector<int> nonGapCounter(length, 0);
	vector<float> entropy(length, 0);
	size_t i;
	int j;
	for (i=0; i<n_architectures; ++i)
	{
		DomainArchitecture &arch = set[i];
		for (j=0; j<length; ++j)
		{
			if (arch[j].id()!=-1)
			{
				++idCounter[j][arch[j].id()];
				++nonGapCounter[j];
			}
		}
	}

	for (j=0; j<length; ++j)
		entropy[j] = columnEntropy(idCounter[j], nonGapCounter[j]);

	for (j=0; j<length; ++j)
		refineMinimumShannonEntropyColumn(set, idCounter, nonGapCounter, entropy, j, true);
	int pos;
	for (i=0; i<n_architectures; ++i)
	{
		DomainArchitecture &arch = set[i];
		pos = -1;
		for (j=0; j<length; ++j)
		{
			if (nonGapCounter[j]!=0)
			{
				arch[++pos]=arch[j];
				idCounter[pos] = idCounter[j];
				nonGapCounter[pos] = nonGapCounter[j];
			}
		}
		arch.resize(++pos);
	}
	length = set.max_length();
	for (j=length-1; j>=0; --j)
		refineMinimumShannonEntropyColumn(set, idCounter, nonGapCounter, entropy, j, false);
	for (i=0; i<n_architectures; ++i)
	{
		DomainArchitecture &arch = set[i];
		pos = -1;
		for (j=0; j<length; ++j)
		{
			if (nonGapCounter[j]!=0)
			{
				arch[++pos]=arch[j];
				idCounter[pos] = idCounter[j];
				nonGapCounter[pos] = nonGapCounter[j];
			}
		}
		arch.resize(++pos);
	}
}


void
refinePercentage(DomainArchitectureSet &set)
{

	int length = set.max_length();
	size_t n_architectures = set.size();
	vector<map<int, int> > idCounter(length);
	vector<int> nonGapCounter(length, 0);
	size_t i;
	int j;
	for (i=0; i<n_architectures; ++i)
	{
		DomainArchitecture &arch = set[i];
		for (j=0; j<length; ++j)
		{
			if (arch[j].id()!=-1)
			{
				++idCounter[j][arch[j].id()];
				++nonGapCounter[j];
			}
		}
	}


	for (j=1; j<length; ++j)
		refinePercentageColumn(set, idCounter, nonGapCounter, j, true);
	int pos;
	for (i=0; i<n_architectures; ++i)
	{
		DomainArchitecture &arch = set[i];
		pos = -1;
		for (j=0; j<length; ++j)
		{
			if (nonGapCounter[j]!=0)
				arch[++pos]=arch[j];
		}
		arch.resize(++pos);
	}
	pos=-1;
	for (j=0; j<length; ++j)
	{
		if (nonGapCounter[j]!=0)
		{
			idCounter[++pos] = idCounter[j];
			nonGapCounter[pos] = nonGapCounter[j];
		}
	}
	length = set.max_length();
	for (j=length-2; j>=0; --j)
		refinePercentageColumn(set, idCounter, nonGapCounter, j, false);
	for (i=0; i<n_architectures; ++i)
	{
		DomainArchitecture &arch = set[i];
		pos = -1;
		for (j=0; j<length; ++j)
		{
			if (nonGapCounter[j]!=0)
				arch[++pos]=arch[j];
		}
		arch.resize(++pos);
	}
	pos=-1;
	for (j=0; j<length; ++j)
	{
		if (nonGapCounter[j]!=0)
		{
			idCounter[++pos] = idCounter[j];
			nonGapCounter[pos] = nonGapCounter[j];
		}
	}
	length = set.max_length();
	for (j=1; j<length; ++j)
		refinePercentageColumn(set, idCounter, nonGapCounter, j, true);
	for (i=0; i<n_architectures; ++i)
	{
		DomainArchitecture &arch = set[i];
		pos = -1;
		for (j=0; j<length; ++j)
		{
			if (nonGapCounter[j]!=0)
				arch[++pos]=arch[j];
		}
		arch.resize(++pos);
	}
	length = set.max_length();

}


/**
 * \brief Fills the Gotoh match matrix with average scores for the alignment of two columns.
 * @param set A set of sequences to be used.
 * @param ids1 The positions of the first set of sequences.
 * @param ids2 The positions of the second set of sequences.
 * @param matrixStack The matrix stack.
 * @param sim_mat The similarity matrix.
 */
/*
void
fillGotohMatrix(const DomainArchitectureSet &set, std::vector<size_t> ids1, std::vector<size_t> ids2, MatrixStack<3,std::pair<float, char> > &matrixStack, const Opt_DA &opts)
{
	size_t i,j;
	std::unordered_map<short, int>::iterator it;
	short c;
	size_t n_seqs1=ids1.size();
	size_t length1 = set[ids1[0]].size();
	std::vector<std::unordered_map<short, int> > prof1(length1);
	for (i=0; i<n_seqs1; ++i)
	{
		const DomainArchitecture &seq = set[ids1[i]];
		for (j=0; j<length1; ++j)
		{
			if (seq[j].id() != -1)
			{
				c=seq[j].id();
				if ((it =prof1[j].find(c)) != prof1[j].end())
					++it->second;
				else
					prof1[j][c]=1;
			}
		}
	}

	size_t length2 = set[ids2[0]].size();
	size_t n_seqs2=ids2.size();
	std::vector<std::unordered_map<short, int> > prof2(length2);

	for (i=0; i<n_seqs2; ++i)
	{
		const DomainArchitecture &seq = set[ids2[i]];
		for (j=0; j<length2; ++j)
		{
			if (seq[j].id() != -1)
			{
				c=seq[j].id();
				if ((it =prof2[j].find(c)) != prof2[j].end())
					++it->second;
				else
					prof2[j][c]=1;
			}
		}
	}

	matrixStack.ensure(length1+1, length2+1);
	Matrix<std::pair<float, char> > &matrix = matrixStack[0];
	std::unordered_map<short, int>::iterator it1,it2,it1_end,it2_end;

	double tmp;
	for (i=0; i<length1; ++i)
	{
		it1_end=prof1[i].end();
		for (j=0; j<length2; ++j)
		{

			matrix[i+1][j+1].first=0;
			it2_end=prof2[i].end();
			tmp=0;
			for (it1=prof1[i].begin(); it1!=it1_end; ++it1)
			{
//				const std::vector<int> &vec=sim_mat[it1->first];
				for (it2=prof2[j].begin(); it2!=it2_end; ++it2)
				{
					get_score(opts, const Domain &dm1, const Domain &dm2)
					matrix[i+1][j+1].first += vec[it2->first] * it1->second * it2->second;
					tmp += it1->second * it2->second;
				}
			}
			matrix[i+1][j+1].first /= tmp;
		}
	}
}*/


void
refine(DomainArchitectureSet &set)
{
	int length = set.max_length();
	size_t n_architectures = set.size();
	size_t i;
	int j;
	int k;
	int best_pos;
	float test_val;
	float best_val;

	vector<map<int, int> > idCounter(length);
	vector<int> nonGapCounter(length, 0);
	for (i=0; i<n_architectures; ++i)
	{
		DomainArchitecture &arch = set[i];
		for (j=0; j<length; ++j)
		{
			if (arch[j].id()!=-1)
			{
				idCounter[j][arch[j].id()] += 1;
				++nonGapCounter[j];
			}
		}
	}

	int id;
	for (i=0; i<n_architectures; ++i)
	{
		DomainArchitecture &arch = set[i];
		for (j=0; j<length; ++j)
		{
			best_pos=j;
			if (arch[j].id()!=-1)
			{
				k=j;
				id = arch[j].id();
				best_val = 1.0*idCounter[j][id]/nonGapCounter[j];
				while ((--k>=0) && (arch[k].id()==-1))
				{

					test_val=1.0 *idCounter[k][id]/nonGapCounter[k];
					if (test_val >= best_val)
					{
						best_val=test_val;
						best_pos=k;
					}
				}
				k=j;
				while ((++k<length) && (arch[k].id()==-1))
				{
					test_val=1.0*idCounter[k][id]/nonGapCounter[k];
					if (test_val >= best_val)
					{
						best_val=test_val;
						best_pos=k;
					}
				}
			}
			if (best_pos!=j)
			{
				arch[best_pos] = arch[j];
				arch[j].id(-1);
				++idCounter[best_pos][id];
				--idCounter[j][id];
				--nonGapCounter[j];
				++nonGapCounter[best_pos];
			}
		}
	}
	for (i=0; i<n_architectures; ++i)
	{
		DomainArchitecture &arch = set[i];
		for (j=length-1; j>=0; --j)
		{
			best_pos=j;
			if (arch[j].id()!=-1)
			{
				k=j;
				id = arch[j].id();
				best_val = 1.0*idCounter[j][id]/nonGapCounter[j];
				while ((--k>=0) && (arch[k].id()==-1))
				{

					test_val=1.0 *idCounter[k][id]/nonGapCounter[k];
					if (test_val >= best_val)
					{
						best_val=test_val;
						best_pos=k;
					}
				}
				k=j;
				while ((++k<length) && (arch[k].id()==-1))
				{
					test_val=1.0*idCounter[k][id]/nonGapCounter[k];
					if (test_val >= best_val)
					{
						best_val=test_val;
						best_pos=k;
					}
				}
			}
			if (best_pos!=j)
			{
				arch[best_pos] = arch[j];
				arch[j].id(-1);
				++idCounter[best_pos][id];
				--idCounter[j][id];
				--nonGapCounter[j];
				++nonGapCounter[best_pos];
			}
		}
	}

	// remove gap columns
	int pos;
	for (i=0; i<n_architectures; ++i)
	{
		DomainArchitecture &arch = set[i];
		pos = -1;
		for (j=0; j<length; ++j)
		{
			if (nonGapCounter[j]!=0)
				arch[++pos]=arch[j];
		}
		arch.resize(++pos);
	}
}



void
removeGapColumn(DomainArchitectureSet &set, const std::vector<bool> &use, size_t length)
{
	size_t nArchis=set.size();
	size_t j,i;
	size_t gapCount;
	int pos=-1;
	for (j=0; j<length; ++j)
	{
		gapCount=0;
		for (i=0; i<nArchis; ++i)
		{
			if (use[i])
			{
				if (set[i][j].id() == -1)
					++gapCount;
			}
		}
		if (gapCount!=use.size()-1)
		{
			++pos;
			for (i=0; i<nArchis; ++i)
				set[i][pos]=set[i][j];
		}
	}
	++pos;
	for (i=0; i<nArchis; ++i)
		set[i].resize(pos);
}


/*
void
realign(DomainArchitectureSet &set, const Opt_DA &opts)
{
	MatrixStack<3, std::pair<float, char> > matrices(1,1);
	size_t nArchis = set.size();
	size_t length = set[0].size();
	std::vector<bool> use(nArchis, true);
	for (size_t i=0; i<nArchis; ++i)
	{
		use[i] = false;
		set[i].removeGaps();
		removeGapColumn(set, use, length);

		use[i] = true;
	}
}
*/




}
