/*
 * domain_aln.hpp
 *
 *  Created on: Sep 11, 2013
 *      Author: ckeme_01
 */


/**
 * \file domain_aln.hpp
 * \brief Several functions related to domain alignment.
 *
 * Many function in this file are based on RADS. (http://rads.uni-muenster.de/)
 */


#ifndef DOMAIN_ALN_HPP_
#define DOMAIN_ALN_HPP_

//C++ header
#include <algorithm>
#include <map>
#include <stack>
#include <utility>
#include <vector>


//MDAT header
#include "Library.hpp"
#include "../Sequence/ProteinSequenceSet.hpp"
#include "../Annotation/DomainArchitecture.hpp"
#include "../utils/MatrixStack.hpp"
#include "../utils/ThreadPool.hpp"
#include "consistency_aln.hpp"



namespace MDAT
{



/**
 * \brief Structure to contain the MDA options.
 */
struct Opt_DA
{
	Opt_DA() : match(0), mismatch(0), matches(), terminal_gop(0), terminal_gep(0), gop(0), gep(0), l_prop(0), scale(0), dmpm_f("")
	{}
	int match;			/**< \brief  The match score of two domains*/
	int mismatch;		/**< \brief  The mismatch scores of two domains*/
	MDAT::Matrix<short> matches;	/**< \brief  Is unequal 0 if two domains are similar*/
	int terminal_gop;	/**< \brief  The terminal gap opening penalties*/
	int terminal_gep;	/**< \brief  The terminal gap extension penalties*/
	int gop;			/**< \brief  The gap opening penalties*/
	int gep;			/**< \brief  The gep extension penalties*/
	int l_prop;
	int scale;
	std::string dmpm_f;
};



/**
 * \brief Aligning two Architectures
 *
 * Function is basically taken from RADS. It is based on the Gotoh-Algorithm but uses scores specifically designed for domains.
 * @param opts The options (scores) to be used by the algorithm.
 * @param doms1 The first domain architecture
 * @param doms2 The second domain architecture
 * @param matrices The matrices for the Gotoh-Algorithm.
 */
void rads_align_archi (const Opt_DA &opts, const DomainArchitecture &doms1, const DomainArchitecture &doms2, MatrixStack<3,std::pair<int, char> > &matrices);


/**
 * \brief Traceback for the RADS dynamic programming.
 *
 * This function does the traceback of the Gotoh-Algorithm of the RADS alignment. The identified matches are stored in the library.
 * @param lib The Library object.
 * @param arch1 The first architecture.
 * @param arch2 The second architecture.
 * @param matrices The dynamic programming matrix
 *
 */
void
rads_traceback_archi(Library<DomainArchitectureSet > &lib, const DomainArchitecture &arch1, const DomainArchitecture &arch2, const MatrixStack<3,std::pair<int, char> > &matrices);


/**
 * \brief Produces all pairwise alignments
 *
 * This functions computes all alignments between every pair of sequences and saves the matches in library.
 * @param lib The Library object
 * @param seqSet The sequence set
 * @param opts The domain alignment options
 * @return A matrix of pairwise similarities
 */
Matrix<float> *
all_rads_pair_align_archi(Library<DomainArchitectureSet > &lib, const ProteinSequenceSet<Default> &seqSet, const Opt_DA &opts);



Matrix<float> *
all_rads_pair_align_archi(Library<DomainArchitectureSet > &lib, const ProteinSequenceSet<Default> &seqSet, const Opt_DA &opts, unsigned int n_threads);


/**
 * \brief Refinement of domain alignment by shifting domains to increase number of identical domains in a column.
 * @param set Domain Alignment.
 */
void
refine(DomainArchitectureSet &set);

void
refinePercentage(DomainArchitectureSet &set);

void
refineMinimumShannonEntropy(DomainArchitectureSet &set);

/*
void
realign(DomainArchitectureSet &set, const Opt_DA &opts);
*/
}



#endif /* DOMAIN_ALN_HPP_ */
