/*
 * seq_align.cpp
 *
 *  Created on: Feb 11, 2014
 *      Author: ckeme_01
 */

#include "seq_align.hpp"

namespace MDAT {

void
reverseGapped2ungapped(const std::string &seq, std::vector<int> &index)
{
	int seqL = seq.size();
	index.clear();
	index.resize(seqL);
	int pos = -1;
	for (int i=seqL-1; i>=0; --i)
	{
		if (seq[i] != '-')
			index[++pos] = seqL-1 - i;
	}
	index.resize(++pos);
}

/**
 * \brief Computes the dynamic programming algorithm for as described by gotoh but limited to a certain width around the diagonal.
 * @param dim1 The first dimension of the matrix.
 * @param dim2 The second dimension of the matrix.
 * @param matrices The dynamic programming matrix.
 * @param gop The gap opening costs.
 * @param gep The gap extension costs.
 * @param band_width The bandwidth to be used around.
 */
void
gotoh_align_banded(int dim1, int dim2, MatrixStack<3, std::pair<float, char> > &matrices, int gop, int gep, size_t band_width)
{
	int diff = abs(dim1-dim2);
	Matrix<std::pair<float, char> > &mat_m = matrices[0];
	Matrix<std::pair<float, char> > &mat_h = matrices[1];
	Matrix<std::pair<float, char> > &mat_v = matrices[2];
	++dim1;
	++dim2;
	int use_gop;
	int match_score;
	// 0=match, 1=insert, 2=deletion

	mat_m[0][0].first= 0;
	mat_h[0][0].first= 0;
	mat_v[0][0].first= gep;
	int i,j;

	for (i=1; i<dim1; ++i)
	{
		mat_h[i][0].first=-FLT_MAX+900;
		mat_m[i][0].first=mat_v[i][0].first=mat_v[i-1][0].first+gep;
		mat_h[i][0].second = 'm';
		mat_v[i][0].second = 'v';
		mat_m[i][0].second = 'v';
	}
	for (j=1; j<dim2; ++j)
	{
		mat_m[0][j].first=mat_h[0][j].first=mat_h[0][j-1].first+gep;
		mat_v[0][j].first=-FLT_MAX+900;
		mat_h[0][j].second = 'h';
		mat_v[0][j].second = 'm';
		mat_m[0][j].second = 'h';
	}


	int start, end;
	for (i=1; i<dim1; ++i)
	{
		end = std::min<int>(dim2, i+diff+band_width);
		if (end<dim2)
			mat_v[i][end].first=mat_h[i][end].first=mat_m[i][end].first =-FLT_MAX+900;
		start = std::max<int>(1,i-diff-band_width);
		if (start>1)
			mat_v[i][start-1].first=mat_h[i][start-1].first=mat_m[i][start-1].first =-FLT_MAX+900;
		for (j=start; j<end; ++j)
		{
			//calculate insert value
			use_gop = (j==(dim2-1))? 0 : gop;
			if (mat_v[i-1][j].first > (mat_m[i-1][j].first +use_gop))
			{
				mat_v[i][j].second = 'v';
				mat_v[i][j].first = mat_v[i-1][j].first;
			}else
			{
				mat_v[i][j].second = 'm';
				mat_v[i][j].first = mat_m[i-1][j].first+use_gop;
			}
			mat_v[i][j].first += gep;

			//calculate deletion value
			use_gop = (i==(dim1-1))? 0 : gop;
			if (mat_h[i][j-1].first > (mat_m[i][j-1].first +use_gop))
			{
				mat_h[i][j].second = 'h';
				mat_h[i][j].first = mat_h[i][j-1].first;
			} else
			{
				mat_h[i][j].second = 'm';
				mat_h[i][j].first = mat_m[i][j-1].first+use_gop;
			}
			mat_h[i][j].first += gep;

			//calculate match value
			match_score=mat_m[i][j].first;
			if (mat_v[i][j].first > mat_h[i][j].first)
			{
				mat_m[i][j].second = 'v';
				mat_m[i][j].first = mat_v[i][j].first;
			}
			else
			{
				mat_m[i][j].second = 'h';
				mat_m[i][j].first = mat_h[i][j].first;
			}

			if (mat_m[i-1][j-1].first + match_score >= mat_m[i][j].first)
			{
				mat_m[i][j].second = 'm';
				mat_m[i][j].first = mat_m[i-1][j-1].first + match_score;
			}
		}
	}

}


/**
 * \brief Computes the dynamic programming algorithm for as described by gotoh.
 * @param dim1 The first dimension of the matrix.
 * @param dim2 The second dimension of the matrix.
 * @param matrices The dynamic programming matrix.
 * @param gop The gap opening costs.
 * @param gep The gap extension costs.
 */
void
gotoh_align(int dim1, int dim2, MatrixStack<3, std::pair<float, char> > &matrices, int gop, int gep)
{
	Matrix<std::pair<float, char> > &mat_m = matrices[0];
	Matrix<std::pair<float, char> > &mat_h = matrices[1];
	Matrix<std::pair<float, char> > &mat_v = matrices[2];
	++dim1;
	++dim2;
	int use_gop;
	int match_score;
	// 0=match, 1=insert, 2=deletion

	mat_m[0][0].first= 0;
	mat_h[0][0].first= 0;
	mat_v[0][0].first= gep;
	int i,j;
	for (i=1; i<dim1; ++i)
	{
		mat_h[i][0].first=-FLT_MAX+900;
		mat_m[i][0].first=mat_v[i][0].first=mat_v[i-1][0].first+gep;
		mat_h[i][0].second = 'v';
		mat_v[i][0].second = 'v';
		mat_m[i][0].second = 'v';
	}
	for (j=1; j<dim2; ++j)
	{
		mat_m[0][j].first=mat_h[0][j].first=mat_h[0][j-1].first+gep;
		mat_v[0][j].first=-FLT_MAX+900;
		mat_h[0][j].second = 'h';
		mat_v[0][j].second = 'h';
		mat_m[0][j].second = 'h';
	}

	for (i=1; i<dim1; ++i)
	{
		for (j=1; j<dim2; ++j)
		{
			//calculate insert value
			use_gop = (j==(dim2-1))? 0 : gop;
			if (mat_v[i-1][j].first > (mat_m[i-1][j].first +use_gop))
			{
				mat_v[i][j].second = 'v';
				mat_v[i][j].first = mat_v[i-1][j].first;
			}else
			{
				mat_v[i][j].second = 'm';
				mat_v[i][j].first = mat_m[i-1][j].first+use_gop;
			}
			mat_v[i][j].first += gep;

			//calculate deletion value
			use_gop = (i==(dim1-1))? 0 : gop;
			if (mat_h[i][j-1].first > (mat_m[i][j-1].first +use_gop))
			{
				mat_h[i][j].second = 'h';
				mat_h[i][j].first = mat_h[i][j-1].first;
			} else
			{
				mat_h[i][j].second = 'm';
				mat_h[i][j].first = mat_m[i][j-1].first+use_gop;
			}
			mat_h[i][j].first += gep;

			//calculate match value
			match_score=mat_m[i][j].first;
			if (mat_v[i][j].first > mat_h[i][j].first)
			{
				mat_m[i][j].second = 'v';
				mat_m[i][j].first = mat_v[i][j].first;
			}
			else
			{
				mat_m[i][j].second = 'h';
				mat_m[i][j].first = mat_h[i][j].first;
			}

			if (mat_m[i-1][j-1].first + match_score >= mat_m[i][j].first)
			{
				mat_m[i][j].second = 'm';
				mat_m[i][j].first = mat_m[i-1][j-1].first + match_score;
			}
		}
	}
}


/**
 * \brief Computes the dynamic programming algorithm for as described by gotoh.
 * @param dim1 The first dimension of the matrix.
 * @param dim2 The second dimension of the matrix.
 * @param matrices The dynamic programming matrix.
 * @param gop The gap opening costs.
 * @param gep The gap extension costs.
 */
void
gotoh_alignExceptRegion(int dim1, int dim2, MatrixStack<3, std::pair<float, char> > &matrices, int gop, int gep, const std::vector<int> &align1, const std::vector<int> &align2)
{
	Matrix<std::pair<float, char> > &mat_m = matrices[0];
	Matrix<std::pair<float, char> > &mat_h = matrices[1];
	Matrix<std::pair<float, char> > &mat_v = matrices[2];
	++dim1;
	++dim2;
	int use_gop;
	int match_score;
	// 0=match, 1=insert, 2=deletion

	mat_m[0][0].first= 0;
	mat_h[0][0].first= 0;
	mat_v[0][0].first= gep;
	int i,j;
	for (i=1; i<dim1; ++i)
	{
		mat_h[i][0].first=-FLT_MAX +900;
		mat_m[i][0].first=mat_v[i][0].first=mat_v[i-1][0].first+gep;
		mat_h[i][0].second = 'v';
		mat_v[i][0].second = 'v';
		mat_m[i][0].second = 'v';
	}
	for (j=1; j<dim2; ++j)
	{
		mat_m[0][j].first=mat_h[0][j].first=mat_h[0][j-1].first+gep;
		mat_v[0][j].first=-FLT_MAX +900;
		mat_h[0][j].second = 'h';
		mat_v[0][j].second = 'h';
		mat_m[0][j].second = 'h';
	}

	for (i=1; i<dim1; ++i)
	{
		for (j=1; j<dim2; ++j)
		{
			// In the middle of a not to be aligned region --> skip calculation
			if ((align1[i]==2) && (align2[j]==2))
				continue;

			// border of a not aligned region --> allow gap passage on the left and lower border
			if ((align1[i]==1) && (align2[j]==1) && (align1[i-1]==2) && (align2[j-1]==2))
			{
				use_gop = (i==(dim1-1))? 0 : gop;
				if (mat_h[i][j-1].first > (mat_m[i][j-1].first + use_gop))
				{
					mat_h[i][j].second = 'h';
					mat_h[i][j].first = mat_h[i][j-1].first;
				}
				else
				{
					mat_h[i][j].second = 'm';
					mat_h[i][j].first = mat_m[i][j-1].first + use_gop;
				}
				mat_h[i][j].first += gep;
				use_gop = (j==(dim2-1))? 0 : gop;
				if (mat_v[i-1][j].first > (mat_m[i-1][j].first + use_gop))
				{
					mat_v[i][j].second = 'v';
					mat_v[i][j].first = mat_v[i-1][j].first;
				}
				else
				{
					mat_v[i][j].second = 'm';
					mat_v[i][j].first = mat_m[i-1][j].first + use_gop;
				}
				mat_v[i][j].first += gep;
				if (mat_v[i][j].first > mat_h[i][j].first)
				{
					mat_m[i][j].second = 'v';
					mat_m[i][j].first = mat_v[i][j].first;
				}
				else
				{
					mat_m[i][j].second = 'h';
					mat_m[i][j].first = mat_h[i][j].first;
				}
				continue;
			}
			if ((align2[j]!=0) && (align1[i]==1) && (align1[i-1]==2))
			{
				use_gop = (i==(dim1-1))? 0 : gop;
				if (mat_h[i][j-1].first > (mat_m[i][j-1].first + use_gop))
				{
					mat_h[i][j].second = 'h';
					mat_h[i][j].first = mat_h[i][j-1].first;
				}
				else
				{
					mat_h[i][j].second = 'm';
					mat_h[i][j].first = mat_m[i][j-1].first + use_gop;
				}
				mat_h[i][j].first += gep;
				mat_m[i][j].second = 'h';
				mat_m[i][j].first = mat_h[i][j].first;
				continue;
			}
			if ((align1[i]!=0) && (align2[j]==1) && (align2[j-1]==2))
			{
				use_gop = (j==(dim2-1))? 0 : gop;
				if (mat_v[i-1][j].first > (mat_m[i-1][j].first + use_gop))
				{
					mat_v[i][j].second = 'v';
					mat_v[i][j].first = mat_v[i-1][j].first;
				}
				else
				{
					mat_v[i][j].second = 'm';
					mat_v[i][j].first = mat_m[i-1][j].first + use_gop;
				}
				mat_v[i][j].first += gep;
				mat_m[i][j].second = 'v';
				mat_m[i][j].first = mat_v[i][j].first;
				continue;
			}
			if ((align1[i]!=0) && (align2[j]!=0))
				continue;

			// normal gotoh
			//calculate insert value
			use_gop = (j==(dim2-1))? 0 : gop;
			if (mat_v[i-1][j].first > (mat_m[i-1][j].first + use_gop))
			{
				mat_v[i][j].second = 'v';
				mat_v[i][j].first = mat_v[i-1][j].first;
			}
			else
			{
				mat_v[i][j].second = 'm';
				mat_v[i][j].first = mat_m[i-1][j].first + use_gop;
			}
			mat_v[i][j].first += gep;

			//calculate deletion value
			use_gop = (i==(dim1-1))? 0 : gop;
			if (mat_h[i][j-1].first > (mat_m[i][j-1].first + use_gop))
			{
				mat_h[i][j].second = 'h';
				mat_h[i][j].first = mat_h[i][j-1].first;
			}
			else
			{
				mat_h[i][j].second = 'm';
				mat_h[i][j].first = mat_m[i][j-1].first + use_gop;
			}
			mat_h[i][j].first += gep;

			//calculate match value
			match_score=mat_m[i][j].first;
			if (mat_v[i][j].first > mat_h[i][j].first)
			{
				mat_m[i][j].second = 'v';
				mat_m[i][j].first = mat_v[i][j].first;
			}
			else
			{
				mat_m[i][j].second = 'h';
				mat_m[i][j].first = mat_h[i][j].first;
			}

			if (mat_m[i-1][j-1].first + match_score >= mat_m[i][j].first)
			{
				mat_m[i][j].second = 'm';
				mat_m[i][j].first = mat_m[i-1][j-1].first + match_score;
			}
		}
	}
}


//struct Region
//{
//	int postition;
//	int start;
//	int end;
//};


void
determineRegionsToAlignWithFixedDomainOrder(std::vector<Region> regions1, std::vector<Region> regions2, Matrix<int> &mat, size_t dim1, size_t dim2)
{
	int nReg1 = regions1.size();
	int nReg2 = regions2.size();
	++dim1;
	++dim2;
	mat.resize(dim1, dim2);
	size_t i,j;
	mat.fill(3);

	//init first values
	size_t start1=1;
	size_t end1;
	size_t start2=1;
	size_t end2;
	//std::cout << "dim: " <<dim1 << " " << dim2 << std::endl;
	int currPos = -1;
	int maxPos = nReg1 + nReg2;
	int pos1 = -1, pos2 = -1;
	int change = 1;
	while (++currPos < maxPos)
	{
		if (pos1+1 < nReg1)
			end1 = regions1[pos1+1].start+1;
		else
			end1 = dim1;
		if (pos2+1 < nReg2)
			end2 = regions2[0].start+1;
		else
			end2 = dim2;
		if (((pos1+1) < nReg1) && (regions1[pos1+1].position == currPos))
		{
			end1 = regions1[++pos1].end+2;
			change=1;
		}
		else
		{
			end2 = regions2[++pos2].end+2;
			change = 2;
		}
		for (i=start1-1; i<end1; ++i)
		{
			if (mat[i][start2-1] == 3)
				mat[i][start2-1] = 1;
		}
	//	std::cout << "se " << start2-1 << " " << end2 << std::endl;
		for (j=start2-1; j<end2; ++j)
		{
			if (mat[start1-1][j] == 3)
				mat[start1-1][j] = 2;
		}
		for (i=start1; i<end1; ++i)
		{
			for (j=start2; j<end2; ++j)
				mat[i][j] = 0;
		}
		if (change==1)
			start1 = end1;
		else
			start2 = end2;
	}
	end1=dim1;
	end2=dim2;
	for (i=start1-1; i<end1; ++i)
	{
		if (mat[i][start2-1] == 3)
			mat[i][start2-1] = 1;
	}
	for (j=start2-1; j<end2; ++j)
	{
		if (mat[start1-1][j] == 3)
			mat[start1-1][j] = 2;
	}
	for (i=start1; i<end1; ++i)
	{
		for (j=start2; j<end2; ++j)
			mat[i][j] = 0;
	}

	/*
	for (i=0; i< dim1; ++i)
	{
		for (j=0; j<dim2; ++j)
			std::cout << " " << mat[i][j];
		std::cout << std::endl;
	}*/

}

void
gotoh_align(int dim1, int dim2, MatrixStack<3, std::pair<float, char> > &matrices, int gop, int gep, const Matrix<int> &alignMat)
{
	Matrix<std::pair<float, char> > &mat_m = matrices[0];
	Matrix<std::pair<float, char> > &mat_h = matrices[1];
	Matrix<std::pair<float, char> > &mat_v = matrices[2];
	++dim1;
	++dim2;
	int use_gop;
	int match_score;
	// 0=match, 1=insert, 2=deletion

	mat_m[0][0].first= 0;
	mat_h[0][0].first= 0;
	mat_v[0][0].first= gep;
	int i,j;
	for (i=1; i<dim1; ++i)
	{
		if ((alignMat[i][0]) == 3)
			break;
		mat_h[i][0].first=-FLT_MAX+900;
		mat_m[i][0].first=mat_v[i][0].first=mat_v[i-1][0].first+gep;
		mat_h[i][0].second = 'v';
		mat_v[i][0].second = 'v';
		mat_m[i][0].second = 'v';
	}
	for (j=1; j<dim2; ++j)
	{
		if ((alignMat[0][j]) == 3)
			break;
		mat_m[0][j].first=mat_h[0][j].first=mat_h[0][j-1].first+gep;
		mat_v[0][j].first=-FLT_MAX+900;
		mat_h[0][j].second = 'h';
		mat_v[0][j].second = 'h';
		mat_m[0][j].second = 'h';
	}

	int aM;
	int startPos=1;
	for (i=1; i<dim1; ++i)
	{
		for (j=startPos; j<dim2; ++j)
		{
			aM=alignMat[i][j];
			if (aM == 1)
				startPos=j;
			if (aM == 3)
				continue;

			//calculate insert value
			if (aM != 2)
			{
				use_gop = (j==(dim2-1))? 0 : gop;
				if (mat_v[i-1][j].first > (mat_m[i-1][j].first +use_gop))
				{
					mat_v[i][j].second = 'v';
					mat_v[i][j].first = mat_v[i-1][j].first;
				}else
				{
					mat_v[i][j].second = 'm';
					mat_v[i][j].first = mat_m[i-1][j].first+use_gop;
				}
				mat_v[i][j].first += gep;

				if (aM == 1)
				{
					mat_m[i][j].second = 'v';
					mat_m[i][j].first = mat_v[i][j].first;
					mat_h[i][j].first=-FLT_MAX+900;
					continue;
				}
			}

			//calculate deletion value
			if (aM != 1)
			{
				use_gop = (i==(dim1-1))? 0 : gop;
				if (mat_h[i][j-1].first > (mat_m[i][j-1].first +use_gop))
				{
					mat_h[i][j].second = 'h';
					mat_h[i][j].first = mat_h[i][j-1].first;
				} else
				{
					mat_h[i][j].second = 'm';
					mat_h[i][j].first = mat_m[i][j-1].first+use_gop;
				}
				mat_h[i][j].first += gep;
				if (aM == 2)
				{
					mat_m[i][j].second = 'h';
					mat_m[i][j].first = mat_h[i][j].first;
					mat_v[i][j].first=-FLT_MAX+900;
					continue;
				}
			}

			//calculate match value
			match_score=mat_m[i][j].first;
			if (mat_v[i][j].first > mat_h[i][j].first)
			{
				mat_m[i][j].second = 'v';
				mat_m[i][j].first = mat_v[i][j].first;
			}
			else
			{
				mat_m[i][j].second = 'h';
				mat_m[i][j].first = mat_h[i][j].first;
			}

			if (mat_m[i-1][j-1].first + match_score >= mat_m[i][j].first)
			{
				mat_m[i][j].second = 'm';
				mat_m[i][j].first = mat_m[i-1][j-1].first + match_score;
			}
		}
	}
}


} /* namespace MDAT */



