/*
 * seq_align.hpp
 *
 *  Created on: Sep 17, 2013
 *      Author: ckeme_01
 */

#ifndef SEQ_ALIGN_HPP_
#define SEQ_ALIGN_HPP_

#include <algorithm>
#include <cmath>
#include <limits>
#include <unordered_map>
#include <map>

//#include "Library.hpp"
//#include "fw_bw.hpp"
#include "../clustering/Tree.hpp"
#include "../clustering/Vector.hpp"
#include "../utils/MatrixStack.hpp"
#include "../Sequence/SplitSet.hpp"
#include "../utils/ThreadPool.hpp"

#include <condition_variable>
#include <thread>
#include <mutex>

using std::cout;
using std::endl;

namespace MDAT
{

struct Region
{
	Region(int pos, int s, int e) :
			position(pos), start(s), end(e)
	{}

	int position;
	int start;
	int end;
};

void
reverseGapped2ungapped(const std::string &seq, std::vector<int> &index);

template<typename MemoryType>
void splitBack(const std::vector<Region> &reg1, const std::vector<Region> &reg2,
		SplitSet<ProteinSequenceSet<MemoryType> > &set, int colID,
		const std::string &editString1, const std::string &editString2)
{
	std::vector<int> index1, index2;
	reverseGapped2ungapped(editString1, index1);
	reverseGapped2ungapped(editString2, index2);

	int pos1 = -1, pos2 = -1;
	int val1, val2;
	int col;
	int start = 0;
	int end = -1;
	size_t nSeqs = set[colID].size();
	int nReg1 = reg1.size();
	int nReg2 = reg2.size();
	size_t i;
	while (((pos1 + 1) < nReg1) || ((pos2 + 1) < nReg2))
	{
		if ((pos1 + 1) < nReg1)
			val1 = reg1[pos1 + 1].position;
		else
			val1 = INT_MAX;
		if ((pos2 + 1) < nReg2)
			val2 = reg2[pos2 + 1].position;
		else
			val2 = INT_MAX;
		if (val1 < val2)
		{
			col = reg1[++pos1].position;
			end = index1[reg1[pos1].start];
			set[col - 1].clear();
			for (i = 0; i < nSeqs; ++i)
				set[col - 1].add_seq(
						substr_no_renaming2(set[colID][i], start, end - start));
			start = end;
			end = index1[reg1[pos1].end];
			set[col].clear();
			for (i = 0; i < nSeqs; ++i)
				set[col].add_seq(substr_no_renaming2(set[colID][i], start, end + 1 - start));
		}
		else
		{
			col = reg2[++pos2].position;
			end = index2[reg2[pos2].start];
			set[col - 1].clear();
			for (i = 0; i < nSeqs; ++i)
				set[col - 1].add_seq(substr_no_renaming2(set[colID][i], start, end - start));
			start = end;
			end = index2[reg2[pos2].end];

			set[col].clear();
			for (i = 0; i < nSeqs; ++i)
				set[col].add_seq(substr_no_renaming2(set[colID][i], start, end + 1 - start));
		}
		start = end + 1;
	}
	int pos;
	int j;
	int len = set[colID][0].size();
	start = end + 1;

	if (start != 0)
	{
		for (i = 0; i < nSeqs; ++i)
		{
			ProteinSequence &seq = set[colID][i];
			pos = -1;
			for (j = start; j < len; ++j)
				seq[++pos] = seq[j];
			seq.resize(++pos);
		}
	}

}

struct NoAlign
{
	size_t seq1Start;
	size_t seq1End;
	size_t seq2Start;
	size_t seq2End;
};

template<typename Data>
struct SequenceGapFunc
{
	bool operator()(const Data &data, int pos) const
	{
		return (data[pos] == '-');
	}
};

struct SequenceSetGapFunc
{
	bool operator()(const ProteinSequenceSet<Default> &set, int column) const
	{
		size_t n_seqs = set.size();
		for (size_t i = 0; i < n_seqs; ++i)
		{
			if (set[i][column] != '-')
				return false;
		}
		return true;
	}
};

void
determineRegionsToAlignWithFixedDomainOrder(std::vector<Region> regions1,
		std::vector<Region> regions2, Matrix<int> &mat, size_t dim1,
		size_t dim2);

/**
 * \brief Computes the dynamic programming algorithm for as described by gotoh but limited to a certain width around the diagonal.
 * @param dim1 The first dimension of the matrix.
 * @param dim2 The second dimension of the matrix.
 * @param matrices The dynamic programming matrix.
 * @param gop The gap opening costs.
 * @param gep The gap extension costs.
 * @param band_width The bandwidth to be used around.
 */
void
gotoh_align_banded(int dim1, int dim2,
		MatrixStack<3, std::pair<float, char> > &matrices, int gop, int gep,
		size_t band_width);

/**
 * \brief Computes the dynamic programming algorithm for as described by gotoh.
 * @param dim1 The first dimension of the matrix.
 * @param dim2 The second dimension of the matrix.
 * @param matrices The dynamic programming matrix.
 * @param gop The gap opening costs.
 * @param gep The gap extension costs.
 */
void
gotoh_align(int dim1, int dim2,
		MatrixStack<3, std::pair<float, char> > &matrices, int gop, int gep);

void
gotoh_align(int dim1, int dim2,
		MatrixStack<3, std::pair<float, char> > &matrices, int gop, int gep,
		const Matrix<int> &alignMat);

/**
 * \brief Computes the dynamic programming algorithm for as described by gotoh.
 * @param dim1 The first dimension of the matrix.
 * @param dim2 The second dimension of the matrix.
 * @param matrices The dynamic programming matrix.
 * @param gop The gap opening costs.
 * @param gep The gap extension costs.
 */
void
gotoh_alignExceptRegion(int dim1, int dim2,
		MatrixStack<3, std::pair<float, char> > &matrices, int gop, int gep,
		const std::vector<int> &align1, const std::vector<int> &align2);


/**
 * \brief Performs the gotoh traceback.
 * \details The result is stored in the edit_strings in the form "mm-mm--". 'm' mean match, '-' a gap is introduced.
 * @param dim1 The size of the first dimension.
 * @param dim2 The size of the second dimension.
 * @param matrices The dynamic programming matrix.
 * @param edit_string1 The resulting edit string for dim1.
 * @param edit_string2 The resulting edit string for dim2.
 */
template<typename MatrixStackType>
void gotoh_traceback(int dim1, int dim2, const MatrixStackType &matrices,
		std::string &edit_string1, std::string &edit_string2)
{
	edit_string1.clear();
	edit_string2.clear();
	size_t i = dim1;
	size_t j = dim2;

	char state = 'c';
	int mat = 0;
	while ((i != 0) && (j != 0))
	{
		state = matrices[mat][i][j].second;
		if (mat == 0)
		{
			if (state == 'm')
			{
				--i;
				--j;
				edit_string1.push_back('m');
				edit_string2.push_back('m');
			}
			else
			{
				if (state == 'v')
					mat = 2;
				else
					mat = 1;
			}
		}
		else
		{
			if (mat == 2)
			{
				--i;
				edit_string1.push_back('m');
				edit_string2.push_back('-');
			}
			else
			{
				--j;
				edit_string1.push_back('-');
				edit_string2.push_back('m');
			}
			if (state == 'm')
				mat = 0;
		}
	}

	while (j > 0)
	{
		--j;
		edit_string1.push_back('-');
		edit_string2.push_back('m');
	}
	while (i > 0)
	{
		--i;
		edit_string1.push_back('m');
		edit_string2.push_back('-');
	}
}

/**
 * \brief Fills the Gotoh match matrix with average scores for the alignment of two columns.
 * @param set A set of sequences to be used.
 * @param ids1 The positions of the first set of sequences.
 * @param ids2 The positions of the second set of sequences.
 * @param matrixStack The matrix stack.
 * @param sim_mat The similarity matrix.
 */
template<typename DataType>
void
fillGotohMatrix(const DataType &set, std::vector<size_t> ids1,
				std::vector<size_t> ids2,
				MatrixStack<3, std::pair<float, char> > &matrixStack,
				const Matrix<int> &sim_mat)
{
	size_t i, j;
	std::unordered_map<short, int>::iterator it;
	short c;
	size_t n_seqs1 = ids1.size();
	size_t length1 = set[ids1[0]].size();
	std::vector<std::unordered_map<short, int> > prof1(length1);
	for (i = 0; i < n_seqs1; ++i)
	{
		const typename DataType::value_type &seq = set[ids1[i]];
		for (j = 0; j < length1; ++j)
		{
			if (seq[j] != '-')
			{
				c = toupper(seq[j]) - 65;
				if ((it = prof1[j].find(c)) != prof1[j].end())
					++it->second;
				else
					prof1[j][c] = 1;
			}
		}
	}

	size_t length2 = set[ids2[0]].size();
	size_t n_seqs2 = ids2.size();
	std::vector<std::unordered_map<short, int> > prof2(length2);

	for (i = 0; i < n_seqs2; ++i)
	{
		const typename DataType::value_type &seq = set[ids2[i]];
		for (j = 0; j < length2; ++j)
		{
			if (seq[j] != '-')
			{
				c = toupper(seq[j]) - 65;
				if ((it = prof2[j].find(c)) != prof2[j].end())
					++it->second;
				else
					prof2[j][c] = 1;
			}
		}
	}

	matrixStack.ensure(length1 + 1, length2 + 1);
	Matrix<std::pair<float, char> > &matrix = matrixStack[0];
	std::unordered_map<short, int>::iterator it1, it2, it1_end, it2_end;

	double tmp;
	for (i = 0; i < length1; ++i)
	{
		it1_end = prof1[i].end();
		for (j = 0; j < length2; ++j)
		{

			matrix[i + 1][j + 1].first = 0;
			it2_end = prof2[j].end();
			tmp = 0;
			for (it1 = prof1[i].begin(); it1 != it1_end; ++it1)
			{
				const std::vector<int> &vec = sim_mat[it1->first];
				for (it2 = prof2[j].begin(); it2 != it2_end; ++it2)
				{
					matrix[i + 1][j + 1].first += vec[it2->first] * it1->second * it2->second;
					tmp += it1->second * it2->second;
				}
			}
			matrix[i + 1][j + 1].first /= tmp;
		}
	}
}

/**
 * \brief Fills the Gotoh match matrix with average scores for the alignment of two columns.
 * @param set A set of sequences to be used.
 * @param ids1 The positions of the first set of sequences.
 * @param ids2 The positions of the second set of sequences.
 * @param matrixStack The matrix stack.
 * @param sim_mat The similarity matrix.
 */
template<typename DataType>
void fillGotohMatrix(const DataType &set, std::vector<size_t> ids1,
		std::vector<size_t> ids2,
		MatrixStack<3, std::pair<float, char> > &matrixStack,
		const Matrix<int> &sim_mat,
		const Matrix<int> &alignMat)
{
	size_t i, j;
	std::unordered_map<short, int>::iterator it;
	short c;
	size_t n_seqs1 = ids1.size();
	size_t length1 = set[ids1[0]].size();
	std::vector<std::unordered_map<short, int> > prof1(length1);
	for (i = 0; i < n_seqs1; ++i)
	{
		const typename DataType::value_type &seq = set[ids1[i]];
		for (j = 0; j < length1; ++j)
		{
			if (seq[j] != '-')
			{
				c = toupper(seq[j]) - 65;
				if ((it = prof1[j].find(c)) != prof1[j].end())
					++it->second;
				else
					prof1[j][c] = 1;
			}
		}
	}

	size_t length2 = set[ids2[0]].size();
	size_t n_seqs2 = ids2.size();
	std::vector<std::unordered_map<short, int> > prof2(length2);

	for (i = 0; i < n_seqs2; ++i)
	{
		const typename DataType::value_type &seq = set[ids2[i]];
		for (j = 0; j < length2; ++j)
		{
			if (seq[j] != '-')
			{
				c = toupper(seq[j]) - 65;
				if ((it = prof2[j].find(c)) != prof2[j].end())
					++it->second;
				else
					prof2[j][c] = 1;
			}
		}
	}

	matrixStack.resize(length1 + 1, length2 + 1);
	matrixStack[0].fill(std::pair<float,char>(0,'x'));
	Matrix<std::pair<float, char> > &matrix = matrixStack[0];
	std::unordered_map<short, int>::iterator it1, it2, it1_end, it2_end;

	double tmp;
	for (i = 0; i < length1; ++i)
	{
		it1_end = prof1[i].end();
		for (j = 0; j < length2; ++j)
		{
			if (alignMat[i + 1][j + 1] != 0)
				continue;

			matrix[i + 1][j + 1].first = 0;
			it2_end = prof2[j].end();
			tmp = 0;
			for (it1 = prof1[i].begin(); it1 != it1_end; ++it1)
			{
				const std::vector<int> &vec = sim_mat[it1->first];
				for (it2 = prof2[j].begin(); it2 != it2_end; ++it2)
				{
					matrix[i + 1][j + 1].first += vec[it2->first] * it1->second
							* it2->second;
					tmp += it1->second * it2->second;
				}
			}
			matrix[i + 1][j + 1].first /= tmp;
		}
	}
}

/**
 * \brief Fills the Gotoh match matrix with average scores for the alignment of two columns.
 * @param set A set of sequences to be used.
 * @param ids1 The positions of the first set of sequences.
 * @param ids2 The positions of the second set of sequences.
 * @param matrixStack The matrix stack.
 * @param sim_mat The similarity matrix.
 * @param band_width The bandwith to be used.
 */
template<typename DataType>
void fillGotohMatrix_banded(const DataType &set, std::vector<size_t> ids1,
		std::vector<size_t> ids2,
		MatrixStack<3, std::pair<float, char> > &matrixStack,
		const Matrix<int> &sim_mat, int band_width)
{
	struct helper
	{
		std::vector<std::pair<int, short> > counter;
		std::vector<int> char2counter;
		int pos;
		helper() :
				counter(), char2counter(26, 0), pos(-1)
		{
		}
	};
	size_t i, j;
	int pos2;
	short c;

	size_t n_seqs1 = ids1.size();
	size_t length1 = set[ids1[0]].size();
	std::vector<helper> prof1(length1);
	for (i = 0; i < n_seqs1; ++i)
	{
		const typename DataType::value_type &seq = set[ids1[i]];
		for (j = 0; j < length1; ++j)
		{
			if (seq[j] != '-')
			{
				c = toupper(seq[j]) - 65;
				if ((pos2 = prof1[j].char2counter[c]) != 0)
					++prof1[j].counter[pos2].first;
				else
				{
					prof1[j].char2counter[c] = ++prof1[j].pos;
					prof1[j].counter.emplace_back(std::pair<int, short>(1, c));
				}
			}
		}
	}

	size_t length2 = set[ids2[0]].size();
	size_t n_seqs2 = ids2.size();
	std::vector<helper> prof2(length2);
	for (i = 0; i < n_seqs2; ++i)
	{
		const typename DataType::value_type &seq = set[ids2[i]];
		for (j = 0; j < length2; ++j)
		{
			if (seq[j] != '-')
			{
				c = toupper(seq[j]) - 65;
				if ((pos2 = prof2[j].char2counter[c]) != 0)
					++prof2[j].counter[pos2].first;
				else
				{
					prof2[j].char2counter[c] = ++prof2[j].pos;
					prof2[j].counter.emplace_back(std::pair<int, short>(1, c));
				}
			}
		}
	}

	matrixStack.ensure(length1 + 50, length2 + 50);
	Matrix<std::pair<float, char> > &matrix = matrixStack[0];
	std::vector<std::pair<int, short> >::const_iterator it1, it2, it1_end,
			it2_end;

	double tmp;
	size_t start, end;
	size_t diff =
			(length1 > length2) ? (length1 - length2) : (length2 - length1);
	for (i = 0; i < length1; ++i)
	{
		it1_end = prof1[i].counter.cend();
		end = std::min<int>(length2, i + diff + band_width);
		start = std::max<int>(0, i - diff - band_width);
		for (j = start; j < end; ++j)
		{
			matrix[i + 1][j + 1].first = 0;
			it2_end = prof2[j].counter.cend();
			tmp = 0;
			for (it1 = prof1[i].counter.cbegin(); it1 != it1_end; ++it1)
			{
				const std::vector<int> &vec = sim_mat[it1->second];
				for (it2 = prof2[j].counter.cbegin(); it2 != it2_end; ++it2)
				{
					matrix[i + 1][j + 1].first += vec[it2->second] * it1->first
							* it2->first;
					tmp += it1->first * it2->first;
				}
			}
			matrix[i + 1][j + 1].first /= tmp;
		}
	}
}

/**
 * \brief Fills the Gotoh match matrix with average scores for the alignment of two columns.
 * @param set A set of sequences to be used.
 * @param ids1 The positions of the first set of sequences.
 * @param ids2 The positions of the second set of sequences.
 * @param matrixStack The matrix stack.
 * @param sim_mat The similarity matrix.
 * @param band_width The bandwith to be used.
 */
template<typename DataType>
void fillGotohMatrix_banded(const DataType &set, std::vector<size_t> ids1,
		std::vector<size_t> ids2,
		MatrixStack<3, std::pair<float, char> > &matrixStack,
		const Matrix<int> &sim_mat, int band_width,
		std::vector<NoAlign> &doNotAlign)
{
	struct helper
	{
		std::vector<std::pair<int, short> > counter;
		std::vector<int> char2counter;
		int pos;
		helper() :
				counter(), char2counter(26, 0), pos(-1)
		{
		}
	};

	std::map<std::pair<int, int>, bool> ignorePairs;
	for (size_t i = 0; i < doNotAlign.size(); ++i)
	{
		for (size_t j = doNotAlign[i].seq1Start; j <= doNotAlign[i].seq1End;
				++j)
		{
			for (size_t k = doNotAlign[i].seq2Start; k <= doNotAlign[i].seq2End;
					++k)
				ignorePairs[std::make_pair(j, k)] = true;
		}
	}

	size_t i, j;
	int pos2;
	short c;

	size_t n_seqs1 = ids1.size();
	size_t length1 = set[ids1[0]].size();
	std::vector<helper> prof1(length1);
	for (i = 0; i < n_seqs1; ++i)
	{
		const typename DataType::value_type &seq = set[ids1[i]];
		for (j = 0; j < length1; ++j)
		{
			if (seq[j] != '-')
			{
				c = toupper(seq[j]) - 65;
				if ((pos2 = prof1[j].char2counter[c]) != 0)
					++prof1[j].counter[pos2].first;
				else
				{
					prof1[j].char2counter[c] = ++prof1[j].pos;
					prof1[j].counter.emplace_back(std::pair<int, short>(1, c));
				}
			}
		}
	}

	size_t length2 = set[ids2[0]].size();
	size_t n_seqs2 = ids2.size();
	std::vector<helper> prof2(length2);
	for (i = 0; i < n_seqs2; ++i)
	{
		const typename DataType::value_type &seq = set[ids2[i]];
		for (j = 0; j < length2; ++j)
		{
			if (seq[j] != '-')
			{
				c = toupper(seq[j]) - 65;
				if ((pos2 = prof2[j].char2counter[c]) != 0)
					++prof2[j].counter[pos2].first;
				else
				{
					prof2[j].char2counter[c] = ++prof2[j].pos;
					prof2[j].counter.emplace_back(std::pair<int, short>(1, c));
				}
			}
		}
	}

	matrixStack.ensure(length1 + 50, length2 + 50);
	Matrix<std::pair<float, char> > &matrix = matrixStack[0];
	std::vector<std::pair<int, short> >::const_iterator it1, it2, it1_end,
			it2_end;

	double tmp;
	size_t start, end;
	size_t diff =
			(length1 > length2) ? (length1 - length2) : (length2 - length1);
	for (i = 0; i < length1; ++i)
	{
		it1_end = prof1[i].counter.cend();
		end = std::min<int>(length2, i + diff + band_width);
		start = std::max<int>(0, i - diff - band_width);
		for (j = start; j < end; ++j)
		{
			if (ignorePairs.find(std::make_pair(i, j)) != ignorePairs.end())
				continue;
			matrix[i + 1][j + 1].first = 0;
			it2_end = prof2[j].counter.cend();
			tmp = 0;
			for (it1 = prof1[i].counter.cbegin(); it1 != it1_end; ++it1)
			{
				const std::vector<int> &vec = sim_mat[it1->second];
				for (it2 = prof2[j].counter.cbegin(); it2 != it2_end; ++it2)
				{
					matrix[i + 1][j + 1].first += vec[it2->second] * it1->first
							* it2->first;
					tmp += it1->first * it2->first;
				}
			}
			matrix[i + 1][j + 1].first /= tmp;
		}
	}
}

/**
 * Calculates a progressive alignment.
 * @param set The sequence set.
 * @param guide_tree The guide tree to to be used.
 * @param sim_mat The similarity matrix.
 * @param gop The gap opening penalty.
 * @param gep The gap extension penalty.
 * \details Uses standard gotoh algorithm to produce alignments.
 */
template<typename DataType>
void seq_progressive_align(DataType &set, std::shared_ptr<Tree> guide_tree,
		const Matrix<int> &sim_mat, int gop, int gep, bool banded,
		MatrixStack<3, std::pair<float, char> > &dyn_matrix)
{
	size_t n_elems = set.size();
	std::stack<std::pair<const TreeNode*, unsigned int> > to_do;
	to_do.push(std::pair<const TreeNode*, int>(guide_tree->root(), 0));
	std::vector<std::pair<unsigned int, unsigned int> > gap1, gap2;

	Matrix<size_t> ids(n_elems, 1);
	for (size_t i = 0; i < n_elems; ++i)
		ids[i][0] = i;
	int *tree_helper = new int[guide_tree->n_species() * 2];
	for (size_t i = 0; i < guide_tree->n_species() * 2; ++i)
		tree_helper[i] = i;
	size_t dim1, dim2, i;
	std::string edit_string1, edit_string2;
	std::vector<float> pos_gops1, pos_gops2;
	std::vector<float> pos_geps1, pos_geps2;
	while (!to_do.empty())
	{
		std::pair<const TreeNode*, unsigned int> &current = to_do.top();
		if (current.first->children.empty())
			to_do.pop();
		else if (current.second == current.first->children.size())
		{
			pos_geps1.clear();
			pos_gops1.clear();
			pos_geps2.clear();
			pos_gops2.clear();
			std::vector<size_t> &ids1 =
					ids[tree_helper[current.first->children[0]->id]];
			dim1 = set[ids1[0]].length();
			std::vector<size_t> &ids2 =
					ids[tree_helper[current.first->children[1]->id]];
			dim2 = set[ids2[0]].length();

			if (banded)
			{
				fillGotohMatrix_banded(set, ids1, ids2, dyn_matrix, sim_mat, 15);
				gotoh_align_banded(dim1, dim2, dyn_matrix, gop, gep, 15);
			}
			else
			{
				fillGotohMatrix(set, ids1, ids2, dyn_matrix, sim_mat);
				gotoh_align(dim1, dim2, dyn_matrix, gop, gep);
			}
			gotoh_traceback(dim1, dim2, dyn_matrix, edit_string1, edit_string2);

			for (i = 0; i < ids1.size(); ++i)
				set[ids1[i]].insert_gaps(edit_string1);
			for (i = 0; i < ids2.size(); ++i)
				set[ids2[i]].insert_gaps(edit_string2);
			ids1.reserve(ids1.size() + ids2.size());
			ids1.insert(ids1.end(), ids2.begin(), ids2.end());
			tree_helper[current.first->id] =
					tree_helper[current.first->children[0]->id];
			to_do.pop();
		}
		else
		{
			to_do.push(
					std::pair<const TreeNode*, int>(
							&(*current.first->children[current.second]), 0));
			++current.second;
		}
	}
	delete[] tree_helper;
}

template<typename DataType>
void seq_progressive_align(DataType &set, const Tree &guide_tree,
		const Matrix<int> &sim_mat, int gop, int gep, bool banded)
{
	MatrixStack<3, std::pair<float, char> > dyn_matrix;
	seq_progressive_align(set, guide_tree, dyn_matrix, sim_mat, gop, gep,
			banded);
}

/**
 * \brief Performs a profile profile alignment.
 * @param set1 The first alignment.
 * @param set2 The second alignment.
 * @param sim_mat The similarity matrix.
 * @param gop The gap opening costs to use.
 * @param gep The gap extension costs to use.
 * @param matrix The dynamic programming matrix.
 */
template<typename DataType>
void seq_prof_prof_align(DataType &set1, DataType &set2,
		const Matrix<int> &sim_mat, int gop, int gep,
		MatrixStack<3, std::pair<float, char> > &matrix)
{
	size_t i;
	std::vector<size_t> ids1, ids2;
	size_t n_seqs1 = set1.size();
	for (i = 0; i < n_seqs1; ++i)
		ids1.push_back(i);
	size_t n_seqs2 = set2.size();
	for (i = 0; i < n_seqs2; ++i)
		ids2.push_back(n_seqs1 + i);

	std::string edit_string1, edit_string2;
	if ((n_seqs2 == 0) && (n_seqs1 == 0))
		return;

	if (n_seqs2 == 0)
	{
		size_t dim1 = set1[ids1[0]].length();
		for (i = 0; i < n_seqs2; ++i)
			set1[i + n_seqs1].append(dim1, '-');
		edit_string1.append(dim1, 'm');
		edit_string2.append(dim1, '-');
	}

	if (n_seqs1 == 0)
	{
		size_t dim2 = set1[ids2[0]].length();
		for (i = 0; i < n_seqs1; ++i)
			set1[i].append(dim2, '-');
		edit_string2.append(dim2, 'm');
		edit_string1.append(dim2, '-');
	}

	if ((n_seqs1 != 0) && (n_seqs2 != 0))
	{
		std::vector<std::pair<unsigned int, unsigned int> > gap1, gap2;
		std::vector<float> pos_gops1, pos_gops2;
		std::vector<float> pos_geps1, pos_geps2;
		set1.transfer(set2);
		size_t dim1 = set1[ids1[0]].length();
		size_t dim2 = set1[ids2[0]].length();
		fillGotohMatrix(set1, ids1, ids2, matrix, sim_mat);
		gotoh_align(dim1, dim2, matrix, gop, gep);
		gotoh_traceback(dim1, dim2, matrix, edit_string1, edit_string2);
	}
	for (i = 0; i < ids1.size(); ++i)
		set1[ids1[i]].insert_gaps(edit_string1);
	for (i = 0; i < ids2.size(); ++i)
		set1[ids2[i]].insert_gaps(edit_string2);
}

/**
 * \brief Performs a profile profile alignment.
 * @param set1 The first alignment.
 * @param set2 The second alignment.
 * @param sim_mat The similarity matrix.
 * @param gop The gap opening costs to use.
 * @param gep The gap extension costs to use.
 * @param matrix The dynamic programming matrix.
 * @param reg1 Regions of seq1 that should not be aligned with regions in reg2.
 * @param reg2 Regions of seq2 that should not be aligned with regions in reg1.
 */
template<typename DataType>
void seq_prof_prof_align(DataType &set1, DataType &set2, int colID,
		const Matrix<int> &sim_mat,
		int gop,
		int gep,
		std::vector<Region> reg1,
		std::vector<Region> reg2,
		MatrixStack<3, std::pair<float, char> > &matrix)
{

	std::vector<int> domPos1, domPos2;
	int val1, val2, num;
	domPos1.clear();
	domPos2.clear();
	int pos1 = -1;
	int pos2 = -1;
	num = -1;
	int nReg1 = reg1.size();
	int nReg2 = reg2.size();
	while (((pos1 + 1) < nReg1) || ((pos2 + 1) < nReg2))
	{
		if ((pos1 + 1) < nReg1)
			val1 = reg1[pos1 + 1].position;
		else
			val1 = INT_MAX;
		if ((pos2 + 1) < nReg2)
			val2 = reg2[pos2 + 1].position;
		else
			val2 = INT_MAX;
		if (val1 < val2)
		{
			++pos1;
			domPos1.push_back(reg1[pos1].position);
			reg1[pos1].position = ++num;
		}
		else
		{
			++pos2;
			domPos2.push_back(reg2[pos2].position);
			reg2[pos2].position = ++num;
		}
	}

	size_t i;
	std::vector<size_t> ids1, ids2;
	size_t n_seqs1 = set1[colID].size();
	for (i = 0; i < n_seqs1; ++i)
		ids1.push_back(i);
	size_t n_seqs2 = set2[colID].size();
	for (i = 0; i < n_seqs2; ++i)
		ids2.push_back(n_seqs1 + i);

	std::vector<std::pair<unsigned int, unsigned int> > gap1, gap2;
	std::string edit_string1, edit_string2;
	std::vector<float> pos_gops1, pos_gops2;
	std::vector<float> pos_geps1, pos_geps2;
	set1[colID].transfer(set2[colID]);

	if ((n_seqs2 == 0) && (n_seqs1 == 0))
		return;

	size_t dim1, dim2;
	if (n_seqs2 == 0)
	{
		size_t dim1 = set1[colID][ids1[0]].length();
		for (i = 0; i < n_seqs2; ++i)
			set1[colID][i + n_seqs1].append(dim1, '-');
		edit_string1.append(dim1, 'm');
		edit_string2.append(dim1, '-');
	}

	if (n_seqs1 == 0)
	{
		dim2 = set1[colID][ids2[0]].length();
		for (i = 0; i < n_seqs1; ++i)
			set1[colID][i].append(dim2, '-');
		edit_string2.append(dim2, 'm');
		edit_string1.append(dim2, '-');
	}

	if ((n_seqs2 != 0) && (n_seqs1 != 0))
	{
		dim1 = set1[colID][ids1[0]].length();
		dim2 = set1[colID][ids2[0]].length();
		Matrix<int> alignMat;
		determineRegionsToAlignWithFixedDomainOrder(reg1, reg2, alignMat, dim1,
				dim2);

		fillGotohMatrix(set1[colID], ids1, ids2, matrix, sim_mat, alignMat);
		gotoh_align(dim1, dim2, matrix, gop, gep, alignMat);
		gotoh_traceback(dim1, dim2, matrix, edit_string1, edit_string2);
	}

	for (i = 0; i < ids1.size(); ++i)
		set1[colID][ids1[i]].insert_gaps(edit_string1);
	for (i = 0; i < ids2.size(); ++i)
		set1[colID][ids2[i]].insert_gaps(edit_string2);

	for (int j = 0; j < nReg1; ++j)
		reg1[j].position = domPos1[j];
	for (int j = 0; j < nReg2; ++j)
		reg2[j].position = domPos2[j];
	splitBack(reg1, reg2, set1, colID, edit_string1, edit_string2);
}

/**
 * \brief Performs a profile profile alignment within a banded region.
 * @param set1 The first alignment.
 * @param set2 The second alignment.
 * @param sim_mat The similarity matrix.
 * @param gop The gap opening costs to use.
 * @param gep The gap extension costs to use.
 * @param matrix The dynamic programming matrix.
 */
template<typename DataType>
void seq_prof_prof_align_banded(DataType &set1,
		DataType &set2,
		const Matrix<int> &sim_mat,
		int gop,
		int gep,
		MatrixStack<3, std::pair<float, char> > &matrix)
{
	size_t i;
	std::vector<size_t> ids1, ids2;
	size_t n_seqs1 = set1.size();
	for (i = 0; i < n_seqs1; ++i)
		ids1.push_back(i);
	size_t n_seqs2 = set2.size();
	for (i = 0; i < n_seqs2; ++i)
		ids2.push_back(n_seqs1 + i);

	std::vector<std::pair<unsigned int, unsigned int> > gap1, gap2;

	std::string edit_string1, edit_string2;
	std::vector<float> pos_gops1, pos_gops2;
	std::vector<float> pos_geps1, pos_geps2;
	set1.transfer(set2);
	size_t dim1 = set1[ids1[0]].length();
	size_t dim2 = set1[ids2[0]].length();
	fillGotohMatrix_banded(set1, ids1, ids2, matrix, sim_mat, 15);
	gotoh_align_banded(dim1, dim2, matrix, gop, gep, 15);
	gotoh_traceback(dim1, dim2, matrix, edit_string1, edit_string2);
	for (i = 0; i < ids1.size(); ++i)
		set1[ids1[i]].insert_gaps(edit_string1);
	for (i = 0; i < ids2.size(); ++i)
		set1[ids2[i]].insert_gaps(edit_string2);
}

/**
 * \brief Calculates a progressive consistency alignment given a set of Sequences.
 * @param set The set of sequences
 */
template<typename DataType>
void seq_progressive_align(DataType &set,
		MatrixStack<3, std::pair<float, char> > &matrices,
		const Matrix<int> &sim_mat, int gop, int gep, bool banded)
{
	size_t n_elems = set.size();
	if (n_elems > 1)
	{
		Matrix<float> *dist_mat = kmer_dist_mat(set);
		Tree guide_tree;
		std::vector<std::string> names(n_elems, "");
		guide_tree.nj(*dist_mat, names);
		delete dist_mat;
		seq_progressive_align(set, matrices, guide_tree, sim_mat, gop, gep,
				banded);
	}
}

template<typename MemoryType>
void same_architecture_aln(ProteinSequenceSet<MemoryType> &set,
		const Matrix<int> &sim_mat, int gop, int gep,
		SplitSet<ProteinSequenceSet<Default> > &splitSet,
		MatrixStack<3, std::pair<float, char> > &matrix)
{
	// Split sequences into domain and interdomain parts and align each separately
	domain_column_split(set, splitSet);
	size_t n_seqs = set.n_seqs();
	// If only one sequence, than already aligned.
	if (n_seqs == 1)
		return;

	// Construct guide tree
	Matrix<float> *dist_mat = kmer_dist_mat(set);
	std::shared_ptr<Tree> guide_tree(new Tree());
	std::vector<std::string> names(set.size(), "");
	std::vector<int> n_members(set.size(), 1);
	guide_tree->upgma(*dist_mat, names, n_members);
	delete dist_mat;

	// Align each piece
	size_t n_pieces = splitSet.size();
	size_t i;
	bool banded = true;
	for (i = 0; i < n_pieces; ++i)
	{
		banded = !banded;
		if (splitSet[i].size() != 0)
			seq_progressive_align(splitSet[i], guide_tree, sim_mat, gop, gep, banded, matrix);
	}
}

/**
 * Refine the domain boundaries by extending into the envelope if score is improved
 * @param left The unaligned data set to the left of the domains
 * @param center The data set containing the aligned domain
 * @param right The unaligned data set to the right of the domains
 * @param i column in the MDA
 * @param all The data sets with the domains
 */
template<typename MemoryType>
void refineBounaries(ProteinSequenceSet<MemoryType> &left,
		ProteinSequenceSet<MemoryType> &center,
		ProteinSequenceSet<MemoryType> &right, size_t colID,
		ProteinSequenceSet<MemoryType> &all, const Matrix<int> &sim_mat)
{
	size_t nSeqs = center.size();
	size_t length = center[0].size();
	//calculate profile for each position
	std::vector<std::map<int, int> > profile(center[0].size(),
			std::map<char, int>());
	size_t i, j;
	for (i = 0; i < nSeqs; ++i)
	{
		for (j = 0; j < length; ++j)
		{
			if (center[i][j] != '-')
				profile[toUpper(center[i][j]) - 65] += 1;
		}
	}
	// Calculate score
	std::vector<double> score(length, 0);
	int pos, start, end, posOther;
	// check if aligned sequence can be extended
	for (i = 0; i < nSeqs; ++i)
	{
		const DomainArchitecture &domArch = all[center[i].id()].dom_archi();
		if ((start = domArch[colID].seq_start())
				> (end = domArch[colID].env_start()))
		{
			pos = 0;
			while (center[i][pos] == '-')
				++pos;
			posOther = center[i].size();
			for (--start; start >= end; --start)
			{
				--pos;
				if (pos < 0)
					break;
				--posOther;

			}

		}
	}

}

template<typename MemoryType>
void same_architecture_aln2(ProteinSequenceSet<MemoryType> &set,
		const Matrix<int> &sim_mat, int gop, int gep,
		SplitSet<ProteinSequenceSet<Default> > &splitSet,
		MatrixStack<3, std::pair<float, char> > &matrix)
{
	// Split sequences into domain and interdomain parts and align each separately
	domain_column_split(set, splitSet);
	size_t n_seqs = set.n_seqs();
	// If only one sequence, than already aligned.
	if (n_seqs == 1)
		return;

	// Construct guide tree
	Matrix<float> *dist_mat = kmer_dist_mat(set);
	std::shared_ptr<Tree> guide_tree(new Tree());
	std::vector<std::string> names(set.size(), "");
	std::vector<int> n_members(set.size(), 1);
	guide_tree->upgma(*dist_mat, names, n_members);
	delete dist_mat;

	// Align each piece
	size_t n_pieces = splitSet.size();
	size_t i;
	for (i = 1; i < n_pieces; i += 2)
	{
		if (splitSet[i].size() != 0)
			seq_progressive_align(splitSet[i], guide_tree, sim_mat, gop, gep, true, matrix);
		//refineBounaries(splitSet[i - 1], splitSet[i], splitSet[i + 1],(i - 1) / 2, set, sim_mat);
	}

	for (i = 0; i < n_pieces; i += 2)
	{
		if (splitSet[i].size() != 0)
			seq_progressive_align(splitSet[i], guide_tree, sim_mat, gop, gep, false, matrix);
	}

}

/**
 * \brief Alignment of sequences with the same domains.
 * @param set The sequence set.
 * @param sim_mat The similarity matrix.
 * @param gop The gap opening costs.
 * @param gep The gap extension penalty.
 * @param splitSet The split set.
 * @param pool The thread pool to use.
 */
template<typename MemoryType>
void same_architecture_aln(ProteinSequenceSet<MemoryType> &set,
		const Matrix<int> &sim_mat, int gop, int gep,
		SplitSet<ProteinSequenceSet<Default> > &splitSet,
		ThreadPool<MatrixStack<3, std::pair<float, char> > > &pool)
{
	// Split sequences into domain and interdomain parts and align each separately
	domain_column_split(set, splitSet);
	size_t n_seqs = set.n_seqs();

	// If only one sequence, than already aligned.
	if (n_seqs == 1)
		return;

	// Construct guide tree
	Matrix<float> *dist_mat = kmer_dist_mat(set);
	std::shared_ptr<Tree> guide_tree(new Tree());
	std::vector<std::string> names(set.size(), "");
	std::vector<int> n_members(set.size(), 1);
	guide_tree->upgma(*dist_mat, names, n_members);
	delete dist_mat;

	// Align each piece
	size_t n_pieces = splitSet.size();
	size_t i;
	bool banded = true;
	for (i = 0; i < n_pieces; ++i)
	{
		banded = !banded;
		if (splitSet[i].size() != 0)
			pool.addTask(
					static_cast<void (*)(ProteinSequenceSet<MemoryType> &,
							std::shared_ptr<Tree>, const Matrix<int> &, int,
							int, bool,
							MatrixStack<3, std::pair<float, char> > &)>(&seq_progressive_align), std::ref(splitSet[i]), guide_tree, std::ref(sim_mat), gop, gep, banded);
		}
	}

	/**
	 * Merges sequence sets into a single one.
	 * @param set The split set.
	 * @param start The first set to merge.
	 * @param end The last set to merge.
	 */
template<typename MemoryType>
void merge_sequences(SplitSet<ProteinSequenceSet<MemoryType> > &set,
		size_t start, size_t end)
{
	size_t i, j;
	size_t n_seqs = -1;
	for (j = start; j <= end; ++j)
	{
		if (set[j].size() > 0)
		{
			n_seqs = set[j].size();
			break;
		}
	}
	ProteinSequenceSet<MemoryType> tmp_set;
	//for (i=0; i<n_seqs; ++i)
	tmp_set.transfer(set[j]);

	for (j = j + 1; j <= end; ++j)
	{
		if (set[j].size() > 0)
		{
			for (i = 0; i < n_seqs; ++i)
				tmp_set[i].append(set[j][i].sequence());
			set[j].clear();
		}
	}
	set[end].transfer(tmp_set);
}

/**
 * Merges sequence sets into a single one.
 * @param set The split set.
 * @param start The first set to merge.
 * @param end The last set to merge.
 */
template<typename MemoryType>
void merge_sequences(SplitSet<ProteinSequenceSet<MemoryType> > &set,
		size_t start, size_t end, std::vector<Region> &domRegions)
{
	domRegions.clear();
	size_t i = -1, j;
	size_t n_seqs = -1;
	int regStart;

	for (j = start; j <= end; ++j)
	{
		if (set[j].size() > 0)
		{
			n_seqs = set[j].size();
			break;
		}
	}

	if (j > end) // no middle sequence
		return;

	ProteinSequenceSet<MemoryType> tmp_set;
	tmp_set.transfer(set[j]);

	// if a domain region add to list of region
	if ((j % 2) == 1)
		domRegions.push_back(Region(j, 0, tmp_set[0].size()));
	for (j = j + 1; j <= end; ++j)
	{
		if (set[j].size() > 0)
		{
			regStart = tmp_set[0].size();
			for (i = 0; i < n_seqs; ++i)
				tmp_set[i].append(set[j][i].sequence());
			// if a domain region add to list of region
			if ((j % 2) == 1)
				domRegions.push_back(
						Region(j, regStart, tmp_set[0].size() - 1));
			set[j].clear();
		}
	}
	set[end].transfer(tmp_set);
	set[end].transfer(tmp_set);
}

/**
 * \brief Sequence alignment of two different domains architectures.
 * @param set1 The first set.
 * @param set2 The second set.
 * @param sim_mat The similarity matrix.
 * @param gop The gap opening costs.
 * @param gep The gep extension costs.
 * @param pool The thread pool to use.
 */
template<typename MemoryType>
void aln_different_architectures(
		SplitSet<ProteinSequenceSet<MemoryType> > &set1,
		SplitSet<ProteinSequenceSet<MemoryType> > &set2,
		const Matrix<int> &sim_mat,
		int gop,
		int gep,
		ThreadPool<MatrixStack<3, std::pair<float, char> > > &pool)
{
	size_t n_pieces = set1.size();
	size_t i, pos = 0;
	std::vector<int> pattern1 = set1.pattern();
	std::vector<int> pattern2 = set2.pattern();
	std::vector<Region> reg1, reg2;
	MatrixStack<3, std::pair<float, char> > mat(10, 10);

	for (i = 0; i < n_pieces; ++i)
	{
		if ((pattern1[i] != 0) && (pattern2[i] != 0))
		{
			merge_sequences(set1, pos, i - 1, reg1);
			merge_sequences(set2, pos, i - 1, reg2);
			pool.addTask(static_cast<void (*)(SplitSet<ProteinSequenceSet<MemoryType> > &, SplitSet<ProteinSequenceSet<MemoryType> > &, int, const Matrix<int> &, int, int, std::vector<Region>, std::vector<Region>, MatrixStack<3, std::pair<float, char> > &)>(seq_prof_prof_align), std::ref(set1), std::ref(set2), i-1, std::cref(sim_mat), gop, gep, reg1, reg2);
			if (pattern1[i] == pattern2[i])
				pool.addTask(static_cast<void (*)(ProteinSequenceSet<MemoryType> &, ProteinSequenceSet<MemoryType> &, const Matrix<int> &, int, int, MatrixStack<3, std::pair<float, char> > &)>(seq_prof_prof_align_banded), std::ref(set1[i]), std::ref(set2[i]), std::cref(sim_mat), gop, gep);
			else
				pool.addTask(static_cast<void (*)(ProteinSequenceSet<MemoryType> &, ProteinSequenceSet<MemoryType> &, const Matrix<int> &, int, int, MatrixStack<3, std::pair<float, char> > &)>(seq_prof_prof_align), std::ref(set1[i]), std::ref(set2[i]), std::cref(sim_mat), gop, gep);
			pos = i + 1;
		}
		else
		{
			if (pattern1[i] == 0)
				pattern1[i] = pattern2[i];
			else
				pattern2[i] = pattern1[i];
		}
	}

	merge_sequences(set1, pos, i - 1, reg1);
	merge_sequences(set2, pos, i - 1, reg2);
	pool.addTask(static_cast<void (*)(SplitSet<ProteinSequenceSet<MemoryType> > &, SplitSet<ProteinSequenceSet<MemoryType> > &, int, const Matrix<int> &, int, int, std::vector<Region> , std::vector<Region> , MatrixStack<3, std::pair<float, char> > &)>(seq_prof_prof_align), std::ref(set1), std::ref(set2), i-1, std::cref(sim_mat), gop, gep, reg1, reg2);
	set1.pattern(pattern1);
	set2.pattern(pattern2);
}


template<typename MemoryType>
void aln_different_architecturesRelaxed(
		SplitSet<ProteinSequenceSet<MemoryType> > &set1,
		SplitSet<ProteinSequenceSet<MemoryType> > &set2,
		const Matrix<int> &sim_mat,
		int gop,
		int gep,
		ThreadPool<MatrixStack<3, std::pair<float, char> > > &pool)
{
	size_t n_pieces = set1.size();
	size_t i, pos = 0;
	std::vector<int> pattern1 = set1.pattern();
	std::vector<int> pattern2 = set2.pattern();
	MatrixStack<3, std::pair<float, char> > mat(10, 10);

	for (i = 0; i < n_pieces; ++i)
	{
		if ((pattern1[i] != 0) && (pattern2[i] != 0))
		{
			merge_sequences(set1, pos, i - 1);
			merge_sequences(set2, pos, i - 1);
			pool.addTask(static_cast<void (*)(ProteinSequenceSet<MemoryType> &, ProteinSequenceSet<MemoryType> &, const Matrix<int> &, int, int, MatrixStack<3, std::pair<float, char> > &)>(seq_prof_prof_align), std::ref(set1[i-1]), std::ref(set2[i-1]), std::cref(sim_mat), gop, gep);
			if (pattern1[i] == pattern2[i])
				pool.addTask(static_cast<void (*)(ProteinSequenceSet<MemoryType> &, ProteinSequenceSet<MemoryType> &, const Matrix<int> &, int, int, MatrixStack<3, std::pair<float, char> > &)>(seq_prof_prof_align_banded), std::ref(set1[i]), std::ref(set2[i]), std::cref(sim_mat), gop, gep);
			else
				pool.addTask(static_cast<void (*)(ProteinSequenceSet<MemoryType> &, ProteinSequenceSet<MemoryType> &, const Matrix<int> &, int, int, MatrixStack<3, std::pair<float, char> > &)>(seq_prof_prof_align), std::ref(set1[i]), std::ref(set2[i]), std::cref(sim_mat), gop, gep);
			pos = i + 1;
		}
		else
		{
			pattern1[i] = 0;
			pattern2[i] = 0;
		}
	}

	merge_sequences(set1, pos, i - 1);
	merge_sequences(set2, pos, i - 1);
	pool.addTask(static_cast<void (*)(ProteinSequenceSet<MemoryType> &, ProteinSequenceSet<MemoryType> &, const Matrix<int> &, int, int, MatrixStack<3, std::pair<float, char> > &)>(seq_prof_prof_align), std::ref(set1[i-1]), std::ref(set2[i-1]), std::cref(sim_mat), gop, gep);
	set1.pattern(pattern1);
	set2.pattern(pattern2);
}


/**
 * \brief Calculates a MSA using domain information
 * @param set The sequence set.
 * @param arch_guide_tree The guide tree to use.
 * @param sim_mat The similarity matrix to use.
 * @param gop The gap opening costs to use.
 * @param gep The gap extension costs to use.
 * @param n_threads The number of threads to use.
 */
template<typename MemoryType>
void msa(ProteinSequenceSet<MemoryType> &set, const Tree &arch_guide_tree,
		const Matrix<int> &sim_mat, int gop, int gep, bool relax, size_t n_threads)
{
	int last = 0;
	std::vector<SplitSet<ProteinSequenceSet<MemoryType> > > splitted_sets(
			set.n_architectures());
	MatrixStack<3, std::pair<float, char> > matrix(10, 10);

	if (set.n_architectures() == 1)
	{
		ThreadPool<MatrixStack<3, std::pair<float, char> > > pool(n_threads, matrix);
		same_architecture_aln(set, sim_mat, gop, gep, splitted_sets[0], pool);
		pool.stop();
	}
	else
	{
		// Align sequences with same architecture
		std::vector<ProteinSequenceSet<MemoryType> > architectureSplits;// = new std::vector<ProteinSequenceSet<MemoryType> >();
		splitByArchitecture(set, architectureSplits);
		size_t n_sets = architectureSplits.size();
		ThreadPool<MatrixStack<3, std::pair<float, char> > > pool(n_threads,
				matrix);
		for (size_t i = 0; i < n_sets; ++i)
			same_architecture_aln(architectureSplits[i], sim_mat, gop, gep,
					splitted_sets[i], pool);
		pool.wait();

		// Align sequence sets with different architecture
		size_t n_elems = set.size();
		std::stack<std::pair<const TreeNode*, unsigned int> > to_do;
		to_do.push(std::pair<const TreeNode*, int>(arch_guide_tree.root(), 0));
		int *tree_helper = new int[arch_guide_tree.n_species() * 2];
		for (size_t i = 0; i < arch_guide_tree.n_species() * 2; ++i)
			tree_helper[i] = i;
		std::vector<std::pair<unsigned int, unsigned int> > gap1, gap2;
		Matrix<size_t> ids(n_elems, 1);
		for (size_t i = 0; i < n_elems; ++i)
			ids[i][0] = i;

		std::string edit_string1, edit_string2;
		std::vector<float> pos_gops1, pos_gops2;
		std::vector<float> pos_geps1, pos_geps2;
		while (!to_do.empty())
		{
			std::pair<const TreeNode*, unsigned int> &current = to_do.top();
			if (current.first->children.empty())
				to_do.pop();
			else if (current.second == current.first->children.size())
			{
				int first = tree_helper[current.first->children[0]->id];
				int second = tree_helper[current.first->children[1]->id];
				tree_helper[current.first->id] =
						tree_helper[current.first->children[0]->id];
				last = first;	//current.first->id;
				if (!relax)
					aln_different_architectures(splitted_sets[first], splitted_sets[second], sim_mat, gop, gep, pool);
				else
					aln_different_architecturesRelaxed(splitted_sets[first], splitted_sets[second], sim_mat, gop, gep, pool);
				pool.wait();
				to_do.pop();
			}
			else
			{
				to_do.push(
						std::pair<const TreeNode*, int>(
								&(*current.first->children[current.second]),
								0));
				++current.second;
			}
		}
		delete[] tree_helper;
		//	pool.stop();
	}

	// Merge
	merge_sequences(splitted_sets[last], 0, splitted_sets[last].size() - 1);
	set.clear();
	set.transfer(splitted_sets[last][splitted_sets[last].size() - 1]);
}

} // namespace MDAT

#endif /* SEQ_ALIGN_HPP_ */
