/*
 * Tree.cpp
 *
 *  Created on: May 14, 2012
 *      Author: Carsten Kemena
 *
 * This file is part of MDAT.
 *
 * MDAT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MDAT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with MDAT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "Tree.hpp"

namespace MDAT {


using namespace std;


Tree::Tree():_root(nullptr), _n_species(0)
{}

Tree::~Tree()
{
	clear();
}


/**
 * \brief Calculates the UPGMA tree
 * @param dist_mat The distance matrix to use
 * @param names The names of the sequences
 * @param n_members The number of sequences at each position.
 */
void
Tree::upgma(const Matrix<float> &dist_mat, const vector<string> &names, vector<int> &n_members)
{
	Matrix<float> distance_mat = dist_mat;
	int n_taxa = dist_mat.dim1();
	_n_species = n_taxa;

	int i,j,k,small,large,min,min_k=-1,min_j=-1,new_dist;
	int n_nodes=n_taxa*2-1;
	int node_pos = n_taxa;
	TreeNode** nodes = new TreeNode*[n_nodes];

	for (i=0; i<n_nodes; ++i)
		if (i<n_taxa)
			nodes[i] = new TreeNode(i, names[i]);
		else
			nodes[i] = new TreeNode(i, "");

	if (n_taxa==2)
	{
		_root=nodes[2];
		_root->children.push_back(std::shared_ptr<TreeNode>(nodes[0]));
		_root->children.push_back(std::shared_ptr<TreeNode>(nodes[1]));
		_root->edge_length=0;
		delete[] nodes;
		return;
	}

	n_nodes=n_taxa;
	// the number of nodes to exist
	int *node_ids = new int[n_nodes];
	for (i=0; i<n_nodes; ++i)
		node_ids[i]=i;
	for (i=1; i<n_nodes; ++i)
	{
		// go through the matrix, find smallest match
		min=INT_MAX;
		for (j=0; j<n_nodes; ++j)
		{
			if (n_members[j]==0)
				continue;
			for (k=j+1; k<n_nodes; ++k)
			{
				if (n_members[k]==0)
					continue;
				if (distance_mat[j][k]<min)
				{
					min=distance_mat[j][k];
					min_k=k;
					min_j=j;
				}
			}
		}
		nodes[node_pos]->children.push_back(std::shared_ptr<TreeNode>(nodes[node_ids[min_k]]));
		nodes[node_pos]->children.push_back(std::shared_ptr<TreeNode>(nodes[node_ids[min_j]]));

		if (min_k<min_j)
		{
			small=min_k;
			large=min_j;
		}
		else
		{
			small=min_j;
			large=min_k;
		}

		// calculate new distances
		for (j=0; j<n_nodes; ++j)
		{
			if ((n_members[j]==0) || (j==min_k) || (j==min_j))
				continue;

			new_dist = distance_mat[j][min_k]*n_members[min_k]+ distance_mat[min_j][j]*n_members[min_j];
			new_dist /= (n_members[min_k]+n_members[min_j]);
			distance_mat[small][j] = distance_mat[j][small]=new_dist;
		}
		node_ids[small]=node_pos++;
		n_members[small]+=n_members[large];
		n_members[large]=0;
	}
	_root=nodes[node_pos-1];
	delete[] nodes;
	delete[] node_ids;
}

/**
 * \brief Calculates the neighbor joining tree.
 * @param dist_mat The distance matrix to use.
 * @param names The names of the sequences.
 */
void Tree::nj(const Matrix<float> &dist_mat, const vector<string> &names)
{

	size_t n_taxa = dist_mat.dim1();
	_n_species = n_taxa;
	Matrix<float> d_mat(n_taxa, n_taxa);
	float *r_val = new float[n_taxa];
	d_mat=dist_mat;
	size_t i,j, k;

	// Calculate nodes
	float min_val = FLT_MAX;
	TreeNode** nodes = new TreeNode*[n_taxa];
	for (i=0; i<n_taxa; ++i)
		nodes[i]=new TreeNode(i, names[i]);

	if (n_taxa==2)
	{
		_root=new TreeNode();
		_root->children.push_back(std::shared_ptr<TreeNode>(nodes[0]));
		_root->children.push_back(std::shared_ptr<TreeNode>(nodes[1]));
		_root->edge_length=0;
		_n_species=n_taxa;
		return;
	}

	size_t new_node_id = n_taxa-1;
	TreeNode *tmpNode;
	float dist_i_new;
	size_t x_taxa = n_taxa;
	size_t min_i=-1, min_j=-1;
	size_t n_inner_nodes = n_taxa-2;

	float tmp_score;
	for (k=0; k<n_inner_nodes; ++k)
	{
		for (i=0; i<n_taxa; ++i)
			r_val[i]=0;
		for (i=0; i<n_taxa; ++i)
		{
			if (nodes[i]==NULL)
				continue;
			for (j=i+1; j<n_taxa; ++j)
			{
				if (nodes[j]==NULL)
					continue;
				r_val[i]+=d_mat[i][j];
				r_val[j]+=d_mat[i][j];
			}
			r_val[i]*=(1.0/(x_taxa-2));
		}
		min_val = FLT_MAX;

		// Find minimum value
		for (i=0; i<n_taxa; ++i)
		{
			if (nodes[i]==NULL)
				continue;
			for (j=i+1; j<n_taxa; ++j)
			{
				if (nodes[j]==NULL)
					continue;
				tmp_score = d_mat[i][j] - (r_val[i]+r_val[j]);
				if (tmp_score < min_val)
				{
					min_val=tmp_score;
					min_i = i;
					min_j = j;
				}
			}
		}

		dist_i_new = (d_mat[min_i][min_j] + r_val[min_i]-r_val[min_j])/2;
		nodes[min_i]->edge_length=dist_i_new;
		nodes[min_j]->edge_length=d_mat[min_i][min_j]-dist_i_new;
		tmpNode = new TreeNode(++new_node_id);
		tmpNode->children.push_back(std::shared_ptr<TreeNode>(nodes[min_i]));
		tmpNode->children.push_back(std::shared_ptr<TreeNode>(nodes[min_j]));
		nodes[min_j]=NULL;
		nodes[min_i]=tmpNode;
		for (i=0; i<n_taxa; ++i)
		{
			if ((nodes[i]!=NULL) && (i!=min_i) && (i!=min_j))
				d_mat[min_i][i]=d_mat[i][min_i]=(d_mat[min_i][i] + d_mat[min_j][i] - d_mat[min_i][min_j])*0.5;
		}
		--x_taxa;
	}
	for (j=0; j<n_taxa; ++j)
	{
		if (nodes[j]!=NULL)
		{
			min_j=j;
			break;
		}
	}
	for (j=j+1; j<n_taxa; ++j)
	{
		if (nodes[j]!=NULL)
		{
			min_i=j;
			break;
		}
	}
	_root=new TreeNode();
	_root->children.push_back(std::shared_ptr<TreeNode>(nodes[min_i]));
	_root->children.push_back(std::shared_ptr<TreeNode>(nodes[min_j]));
	delete[] nodes;
	delete[] r_val;
}

/**
 * \brief
 */
void
Tree::clear()
{
	this->_n_species=0;
	delete _root;
	this->_root=NULL;

/*	unsigned int i;
	stack<TreeNode* > to_do;
	TreeNode *current_node;
	to_do.push(_root);
	while (!to_do.empty())
	{
		current_node = to_do.top();
		to_do.pop();
		for (i=0; i< current_node->children.size(); ++i)
		{
			if (current_node->children[i]->children.empty())
				delete current_node->children[i];
			else
				to_do.push(current_node->children[i]);
		}
		delete current_node;
	}
	_n_species=0;*/
}



} /* namespace MDAT */
