/*
 * Tree.h
 *
 *  Created on: May 14, 2012
 *      Author: Carsten Kemena
 *
 * This file is part of MDAT.
 *
 * MDAT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MDAT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with MDAT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/*! \file Tree.hpp
 * \brief Tree management
 */

#ifndef TREE_H_
#define TREE_H_


// C header
#include <cfloat>
#include <cstdio>
#include <cmath>
#include <cstdlib>

// C++ header
#include <map>
#include <stack>
#include <string>
#include <vector>
#include <memory>

// MDAT header
#include "../Sequence/SequenceSet.hpp"
#include "../utils/Matrix.hpp"


namespace MDAT {

/**
 * \brief A node to represent a unit in a tree.
 */
struct TreeNode
{
	unsigned int id; // The id of the node.
	std::string name; // The name of the node.
	float edge_length; // The length of the edge.
	std::vector<std::shared_ptr<TreeNode> > children; //The children of this node.


	TreeNode():id(0), name(""), edge_length(-1), children()
	{}

	TreeNode(unsigned int id_):id(id_), name(""), edge_length(-1), children()
	{}

	TreeNode(const std::string &name_):id(0), name(name_), edge_length(-1), children()
	{}

	TreeNode(unsigned int id_, const std::string &name_):id(id_),name(name_), edge_length(-1), children()
	{}

	bool
	is_leaf() const
	{
		return (!children.empty());
	}


};


/**
 * \brief This class provides several methods to read, compute and manipulate trees.
 */
class Tree {
private:
	TreeNode *_root;
	size_t _n_species;

public:

	/**@{*/
	//! \name Constructors & Destructors
	/**
	 * Simple Constructor for a tree.
	 */
	Tree();
	Tree(const MDAT::Tree&) = delete;
	virtual ~Tree();
	/**@}*/

	Tree& operator=(const Tree&) = delete;
	Tree& operator=(Tree &&other)
	{
		if (this != &other)
		{
			clear();
			_root=other._root;
			_n_species=other._n_species;
			other._root=nullptr;
			other._n_species=0;
		}
		return *this;
	}

	/**@{*/
	//! \name Tree computation methods

	/**
	 * \brief Tree construction using the neighbour joining method.
	 *
	 * This method returns a rooted binary tree. As the method usually returns a unrooted tree
	 * one of the edge lengths of the tree has been set to 0. This method has been first described in:
	 * Saitou N, Nei M. "The neighbor-joining method: a new method for reconstructing phylogenetic trees." Molecular Biology and Evolution, volume 4, issue 4, pp. 406-425, July 1987.
	 * \param dist_mat A matrix giving the distances between the species.
	 * \param names The names of the species.
	 */
	void nj(const Matrix<float> &dist_mat, const std::vector<std::string> &names);

	/**
	 * \brief Tree construction using the UPGMA method.
	 *
	 * The UPGMA (Unweighted Pair Group Method with Arithmetic Mean) method has been first described in:
	 * Sokal R and Michener C (1958). "A statistical method for evaluating systematic relationships". University of Kansas Science Bulletin 38: 1409–1438.
	 * \param dist_mat A matrix giving the distances between the species.
	 * \param names The names of the species.
	 */
	//void upgma(const Matrix<float> &dist_mat, const std::vector<std::string> &names);
	void upgma(const Matrix<float> &dist_mat, const std::vector<std::string> &names, std::vector<int> &n_members);
	/**@}*/


	/**
	 * \brief Deletes all nodes from the tree.
	 */
	void clear();

	/**
	 * \brief Returns the number of species in this tree.
	 * \return The number of species
	 */
	size_t n_species() const
	{
		return _n_species;
	}

	/**
	 * \brief Returns pointer to the root of the tree.
	 * @return Pointer to the root.
	 */
	const TreeNode* root() const
	{
		return _root;
	}

	TreeNode* root()
	{
		return _root;
	}

	/**
	 * \brief Returns true if no species are in the node.
	 * \return True if empty else false.
	 */
	bool empty() const
	{
		return (_n_species==0);
	}
};


} /* namespace MDAT */

#endif /* TREE_H_ */
