/*
 * Vector.hpp
 *
 *  Created on: Oct 9, 2013
 *      Author: ckeme_01
 */

#ifndef VECTOR_HPP_
#define VECTOR_HPP_

#include <vector>
#include <algorithm>

#include "../utils/Matrix.hpp"


namespace MDAT
{

/**
 * \brief A class to represent a simple vector.
 */
template<typename DataType>
class Vector
{

private:
	std::vector<DataType> _vec;
	size_t _id;

public:
	Vector(size_t length, size_t id_value=0):_vec(std::vector<DataType>(length)), _id(id_value)
	{}

	virtual ~Vector()
	{}

	DataType &operator[](unsigned int index)
	{
		return _vec[index];
	}

	const DataType &operator[](unsigned int index) const
	{
		return _vec[index];
	}

	size_t id()
	{
		return _id;
	}

	void
	id(size_t value)
	{
		_id=value;
	}

	void
	resize(size_t new_size)
	{
		_vec.resize(new_size);
	}


	float
	dist(const Vector<DataType> &other)
	{
		typename std::vector<DataType>::const_iterator it1, it2=other._vec.begin(), it1_end=_vec.end();
		DataType val=0;
		DataType sum=0;
		for (it1=_vec.begin(); it1!=it1_end; ++it1)
		{
			val+=std::min(*it1,*it2);
			sum+=std::max(*it1, *it2);
			++it2;
		}
		return 100-(100.0*val/sum);
	}
};

/**
 * \brief Turns a sequence into a vector.
 * \details Counts for each kmer the number of occurrences.
 * @param seq The sequence.
 * @param coded_alphabet The coded alphabet.
 * @param start The start from which to calculate the kmers from.
 * @param end To where the kmers need to be calculated.
 * @return A vector of kmer counts.
 */
template<typename SequenceType>
Vector<int> *
calc_km_vec(const SequenceType &seq, std::vector<short> &coded_alphabet, size_t start=0, size_t end=0)
{

	Vector<int> *vec_p = new Vector<int>(400, seq.id());
	Vector<int> &vec = *vec_p;

    unsigned int value = 20 * coded_alphabet[(int)seq[start]] + coded_alphabet[(int)seq[start+1]];
    short bad=0;
    if (coded_alphabet[(int)seq[start]]==20)
    	++bad;
    if (coded_alphabet[(int)seq[++start]]==20)
    	++bad;
    if (bad==0)
    	++vec[value];
    short c;
    ++start;
    for (size_t i = start; i<end+1; ++i)
    {
    	c = coded_alphabet[(int)seq[i-2]];
    	if (c==20)
    		--bad;
    	value -= 20*c;
    	value *= 20;

    	value += coded_alphabet[(int)seq[i]];
    	if (coded_alphabet[(int)seq[i]]==20)
    		++bad;

    	if (bad==0)
    		++vec[value];
    }
    return vec_p;
}


/**
 * \brief Calculates a distance matrix based on a sequence set.
 * @param set The sequence set.
 * @return The distance matrix.
 */
template<typename SequenceSetType>
Matrix<float> *
kmer_dist_mat(const SequenceSetType &set)
{
	int n_seqs=set.n_seqs();

	if (n_seqs==1)
		return NULL;

	std::vector<short> coded_alphabet(256);
	int i,j;
	for (i=0; i<256;++i)
		coded_alphabet[i]=20;
	char aa_alphabet[20]={'A','V','L','I','P','M','F','W','G','S','T','C','N','Q','Y','D','E','K','R','H'};
	short val=-1;
	for (i=0; i<20; ++i)
	{
		coded_alphabet[toupper(aa_alphabet[i])-0]=++val;
		coded_alphabet[tolower(aa_alphabet[i])-0]=val;
	}

	// turn sequences into vectors for guide tree
	std::vector<Vector<int>* >vec_set(set.size());

	for (i=0; i<n_seqs; ++i)
		vec_set[i]=calc_km_vec(set[i], coded_alphabet, 0, set[i].length());


	Matrix<float> *dist_mat_p=new Matrix<float>(n_seqs, n_seqs, 0);
	Matrix<float> &dist_mat=*dist_mat_p;
	for (i=0; i<n_seqs; ++i)
	{
		dist_mat[i][i]=0;
		for (j=i+1; j<n_seqs; ++j)
			dist_mat[i][j]=dist_mat[j][i]=vec_set[i]->dist(*(vec_set[j]));
	}
	for (i=0; i<n_seqs; ++i)
		delete vec_set[i];
	return dist_mat_p;
}

}



#endif /* VECTOR_HPP_ */
