/*
 * CRS.hpp
 *
 *  Created on: Jul 26, 2013
 *      Author: Carsten Kemena
 */

/*! \file CRS_Mat.hpp
    \brief PFAM domain mat
*/


#ifndef CRS_Mat_HPP_
#define CRS_Mat_HPP_

// C++ header
#include <string>
#include <fstream>
#include <vector>

namespace MDAT {


/**
 * \brief A class to store the Pfam domain matches in a compact way.
 *
 * The domain matrix is symmetric so that only half of the entries are actually stored.
 * The matrix can be read from a binary file of the following format:
 *
 * <pre>
 * ID var name    length     type     variable_name
 * 0  name        10         char     name
 * 1  n_domains   1          int      n_domains
 * 2  n_values    1          int      n_values
 * 3  ids         n_domains  int      The domain identifier as int (eg PF00005 = 5)
 * 4  row_ids     n_domains  int      The row ids
 * 5  col_ids     n_values   int      The column ids
 * 6  values      n_values   DataType The actual values stored in the matrix
 * </pre>
 */
template<typename DataType>
class CRS_Mat
{
private:
	std::string _name; //stores the matrix name
	int _n_domains; // Number of domains
	int _n_vals; // Number of values stored
	std::vector<int> _ids; // Contains the PFAM identifiers as number (e.g. PF00005 = 5)
	std::vector<int> _row_ids; // Contains the row id
	std::vector<int> _col_ids; // Contains the column_id
	std::vector<DataType> _vals; // Contains the values stored in the matrix

public:
	// Constructors & Destructors

	/**
	 * \brief Standard Constructor
	 */
	CRS_Mat();

	/**
	 * \brief Constructor which immediately reads the file
	 * @param mat_f
	 */
	CRS_Mat(const std::string &mat_f);

	/**
	 * \brief Destructor
	 */
	~CRS_Mat();

	// Basic functions
	/**
	 * \brief Returns the identifier of the Matrix
	 * @return
	 */
	std::string name()
	{
		return _name;
	}

	/**
	 * \brief Reads a CRS matrix from a binary file.
	 * @param mat_f The file to read.
	 */
	void read(const std::string &mat_f);

	/**
	 * \brief Returns the values stored in the matrix.
	 * @param i The row id
	 * @param j The column id
	 * @return The value stored in the matrix.
	 */
	DataType value(int i, int j) const;

};


template<typename DataType>
CRS_Mat<DataType>::CRS_Mat():_name(""), _n_domains(0), _n_vals(0), _ids(), _row_ids(), _col_ids(), _vals()
{}

template<typename DataType>
CRS_Mat<DataType>::CRS_Mat(const std::string &mat_f):_name(""), _n_domains(0), _n_vals(0), _ids(), _row_ids(), _col_ids(), _vals()
{
	read(mat_f);
}

template<typename DataType>
CRS_Mat<DataType>::~CRS_Mat()
{}


template<typename DataType>
void
CRS_Mat<DataType>::read(const std::string &mat_f)
{
	std::ifstream mat_F;
	mat_F.open(mat_f.c_str(), std::ios::in|std::ios::binary);
	mat_F.exceptions( std::ifstream::failbit | std::ifstream::badbit);
	_name.resize(11);
	mat_F.read(&_name[0], 10*sizeof(char));
	_name[10]='\0';
	int i;
	for (i=0; i<11; ++i)
		if (_name[i]=='\0')
			break;
	_name.resize(i);

	mat_F.read((char*)&_n_domains, sizeof(int));
	mat_F.read((char*)&_n_vals, sizeof(int));
	_ids.resize(_n_domains);
	_row_ids.resize(_n_domains);
	_col_ids.resize(_n_vals);
	_vals.resize(_n_vals);
	mat_F.read((char*)&_ids[0], sizeof(int)*_n_domains);
	mat_F.read((char*)&_row_ids[0], sizeof(int)*_n_domains);
	mat_F.read((char*)&_col_ids[0], sizeof(int)*_n_vals);
	mat_F.read((char*)&_vals[0], sizeof(DataType)*_n_vals);
	mat_F.close();
}




template<typename DataType>
DataType
CRS_Mat<DataType>::value(int i, int j) const
{
	std::vector<int>::const_iterator it1, it2, it_begin=_ids.begin(), it_end=_ids.end();

	it1=std::lower_bound (it_begin, it_end, i);
	it2=std::lower_bound (it_begin, it_end, j);
	if ((it1 == it_end) || (it2 == it_end))
		return -1;

	if (j<i)
		swap(it1, it2);

	int id1=it1-it_begin;
	int id2=it2-it_begin;

	int id_col=_row_ids[id1];
	int end=(id1!=(_n_domains-1)) ? _row_ids[id1+1] : _n_vals;
	for (int i =id_col; i<end; ++i)
	{
		if (_col_ids[i]==id2)
			return _vals[i];
	}
	return -1;
}

}


#endif /* CRS_Mat_HPP_ */
