/*
 * Matrix.h
 *
 *  Created on: Apr 12, 2012
 *      Author: Carsten Kemena
 *
 * This file is part of MDAT.
 *
 * MDAT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MDAT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with MDAT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#ifndef MATRIX_HPP_
#define MATRIX_HPP_


// C header
#include <cstdlib>
#include <cstdio>

// C++ header
#include <vector>



namespace MDAT {

/**
 * \brief A simple class to produce 2 dimensional matrices.
 */
template <typename DataType>
class Matrix {


private:
	std::vector<std::vector<DataType> > _matrix;
public:
	/**
	 * \brief Standard constructor
	 */
	Matrix():_matrix()
	{}

	/**
	 * \brief Constructor setting the size.
	 * @param dim1 Size of the first dimension.
	 * @param dim2 Size of the second dimension.
	 */
	Matrix(size_t dim1, size_t dim2);

	/**
	 * \brief Constructor allowing initialization of fields.
	 * @param dim1 Size of the first dimension.
	 * @param dim2 Size of the second dimension.
	 * @param init Value to initialize the fields with.
	 */
	Matrix(size_t dim1, size_t dim2, DataType init);
    Matrix(const Matrix&) = default;
    Matrix & operator=(const Matrix&) = default;
	/**
	 * \brief Destructor
	 */
	virtual ~Matrix()
	{

	}

	/**
	 * \brief Access operator
	 * @param index The index to access.
	 * @return Reference to the field.
	 */
	std::vector<DataType> &operator[](unsigned int index)
	{
		return _matrix[index];
	}

	/**
	 * \brief Access operator
	 * @param index The index to access.
	 * @return Reference to the field.
	 */
	const std::vector<DataType> &operator[](unsigned int index) const
	{
		return _matrix[index];
	}


	/**
	 * \brief Resizes the matrix to the new dimensions.
	 * @param dim_1 The first dimension
	 * @param dim_2 The second dimension
	 */
	void resize(size_t dim_1, size_t dim_2)
	{
		_matrix.resize(dim_1);
		for (size_t i=0; i<dim_1; ++i)
			_matrix[i].resize(dim_2);
	}

	void ensure(size_t dim_1, size_t dim_2)
	{
		if (dim_1 > _matrix.size())
			_matrix.resize(dim_1);
		for (size_t i=0; i<dim_1; ++i)
		{
			if (_matrix[i].size() <dim_2)
				_matrix[i].resize(dim_2);
		}
	}


	/**
	 * \brief Returns the size of the first dimension.
	 * @return The size of the first dimension.
	 */
	size_t dim1() const
	{
		return _matrix.size();
	}

	/**
	 * \brief Returns the size of the second dimension.
	 * @return The size of the second dimension.
	 */
	size_t dim2() const
	{
		return (_matrix.size() > 0) ? _matrix[0].size() : 0;
	}

	void
	fill(const DataType &value)
	{
		size_t dim_1 = _matrix.size();
		size_t dim_2= (dim_1>0) ? _matrix[0].size() : 0;
		size_t j;
		for(size_t i=0; i<dim_1; ++i)
		{
			std::vector<DataType> &vec=_matrix[i];
			for (j=0; j<dim_2; ++j)
				vec[j] = value;
		}
	}
};


template <typename DataType>
Matrix<DataType>::Matrix(size_t dim_1, size_t dim_2, DataType init):_matrix(dim_1, std::vector<DataType>(dim_2, init))
{}


template <typename DataType>
Matrix<DataType>::Matrix(size_t dim_1, size_t dim_2):_matrix(dim_1, std::vector<DataType>(dim_2))
{}



} /* namespace MDAT */

#endif /* MATRIX_HPP_ */
