/*
 * MatrixStack.hpp
 *
 *  Created on: Sep 11, 2013
 *      Author: ckeme_01
 */

/**
 * \file MatrixStack.hpp
 * \brief File to contain the MatrixStack class.
 */

#ifndef MATRIXSTACK_HPP_
#define MATRIXSTACK_HPP_



// C header
#include <cstdlib>

// C++ header
#include <vector>

// MDAT
#include "Matrix.hpp"

namespace MDAT {

/**
 * \brief A simple class to contain a set of 2 dimensional matrices.
 * \tparam n The number of matrices
 * \tparam DataType The type of data to be stored in the matrices.
 */
template <int n, typename DataType>
class MatrixStack
{

private:
	std::vector<Matrix<DataType> > _stack;

public:

	/**
	 * \brief Default constructor.
	 */
	MatrixStack():_stack(n,Matrix<DataType>(1,1))
	{}

	/**
	 * \brief Initializes the Stack with the given sizes.
	 * @param dim1 Size of the first dimension.
	 * @param dim2 Size of the second dimension.
	 */
	MatrixStack(size_t dim1, size_t dim2):_stack(n,Matrix<DataType>(dim1,dim2))
	{}

	/**
	 * \brief The copy constructor.
	 * @param The object to copy.
	 */
    MatrixStack(const MatrixStack&) = default;

    /**
     * \brief The move constructor
     * @param The object to move.
     */
    MatrixStack(MatrixStack&&) = default;

    /**
     * \brief Copy assignment operator.
     * @param The object to copy.
     * @return Reference to object.
     */
    MatrixStack & operator=(const MatrixStack&) = default;

    /**
     * \brief Move assignment operator.
     * @param The object to move.
     * @return Reference to object.
     */
    MatrixStack & operator=(MatrixStack&&) = default;


	virtual ~MatrixStack()
	{}


	/**
	 * \brief Access operator
	 * @param index The index to acess.
	 * @return Reference to the field.
	 */
	Matrix<DataType> &operator[](unsigned int index)
	{
		return _stack[index];
	}

	/**
	 * \brief Access operator
	 * @param index The index to acess.
	 * @return Reference to the field.
	 */
	const Matrix<DataType> &operator[](unsigned int index) const
	{
		return _stack[index];
	}


	/**
	 * \brief Returns size of the first dimension.
	 * @return The size of the first dimension.
	 */
	size_t
	dim1()
	{
		return _stack[0].dim1();
	}

	/**
	 * \brief Returns size of the second dimension.
	 * @return The size of the second dimension.
	 */
	size_t
	dim2()
	{
		return _stack[0].dim2();
	}


	/**
	 * \brief Resizes all matrices in the stack.
	 * @param new_size1 The new size of the first dimension.
	 * @param new_size2 The new size of the second dimension.
	 */
	void
	resize(size_t new_size1, size_t new_size2)
	{
		for (size_t i=0; i<n; ++i)
			_stack[i].resize(new_size1, new_size2);
	}

	/**
	 * \brief Will make sure that the matrix is at least of this size.
	 * @param new_size1 The new size of dimension 1.
	 * @param new_size2 The new size of dimension 2.
	 */
	void
	ensure(size_t new_size1, size_t new_size2)
	{
		for (size_t i=0; i<n; ++i)
			_stack[i].ensure(new_size1, new_size2);
	}

	/**
	 * \brief Fills all matrixes with the same value.
	 * @param value The value to be used for the filling.
	 */
	void
	fill(const DataType &value)
	{
		for (size_t i=0; i<n; ++i)
			_stack[i].fill(value);
	}
};

}

#endif /* MATRIXSTACK_HPP_ */
