/*
 * ThreadPool2.hpp
 *
 *  Created on: Nov 8, 2013
 *      Author: ckeme_01
 */

// based on https://github.com/progschj/ThreadPool/blob/master/ThreadPool.h

#ifndef THREADPOOL2_HPP_
#define THREADPOOL2_HPP_


#include <atomic>
#include <vector>
#include <queue>
#include <memory>
#include <thread>
#include <mutex>
#include <condition_variable>
#include <future>
#include <functional>
#include <stdexcept>
//#include <chrono>


template<typename T>
class ThreadPool {

private:
    // need to keep track of threads so we can join them
    std::vector< std::thread > workers;
    std::atomic<unsigned int> _running;
    // the task queue
    std::queue<std::function<void(T&)> > _tasks;

    // synchronization
    std::mutex queue_mutex;
    std::condition_variable condition;

    std::mutex _returned_mutex;
    std::condition_variable _returned;
    bool _stop;
    void _join();

public:
    ThreadPool(size_t, const T &local_data);
    ThreadPool(const ThreadPool&&) = delete;

    template<class F, class... Args>
    void addTask(F&& f, Args&&... args);
    ~ThreadPool();

    void stop();
    void wait();
    int running()
    {
    	return _running.load();
    }

};

// the constructor just launches some amount of workers
template<typename T>
inline ThreadPool<T>::ThreadPool(size_t threads, const T &local_data)
    : workers(), _running(0), _tasks(), queue_mutex(), condition(), _returned_mutex(), _returned(), _stop(false)
{
    for(size_t i = 0;i<threads;++i)
    {
        workers.emplace_back(
            [this] (const T &local_data)
            {
    			std::function<void(T&)> task;
    			T data=local_data;
                while(true)
                {
                	{
						std::unique_lock<std::mutex> lock(this->queue_mutex);
						while(!this->_stop && _tasks.empty())
							this->condition.wait(lock);
						if(this->_stop && _tasks.empty())
							return;
						task = _tasks.front();
						_tasks.pop();
                	}
                    task(data);
                    {
                    	std::unique_lock<std::mutex> lock(_returned_mutex);
                    	--(this->_running);
                    }
                    this->_returned.notify_one();
                }
            },
            local_data
        );
    }
}

// add new work item to the pool
template<typename T>
template<class F, class... Args>
void ThreadPool<T>::addTask(F&& f, Args&&... args)
{
    // don't allow enqueueing after stopping the pool
    if(_stop)
    	throw std::runtime_error("enqueue on stopped ThreadPool");

	auto task = std::function<void(T&)> (
    		std::bind(std::forward<F>(f), std::forward<Args>(args)..., std::placeholders::_1)
		);
    {
    	++_running;
        std::unique_lock<std::mutex> lock(queue_mutex);
        _tasks.push(task);
        condition.notify_all();
    }
    {
		std::unique_lock<std::mutex> lock(_returned_mutex);
		while(_running.load()==workers.size())
			this->_returned.wait(lock);
    }

}

template<typename T>
void
ThreadPool<T>::_join()
{
	{
		std::unique_lock<std::mutex> lock(queue_mutex);
		_stop = true;
	}
	condition.notify_all();
	for(size_t i = 0; i<workers.size(); ++i)
		workers[i].join();
}

template<typename T>
void
ThreadPool<T>::stop()
{
	if (!_stop)
		_join();
}


template<typename T>
void
ThreadPool<T>::wait()
{
	std::unique_lock<std::mutex> lock(_returned_mutex);
	while(_running.load()!=0)
	{
    	this->_returned.wait(lock);
	}
}


// the destructor joins all threads if neccessary
template<typename T>
ThreadPool<T>::~ThreadPool()
{
	if (!_stop)
		_join();

}




#endif /* THREADPOOL2_HPP_ */
