/*
 * mda_pdf_out.cpp
 *
 *  Created on: 22 Sep 2013
 *      Author: ck
 */


#include "mda_pdf_out.hpp"

using namespace MDAT;
using namespace std;
using namespace boost;

/**
 * \brief Simple color to store RGB information.
 * \details In this context the values are percentages.
 */
struct Color
{
	Color(float r_, float g_, float b_):r(r_), g(g_), b(b_)
	{}

	float r;
	float g;
	float b;
};


// Color generation as described in http://martin.ankerl.com/2009/12/09/how-to-create-random-colors-programmatically/
/**
 * \param Turns a HSV color to RGB.
 * @param h The hue.
 * @param s The saturation
 * @param v The value
 * @return The color in RGB format.
 */
Color
hsv_to_rgb(float h, float s, float v)
{
  int h_i = (h*6);
  float f = h*6 - h_i;
  float p = v * (1 - s);
  float q = v * (1 - f*s);
  float t = v * (1 - (1 - f) * s);
  if (h_i==0)
	  return (Color(v, t, p));
  if (h_i==1)
	  return (Color(q, v, p));
  if (h_i==2)
	  return (Color(p, v, t));
  if (h_i==3)
	  return (Color(p, q, v));
  if (h_i==4)
	  return (Color(t, p, v));
  else
	  return (Color(v, p, q));
}


/**
 * \brief Produces a color.
 * @param Random_num The random number to use.
 * @return The resulting color.
 */
Color
generate_color(float random_num)
{
	double golden_ratio_conjugate = 0.618033988749895;
	double h = random_num;// # use random start value
	h += golden_ratio_conjugate;
	h = fmod(h,1);
	return hsv_to_rgb(h, 0.5, 0.95);
}


/**
 * \brief Produced a set of random colors to be more or less distinguishable.
 * @param[in] number The number of colors to produce.
 * @param[out] color_vec The produced colors.
 */
void
produce_colors(int number, vector<Color> &color_vec)
{
	std::default_random_engine generator(48);
	std::uniform_real_distribution<double> distribution(0.0,1.0);
	for (int j = 0; j < number; ++j)
	   color_vec.push_back(generate_color(distribution(generator)));
}


/**
 * \brief Draws a rectangle with rounded corners.
 * @param cr The surface to draw on.
 * @param x The start in x.
 * @param y The start in y.
 * @param width The width of rectangle.
 * @param height The height of the rectangle.
 */
void
draw_rectangle(cairo_t *cr, double x, double y, double width, double height)
{
    double aspect        = 1.0;     /* aspect ratio */
    double corner_radius = height / 10.0;   /* and corner curvature radius */
	double radius = corner_radius / aspect;
	double degrees =  3.14159265 / 180.0;
	cairo_new_sub_path (cr);
	cairo_arc (cr, x + width - radius, y + radius, radius, -90 * degrees, 0 * degrees);
	cairo_arc (cr, x + width - radius, y + height - radius, radius, 0 * degrees, 90 * degrees);
	cairo_arc (cr, x + radius, y + height - radius, radius, 90 * degrees, 180 * degrees);
	cairo_arc (cr, x + radius, y + radius, radius, 180 * degrees, 270 * degrees);
	cairo_close_path (cr);
	cairo_fill_preserve (cr);
	cairo_set_source_rgb(cr, 1,1, 1);
	cairo_set_line_width (cr, 0.5);
	cairo_stroke (cr);


}


/**
 * \brief Draws an empty domain (just a line).
 * @param cr The sruface to draw on.
 * @param x The start in x.
 * @param y The start in y.
 */
void
draw_empty_domain(cairo_t *cr, double x, double y)
{

	double width=30;
	double height=10;

	cairo_set_line_width(cr, 1);
	cairo_set_source_rgb(cr, 0, 0, 0);
	cairo_move_to(cr, x-2.5, y+height/2);
	cairo_line_to(cr, x+width+2.5, y+height/2);
	cairo_stroke(cr);
}


/**
 * \brief Draws a domain.
 * @param ctx Surface to draw on.
 * @param x The start in x.
 * @param y The stop in y.
 * @param dom The domain to draw.
 * @param c_dom The color to use for the domain.
 * @param c_text The color to use for the text.
 */
void
draw_domain(cairo_t *ctx, float x, float y, const MDAT::Domain &dom, const Color &c_dom, const Color &c_text)
{
	cairo_set_line_width(ctx, 1);
	cairo_set_source_rgb(ctx, 0, 0, 0);
	cairo_move_to(ctx, x-2.5, y+5);
	cairo_line_to(ctx, x+30+2.5, y+5);
	cairo_stroke(ctx);
	cairo_set_source_rgb(ctx, c_dom.r, c_dom.g, c_dom.b);
	draw_rectangle(ctx, x, y, 30, 10);
	cairo_move_to(ctx, x+4, y+7);
	cairo_set_source_rgb(ctx, c_text.r, c_text.g, c_text.b);
	cairo_set_font_size (ctx, 4);
	string label = dom.accession() + "*" +lexical_cast<string>(dom.repeated());
	cairo_show_text(ctx, label.c_str());
}


/**
 * \brief Draws a domain.
 * @param ctx The surface to draw on.
 * @param x The start in x.
 * @param y The start in y.
 * @param name The domain name to plot.
 * @param c_dom The color of the domain.
 * @param c_text The color of the text.
 */
void
draw_domain(cairo_t *ctx, float x, float y, const string &name, const Color &c_dom, const Color &c_text)
{
	cairo_set_source_rgb(ctx, c_dom.r, c_dom.g, c_dom.b);
	draw_rectangle(ctx, x, y, 30, 10);
	cairo_move_to(ctx, x+4, y+7);
	cairo_set_source_rgb(ctx, c_text.r, c_text.g, c_text.b);
	cairo_set_font_size (ctx, 4);
	string label = name;// + "*" +lexical_cast<string>(dom.repeated());
	cairo_show_text(ctx, label.c_str());
}


/**
 * Draws the sequence names.
 * @param ctx The surface.
 * @param x The start point in x.
 * @param y The start point in y.
 * @param label_size The label size.
 * @param seqSet The sequence set.
 * @param height The size to be moved after each step.
 */
void
draw_names(cairo_t *ctx, double x, double y, double label_size, const ProteinSequenceSet<Default> &seqSet, double height)
{
	size_t n_seqs = seqSet.size();
	int n_archs = seqSet.dom_archis().size();
	int i;
	size_t j;
	string label;
	double current_size;
	cairo_text_extents_t text_size;
	for (i=0; i<n_archs; ++i)
	{
		int id = seqSet.dom_archis()[i].id();
		label=lexical_cast<string>(i+1) + ": ";
		cairo_text_extents (ctx, label.c_str(), &text_size);
		current_size = text_size.x_advance;
		cairo_move_to(ctx, x+label_size-current_size , y + height);
		cairo_show_text(ctx, label.c_str());

		for (j=0; j<n_seqs; ++j)
		{
			if (seqSet[j].dom_archi().id()==id)
			{
				label= seqSet[j].name();
				y+=height;
				cairo_move_to(ctx, x+label_size , y);
				cairo_show_text(ctx, label.c_str());
			}
		}
	}
}


/**
 * \brief Draws the Domain Matrix used in this alignment.
 * @param ctx The surface to draw on.
 * @param x The start position in x.
 * @param y The start position in y.
 * @param seqSet The sequences Set.
 * @param dom_mat The domain similarity matrix.
 * @param vec_color The colors for each domain.
 */
void
draw_legend(cairo_t *ctx, double x, double y, const ProteinSequenceSet<Default> &seqSet, const MDAT::Matrix<short> &dom_mat, const vector<Color> &vec_color)
{
	int n_diff_domains = seqSet.n_domains();
	int i,j;
	for (i=1; i<=n_diff_domains; ++i)
		draw_domain(ctx, x+i*35, y, seqSet.domain_name(i-1), vec_color[i-1], Color(0,0,0));
	for (i=1; i<=n_diff_domains; ++i)
		draw_domain(ctx, x, y+i*12, seqSet.domain_name(i-1), vec_color[i-1], Color(0,0,0));
	cairo_text_extents_t text_size;
	string label;
	for (i=1; i<=n_diff_domains; ++i)
	{
		for (j=1; j<=n_diff_domains; ++j)
		{
			if (i!=j)
			{
				if (dom_mat[i-1][j-1]>0)
					label=lexical_cast<string>(dom_mat[i-1][j-1]);
				else
					label="0";
				cairo_text_extents (ctx, label.c_str(), &text_size);
				cairo_move_to(ctx, x+i*35+30-text_size.x_advance-2, y+j*12+8);
				cairo_show_text(ctx,label.c_str());
			}
		}
	}
}


/**
 * \brief Calculates the domain color depending on similarity.
 * \details Domains are colored according to their average similarity to the other domains in the same column. The gradient begins at blue
 * (lowest similarity) and ends with ref (highest similarity). In total 11 colors are used in steps of ten, with the exceptions of the first
 * and end step (<=5, <=15, <=25, ..., <=95, <=100).
 * @param[in] mda The domain alignment.
 * @param[in] dom_mat The domain similarity matrix.
 * @param[out] colors The colors to be used.
 */
void
calculate_similarity_colors(const DomainArchitectureSet &mda, const MDAT::Matrix<short> &dom_mat, vector<map<int, Color> > &colors)
{
	int arch_length = mda[0].size();
	int n_sets=mda.size();
	map<int, int> counter;
	map<int, int>::const_iterator it1, it2, it_end;
	int i,j;
	int score, count;

	// Standard color gradient.
	vector<Color> colorGradient;
	colorGradient.emplace_back(0.235, 0.616, 0.976);
	colorGradient.emplace_back(0.560, 0.855, 0.949);
	colorGradient.emplace_back(0.200, 0.922, 0.757);
	colorGradient.emplace_back(0.180, 0.894, 0.475);
	colorGradient.emplace_back(0.165, 0.867, 0.196);
	colorGradient.emplace_back(0.365, 0.839, 0.149);
	colorGradient.emplace_back(0.592, 0.812, 0.133);
	colorGradient.emplace_back(0.784, 0.761, 0.118);
	colorGradient.emplace_back(0.757, 0.498, 0.102);
	colorGradient.emplace_back(0.729, 0.243, 0.090);
	colorGradient.emplace_back(0.698, 0.075, 0.149);

	// for each column
	int id;
	colors.resize(arch_length);
	for (i=0; i<arch_length; ++i)
	{
		// count occurrence of each domain
		counter.clear();
		for (j=0; j<n_sets; ++j)
		{
			if (mda[j][i].id() != -1)
				counter[mda[j][i].id()]+=1;
		}
		it_end = counter.end();
		for (it1=counter.begin(); it1!=it_end; ++it1)
		{
			score = 0;
			count = 0;
			for (it2=counter.begin(); it2!=it_end; ++it2)
			{
				if (it1==it2)
				{
					score += (it1->second-1) * 100;
					count += (it1->second-1);
				}
				else
				{
					score += it2->second * dom_mat[it2->first][it1->first];
					count += it2->second;
				}
			}
			if (count==0)
				score=100;
			else
				score /= count;
			id = round(1.0*score/10);
			colors[i].insert(std::pair<int, Color>(it1->first,colorGradient[id]));
		}
	}
}


void
write_mda_as_pdf(const ProteinSequenceSet<Default> &seqSet, const MDAT::Matrix<short> &dom_mat, const Opt_graphic &graphic_opts)
{
	const DomainArchitectureSet &mda = seqSet.dom_archis();
	int n_diff_domains = seqSet.n_domains();
	size_t arch_length = mda[0].size();
	size_t n_sets=mda.size();
	size_t n_seqs=seqSet.size();
	size_t max_arch_members = 0;
	size_t i,j;
	for (i=0; i<n_sets; ++i)
	{
		if (mda[i].n_members() > max_arch_members)
			max_arch_members=mda[i].n_members();
	}

	// produce colors
	vector<Color> vec_color;
	produce_colors(n_diff_domains, vec_color);

	//calculate size
	double total_x=0;
	double total_y=0;
	double simMat_x=0;
	double simMat_y=0;

	//double seqName_x=0;
	double seqName_y=0;
	if (graphic_opts.show_sim_mat)
	{
		simMat_y = (1+n_diff_domains)*12+10;
		simMat_x = (1+n_diff_domains)*35;
		total_x +=simMat_x;
		total_y +=simMat_y;
	}

	// draw front labels
	cairo_surface_t *csurface = cairo_pdf_surface_create(graphic_opts.graphic_f.c_str(), 10, 10);
	cairo_t *ctx = cairo_create(csurface);
	cairo_text_extents_t text_size;

	// identify textsize of labels;
	string label1, label2;
	label1 = lexical_cast<string>(n_sets)+": ";
	double current_size;
	cairo_text_extents (ctx, label1.c_str(), &text_size);
	double max_length_arch_id = text_size.x_advance;
	label2 = lexical_cast<string>(max_arch_members)+"x  ";
	cairo_text_extents (ctx, label2.c_str(), &text_size);
	double max_length_n_members = text_size.x_advance;
	double complete_length = max_length_arch_id + max_length_n_members;

	double mda_x=arch_length *35.0+complete_length;
	double mda_y=n_sets*12.5;
	total_y+=mda_y;
	total_x = max(simMat_x, mda_x);

	vector<map<int, Color> > colors;

	double text_height=0;

	if (graphic_opts.show_sequence_names)
	{
		cairo_set_font_size (ctx, 6);
		string x="|";
		cairo_text_extents (ctx, x.c_str(), &text_size);
		text_height = text_size.height+2;
		seqName_y = n_seqs*(text_height);
		total_y+=seqName_y+10;
	}

	cairo_surface_finish(csurface);
	cairo_destroy(ctx);
	cairo_surface_destroy(csurface);
	double start_x = 10;
	double start_y = 10;

	total_x+=20; //space for borders
	total_y+=20; //space for borders
	csurface = cairo_pdf_surface_create(graphic_opts.graphic_f.c_str(), total_x, total_y);
	ctx = cairo_create(csurface);
	cairo_rectangle(ctx, 0.0, 0.0, total_x, total_y);
	cairo_set_source_rgb(ctx, 0.9, 0.9, 0.9);
	cairo_fill(ctx);

	if (graphic_opts.show_sim_mat)
	{
		draw_legend(ctx, start_x, start_y, seqSet, dom_mat, vec_color);
		start_y+=simMat_y;
	}
	Color black(0, 0, 0);
	cairo_set_source_rgb(ctx, black.r, black.g, black.b);
	int color_id;
	cairo_set_font_size (ctx, 6);

	//plot labels
	for (i=0; i<n_sets; ++i)
	{
		label1=lexical_cast<string>(i+1) + ": ";
		label2=lexical_cast<string>(mda[i].n_members())+ "x  ";
		cairo_text_extents (ctx, label1.c_str(), &text_size);
		current_size = text_size.x_advance;
		cairo_move_to(ctx, start_x+complete_length-max_length_n_members-current_size , start_y + i*12.5+7);
		cairo_show_text(ctx, label1.c_str());
		cairo_text_extents (ctx, label2.c_str(), &text_size);
		current_size = text_size.x_advance;
		cairo_move_to(ctx, start_x+complete_length-current_size , start_y + i*12.5+7);
		cairo_show_text(ctx, label2.c_str());
	}


	// draw domains
	double start_dom = start_x + complete_length;

	if (graphic_opts.color_by_similarity)
	{
		vector<map<int, Color> > colors;
		calculate_similarity_colors(mda, dom_mat, colors);
		for (i=0; i<n_sets; ++i)
		{
			for (j=0; j<arch_length; ++j)
			{
				color_id = mda[i][j].id();
				if (color_id < 0)
					draw_empty_domain(ctx, start_dom+j*35, start_y);
				else
					draw_domain(ctx, start_dom+j*35, start_y, mda[i][j], colors[j].at(color_id), black);
			}
			start_y+=12.5;
		}
	}
	else
	{
		for (i=0; i<n_sets; ++i)
		{
			for (j=0; j<arch_length; ++j)
			{
				color_id = mda[i][j].id();
				if (color_id < 0)
					draw_empty_domain(ctx, start_dom+j*35, start_y);
				else
					draw_domain(ctx, start_dom+j*35,start_y, mda[i][j], vec_color[color_id], black);
			}
			start_y+=12.5;
		}
	}

	start_y=start_y-2.5+10;
	cairo_set_source_rgb(ctx, black.r, black.g, black.b);
	if (graphic_opts.show_sequence_names)
		draw_names(ctx, start_x, start_y, max_length_arch_id, seqSet, text_height);

	cairo_show_page(ctx);
	cairo_surface_finish(csurface);
	// Destroying cairo context
	cairo_destroy(ctx);


	// Destroying PDF surface
	cairo_surface_destroy(csurface);
}




