/*
 * mda.cpp
 *
 *  Created on: 19 Jul 2013
 *      Author: Carsten Kemena
 *
 * This file is part of MDAT.
 *
 * MDAT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MDAT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with MDAT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

// C++ header
#include <algorithm>
#include <cstdlib>
#include <functional>

// Boost header
#include <boost/program_options.hpp>
#include <boost/lexical_cast.hpp>

// MDAT header
#include "../../lib/Sequence/Sequence.hpp"
#include "../../lib/Sequence/SplitSet.hpp"
#include "../../lib/utils/CRS_Mat.hpp"
#include "../../lib/align/domain_aln.hpp"
#include "../../lib/align/seq_align.hpp"
#include "../../lib/align/consistency_aln.hpp"
#include "../../lib/clustering/Tree.hpp"
#include "../../lib/utils/filesystem.h"

// MDA header
#include "options.hpp"
#include "../../../cmake_config/config.h" //This file is generated by CMAKE from conig.h.in and defines HAVE_CAIRO_H (or not) depending on availability.

#ifdef HAVE_CAIRO_H
	#include "mda_pdf_out.hpp"
#endif

using namespace MDAT;
using namespace std;
using namespace boost;
namespace po = boost::program_options;




Matrix<int>*
read_similarity_matrix(const char *mat_f)
{
	FILE *mat_F = fopen(mat_f, "r");
	if (!mat_F)
		return NULL;
	const int LINE_LENGTH=1001;
	char line[LINE_LENGTH];
	while (fgets(line, LINE_LENGTH, mat_F) != NULL)
	{
		if (line[0]!='#')
			break;
	}


	Matrix<int> *sim_mat_p =new Matrix<int>(26,26,-999);
	Matrix<int> &sim_mat = *sim_mat_p;
	int key[256];
	/*  A  R  N  D  C  Q  E  G  H  I  L  K  M  F  P  S  T  W  Y  V  B  Z  X  *
	A  4 -1 -2 -2  0 -1 -1  0 -2 -1 -1 -1 -1 -2 -1  1  0 -3 -2  0 -2 -1  0 -4
	R*/
	char *tmp=strtok(line, " ");
	int pos =0;
	key[pos]=toupper(tmp[0])-65;
	while ((tmp=strtok(NULL, " ")) != NULL)
	{
		key[++pos]=toupper(tmp[0])-65;
		if ((key[pos]<0) || (key[pos]>25))
			key[pos]=-1;
	}

	int row_id;
	int col_id;
	while (fgets(line, LINE_LENGTH, mat_F) != NULL)
	{
		col_id=-1;
		tmp=strtok(line, " \n");
		if (tmp[0]=='*')
			continue;
		row_id=toupper(tmp[0])-65;
		while ((tmp=strtok(NULL, " \n")) != NULL)
		{
			if (key[++col_id] != -1)
				sim_mat[row_id][key[col_id]]=sim_mat[key[col_id]][row_id]=atoi(tmp);
		}
	}

	fclose(mat_F);
	return sim_mat_p;
}

/**
 * \brief Functor to identify if a Domain is a gap.
 */
struct DomainArchitectureGap
{
  bool operator() (const DomainArchitecture &arch, size_t i) const
  {
	  return (arch[i].id()==-1);
  }
} DomainArchitectureGap_obj;


/**
 * \brief Sets the scores for each pair of domains in the set.
 * @param set The Domain set.
 * @param opts The place to save the scores in.
 */
void
fill_matches (const ProteinSequenceSet<Default> &set, Opt_DA &opts)
{
	CRS_Mat<short> domain_mat(opts.dmpm_f);
	int n_domains = set.n_domains();
	opts.matches.resize(n_domains, n_domains);
	int i,j, score;
	for (i=0; i<n_domains; ++i)
	{
		opts.matches[i][i]=opts.match;
		for (j=i+1; j<n_domains; ++j)
		{
			score = domain_mat.value(atoi(&set.domain_name(i).c_str()[2]), atoi(&set.domain_name(j).c_str()[2]));
			opts.matches[i][j]=score;
			opts.matches[j][i]=score;
		}
	}
}


int
main(int argc, char *argv[])
{

	Opt_DA mda_opts;
	Opt_msa msa_opts;
	Opt_general general_opts;
	Opt_dom dom_opts;
	#ifdef HAVE_CAIRO_H
		Opt_graphic graphic_opts;
	#endif
	po::options_description all("MDA v1.0  Copyright (C) 2013,2014  Carsten Kemena\nThis program comes with ABSOLUTELY NO WARRANTY;\n\nAllowed options are displayed below.");
	po::options_description general("General options");
	general.add_options()
		("help,h", "Produces this help message")
		("sequence,s", po::value<string>(&general_opts.sequence_f)->required(), "Sequence file")
		("domain_f,d", po::value<string>(&general_opts.domain_f)->required(), "Domain file")
		("msa_out,o", po::value<string>(&general_opts.msa_out_f), "File to write the MSA to.")
		("mda_out,a", po::value<string>(&general_opts.mda_out_f), "File to write the MDA to.")
		("format,f", po::value<string>(&general_opts.out_msa_format)->default_value("fasta"), "Output format for the MSA")
		("threads,t", po::value<int>(&general_opts.n_threads)->default_value(1), "Number of threads to use")
	;

	po::options_description dom_options("Domain options");
	dom_options.add_options()
		("no_repeat_collapsing", po::value<bool>(&dom_opts.no_rep_col)->zero_tokens()->default_value(false), "Do not collapses repeateas for the alignment into a single domain")
		("domain_program,p", po::value<string>(&general_opts.domain_program), "Domain detection program")
		("domain_hmms", po::value<string>(&general_opts.domain_database), "Domain database path")
	//	("refine_boundaries", po::value<bool>(&dom_opts.refine_boundaries)->zero_tokens()->default_value(false), "Reestimates the domain boundaries")
		("use_envelope", po::value<bool>(&dom_opts.use_envelope)->zero_tokens()->default_value(false), "Uses the envelope to calculate the MDA")
		("no_merging", po::value<bool>(&dom_opts.no_merging)->zero_tokens()->default_value(false), "Splitted domains are not merged")
	;

	po::options_description msa_options("MSA options");
	msa_options.add_options()
		("gop", po::value<int>(&msa_opts.gop)->default_value(-11), "gap opening penalty")
		("gep", po::value<int>(&msa_opts.gep)->default_value(-1), "gap extension penalty")
		("sim_mat", po::value<string>(&msa_opts.simmat_name)->default_value("BLOSUM62.txt"), "Similarity matrix to use")
		("relax,r", po::value<bool>(&msa_opts.relax)->zero_tokens()->default_value(false), "Relax MDA constraints")
	;

	po::options_description mda_options("Domain alignment options");
	mda_options.add_options()
		("mda_match", po::value<int>(&mda_opts.match)->default_value(500),"Value used for the exact_match of two domains")
		("mda_mismatch", po::value<int>(&mda_opts.mismatch)->default_value(-100), "")
		("mda_gop", po::value<int>(&mda_opts.gop)->default_value(-50), "Domain gap opening cost")
		("mda_gep", po::value<int>(&mda_opts.gep)->default_value(-10), "Domain gap extension cost")
		("mda_term_gop", po::value<int>(&mda_opts.terminal_gop)->default_value(0), "Terminal domain gap opening cost")
		("mda_term_gep", po::value<int>(&mda_opts.terminal_gep)->default_value(-5), "Terminal domain gap extension cost")
		("mda_l_prob", po::value<int>(&mda_opts.l_prop)->default_value(1), "")
		("scale", po::value<int>(&mda_opts.scale)->default_value(100), "")
		("dmpm_f", po::value<string>(&mda_opts.dmpm_f), "Domain match probability matrix")
	;

	all.add(general).add(dom_options).add(mda_options).add(msa_options);

#ifdef HAVE_CAIRO_H
	po::options_description graphic_options("Graphic output options");
	graphic_options.add_options()
		("graphical_out", po::value<string>(&graphic_opts.graphic_f)->default_value(""), "Produces graphical output of MDA")
		("show_sim_mat", po::value<bool>(&graphic_opts.show_sim_mat)->zero_tokens()->default_value(false), "Shows the similarity matrix")
		("show_seq_names", po::value<bool>(&graphic_opts.show_sequence_names)->zero_tokens()->default_value(false), "Shows the sequence names")
		("color_by_similarity", po::value<bool>(&graphic_opts.color_by_similarity)->zero_tokens()->default_value(false), "Colors the MDA according to similarity")
	;
	all.add(graphic_options);
#endif


	try
	{
		po::variables_map vm;
		po::store(po::command_line_parser(argc, argv).options(all).run(), vm);
		if (vm.count("help"))
		{
			cout << all<< "\n";
			return EXIT_SUCCESS;
		}
		po::notify(vm);
	}
	catch (boost::program_options::error &e)
	{
		cerr << "An error occurred parsing the commandline: \n";
		cerr << e.what() << endl;
		cerr << "Use -h/--help for information of available options\n";
		return EXIT_FAILURE;
	}
	ProteinSequenceSet<Default> seqSet;
	try
	{
		seqSet.read(general_opts.sequence_f, true);
	}
	catch (std::exception &e)
	{
		cerr << "Could not read sequences from file " << general_opts.sequence_f << ": " << e.what() << endl;
		exit(EXIT_FAILURE);
	}

	// In case no domain file is given, determine domains using HMMScan
	if (general_opts.domain_f.empty())
	{
		string command("");
		general_opts.domain_f=general_opts.sequence_f + ".doms";
		if (general_opts.domain_program=="hmmscan")
			command = general_opts.domain_program + " --cpu " + boost::lexical_cast<string>(general_opts.n_threads) + " --domtblout " + general_opts.domain_f + " "  + general_opts.domain_database + " " + general_opts.sequence_f + " >/dev/null";
		else
		{
			if (general_opts.domain_program=="pfam_scan")
				command = "pfam_scan.pl -cpu "  + boost::lexical_cast<string>(general_opts.n_threads) + " -outfile " + general_opts.domain_f + " -fasta " +   general_opts.sequence_f + " -dir " + general_opts.domain_database + " >/dev/null";
			else
			{
				cerr << "Unsupported domain detection program " << std::endl;
				exit(EXIT_FAILURE);
			}
		}
		if (system(command.c_str()))
		{
			cerr << "Failure running domain detection program" << endl;
			exit(EXIT_FAILURE);
		}
	}

	try
	{
		seqSet.add_pfam_domains(general_opts.domain_f);
	}
	catch (std::exception &e)
	{
		cerr << "Could not read domain information from: "<< general_opts.domain_f + " " + e.what() << endl;
		exit(EXIT_FAILURE);
	}

	if (mda_opts.dmpm_f.empty())
	{
		mda_opts.dmpm_f = getenv("HOME");
		mda_opts.dmpm_f.append("/.mdat/dmpm_pfam27_crs.dat");
	}

	try
	{
		fill_matches(seqSet, mda_opts);
	}
	catch (std::exception &e)
	{
		cerr << "Could not read domain similarity matrix from: "<< mda_opts.dmpm_f << endl;
		exit(EXIT_FAILURE);
	}

	size_t n_seqs = seqSet.n_seqs();
	if (dom_opts.use_envelope)
	{
		size_t j, n_domains;
		for (size_t i=0; i<n_seqs; ++i)
		{
			n_domains = seqSet[i].n_domains();
			DomainArchitecture &doms = seqSet[i].dom_archi();
			for (j=0; j<n_domains; ++j)
			{
				doms[j].seq_start(doms[j].env_start());
				doms[j].seq_end(doms[j].env_end());
			}
		}
	}

	char cleaning_opts=CleanNested;
	if (!dom_opts.no_merging)
		cleaning_opts|=CleanMerged;
	seqSet.clean_up_domains(cleaning_opts);

	if (!dom_opts.no_rep_col)
	{
		for (size_t i=0; i<n_seqs; ++i)
			seqSet[i].dom_archi().collapse_repeats();
	}
	// TODO refine_boundaries is not yet implemented!
	//if ((!dom_opts.use_envelope) && (dom_opts.refine_boundaries))
	//	seqSet.refine_boundaries();

	// make MDA
	seqSet.extract_architectures();
	size_t n_archis=seqSet.n_architectures();


	Tree architecture_guide_tree;
	if (n_archis>1)
	{
		Library<DomainArchitectureSet > lib(seqSet.dom_archis());
		Matrix<float> *dist_mat=all_rads_pair_align_archi(lib, seqSet, mda_opts, general_opts.n_threads);

		// produce progressive alignment
		lib.relax(static_cast<const double& (*)(const double&, const double&)>(std::min), general_opts.n_threads);
		// Calculate guide tree
		// Turn similarity matrix into distance matrix
		int max_val=INT_MIN;
		size_t i,j;
		for (i=0; i<n_archis; ++i)
			for (j=i+1; j<n_archis; ++j)
				if ((*dist_mat)[i][j]>max_val)
					max_val=(*dist_mat)[i][j];
		for (i=0; i<n_archis; ++i)
		{
			for (j=i+1; j<n_archis; ++j)
			{
				(*dist_mat)[i][j] = (*dist_mat)[i][j]*(-1)+max_val;
				(*dist_mat)[j][i] = (*dist_mat)[i][j];
			}
		}
		vector<string> names(n_archis, "");
		vector<int> n_members;
		for (i=0; i<n_archis; ++i)
			n_members.push_back(seqSet.dom_archis()[i].n_members());
		architecture_guide_tree.upgma(*dist_mat, names, n_members);

		delete dist_mat;
		progressive_consistency_align(lib, architecture_guide_tree, seqSet.dom_archis(), DomainArchitectureGap_obj);
	}

	refinePercentage(seqSet.dom_archis());
	seqSet.dom_archis().sort("gap_pattern");

	#ifdef HAVE_CAIRO_H
	if (!graphic_opts.graphic_f.empty())
	{
		string graphic_out_f = graphic_opts.graphic_f;
		write_mda_as_pdf(seqSet, mda_opts.matches,graphic_opts);
	}
	#endif
	if (!general_opts.mda_out_f.empty())
	{
		try
		{
			seqSet.write_domArchitecture(general_opts.mda_out_f);
		}
		catch (std::exception &e)
		{
			cerr << "Could not write to file " << general_opts.mda_out_f << ": " << e.what() << endl;
			exit(1);
		}
	}

	// make MSA using MDA
	if (!general_opts.msa_out_f.empty())
	{
		msa_opts.sim_mat = read_similarity_matrix(msa_opts.simmat_name.c_str());
		if (msa_opts.sim_mat==NULL)
		{
			string matrix_path = getenv("HOME");
			matrix_path.append("/.mdat/"+msa_opts.simmat_name);
			msa_opts.sim_mat = read_similarity_matrix(matrix_path.c_str());
			if (msa_opts.sim_mat==NULL)
			{
				cerr << "Could not read similarity matrix file " +matrix_path << endl;
				exit(EXIT_FAILURE);
			}
		}
		seqSet.dom_archis().sort("id");

		seqSet.clean_up_domains(CleanOverlap);
		msa(seqSet, architecture_guide_tree, *msa_opts.sim_mat, msa_opts.gop, msa_opts.gep, msa_opts.relax, general_opts.n_threads);
		delete msa_opts.sim_mat;

		// write output to file
		try
		{
			seqSet.sort("id");
			seqSet.write(general_opts.msa_out_f, general_opts.out_msa_format);
		}
		catch (std::exception &e)
		{
			cerr << "Could not write to file: " << general_opts.msa_out_f << ": " << e.what() << endl;
			exit(1);
		}
	}
	return EXIT_SUCCESS;
}
